'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var view = require('@codemirror/view');
var state = require('@codemirror/state');
var history = require('@codemirror/history');
var fold = require('@codemirror/fold');
var language = require('@codemirror/language');
var commands = require('@codemirror/commands');
var matchbrackets = require('@codemirror/matchbrackets');
var closebrackets = require('@codemirror/closebrackets');
var search = require('@codemirror/search');
var autocomplete = require('@codemirror/autocomplete');
var comment = require('@codemirror/comment');
var rectangularSelection = require('@codemirror/rectangular-selection');
var highlight = require('@codemirror/highlight');
var lint = require('@codemirror/lint');
var React = require('react');
var dequal = require('dequal');
var langJavascript = require('@codemirror/lang-javascript');
var langJson = require('@codemirror/lang-json');
var langSql = require('@codemirror/lang-sql');
var langPython = require('@codemirror/lang-python');
var gutter = require('@codemirror/gutter');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var React__default = /*#__PURE__*/_interopDefaultLegacy(React);

/// This is an extension value that just pulls together a whole lot of
/// extensions that you might want in a basic editor. It is meant as a
/// convenient helper to quickly set up CodeMirror without installing
/// and importing a lot of packages.
///
/// Specifically, it includes...
///
///  - [the default command bindings](#commands.defaultKeymap)
///  - [line numbers](#gutter.lineNumbers)
///  - [special character highlighting](#view.highlightSpecialChars)
///  - [the undo history](#history.history)
///  - [a fold gutter](#fold.foldGutter)
///  - [custom selection drawing](#view.drawSelection)
///  - [multiple selections](#state.EditorState^allowMultipleSelections)
///  - [reindentation on input](#language.indentOnInput)
///  - [the default highlight style](#highlight.defaultHighlightStyle) (as fallback)
///  - [bracket matching](#matchbrackets.bracketMatching)
///  - [bracket closing](#closebrackets.closeBrackets)
///  - [autocompletion](#autocomplete.autocompletion)
///  - [rectangular selection](#rectangular-selection.rectangularSelection)
///  - [active line highlighting](#view.highlightActiveLine)
///  - [selection match highlighting](#search.highlightSelectionMatches)
///  - [search](#search.searchKeymap)
///  - [commenting](#comment.commentKeymap)
///  - [linting](#lint.lintKeymap)
///
/// (You'll probably want to add some language package to your setup
/// too.)
///
/// This package does not allow customization. The idea is that, once
/// you decide you want to configure your editor more precisely, you
/// take this package's source (which is just a bunch of imports and
/// an array literal), copy it into your own code, and adjust it as
/// desired.
const basicSetup = [
    view.highlightSpecialChars(),
    history.history(),
    fold.codeFolding({}),
    fold.foldGutter(),
    view.drawSelection(),
    state.EditorState.allowMultipleSelections.of(true),
    language.indentOnInput(),
    highlight.defaultHighlightStyle.fallback,
    matchbrackets.bracketMatching(),
    closebrackets.closeBrackets(),
    autocomplete.autocompletion(),
    rectangularSelection.rectangularSelection(),
    view.highlightActiveLine(),
    search.highlightSelectionMatches(),
    view.keymap.of([
        ...closebrackets.closeBracketsKeymap,
        ...commands.defaultKeymap,
        ...search.searchKeymap,
        ...history.historyKeymap,
        ...fold.foldKeymap,
        ...comment.commentKeymap,
        ...autocomplete.completionKeymap,
        ...lint.lintKeymap,
        {
            key: "Tab",
            run: (view) => {
                view.dispatch({
                    changes: [
                        {
                            from: view.state.selection.main.head,
                            insert: " ".repeat(view.state.tabSize),
                        },
                    ],
                    selection: {
                        anchor: view.state.selection.main.head + view.state.tabSize,
                        head: view.state.selection.main.head + view.state.tabSize,
                    },
                });
                return true;
            },
        },
    ]),
];

/**
 * hooks for mount callback
 * @param cb
 * @returns
 */
function useMount(cb) {
    React.useEffect(() => {
        cb && cb();
    }, []);
    return null;
}
/**
 * hook for unmount callback
 * @param cb
 * @returns
 */
function useUnmount(cb) {
    React.useEffect(() => {
        return () => {
            cb();
        };
    }, []);
    return null;
}

const panelConfig = /*@__PURE__*/state.Facet.define({
    combine(configs) {
        let topContainer, bottomContainer;
        for (let c of configs) {
            topContainer = topContainer || c.topContainer;
            bottomContainer = bottomContainer || c.bottomContainer;
        }
        return { topContainer, bottomContainer };
    }
});
const panelPlugin = /*@__PURE__*/view.ViewPlugin.fromClass(class {
    constructor(view) {
        this.input = view.state.facet(showPanel);
        this.specs = this.input.filter(s => s);
        this.panels = this.specs.map(spec => spec(view));
        let conf = view.state.facet(panelConfig);
        this.top = new PanelGroup(view, true, conf.topContainer);
        this.bottom = new PanelGroup(view, false, conf.bottomContainer);
        this.top.sync(this.panels.filter(p => p.top));
        this.bottom.sync(this.panels.filter(p => !p.top));
        for (let p of this.panels) {
            p.dom.classList.add("cm-panel");
            // FIXME drop on next breaking release
            if (p.class)
                p.dom.classList.add(p.class);
            if (p.mount)
                p.mount();
        }
    }
    update(update) {
        let conf = update.state.facet(panelConfig);
        if (this.top.container != conf.topContainer) {
            this.top.sync([]);
            this.top = new PanelGroup(update.view, true, conf.topContainer);
        }
        if (this.bottom.container != conf.bottomContainer) {
            this.bottom.sync([]);
            this.bottom = new PanelGroup(update.view, false, conf.bottomContainer);
        }
        this.top.syncClasses();
        this.bottom.syncClasses();
        let input = update.state.facet(showPanel);
        if (input != this.input) {
            let specs = input.filter(x => x);
            let panels = [], top = [], bottom = [], mount = [];
            for (let spec of specs) {
                let known = this.specs.indexOf(spec), panel;
                if (known < 0) {
                    panel = spec(update.view);
                    mount.push(panel);
                }
                else {
                    panel = this.panels[known];
                    if (panel.update)
                        panel.update(update);
                }
                panels.push(panel);
                (panel.top ? top : bottom).push(panel);
            }
            this.specs = specs;
            this.panels = panels;
            this.top.sync(top);
            this.bottom.sync(bottom);
            for (let p of mount) {
                p.dom.classList.add("cm-panel");
                // FIXME drop on next breaking release
                if (p.class)
                    p.dom.classList.add(p.class);
                if (p.mount)
                    p.mount();
            }
        }
        else {
            for (let p of this.panels)
                if (p.update)
                    p.update(update);
        }
    }
    destroy() {
        this.top.sync([]);
        this.bottom.sync([]);
    }
}, {
    provide: /*@__PURE__*/view.PluginField.scrollMargins.from(value => ({ top: value.top.scrollMargin(), bottom: value.bottom.scrollMargin() }))
});
class PanelGroup {
    constructor(view, top, container) {
        this.view = view;
        this.top = top;
        this.container = container;
        this.dom = undefined;
        this.classes = "";
        this.panels = [];
        this.syncClasses();
    }
    sync(panels) {
        this.panels = panels;
        this.syncDOM();
    }
    syncDOM() {
        if (this.panels.length == 0) {
            if (this.dom) {
                this.dom.remove();
                this.dom = undefined;
            }
            return;
        }
        if (!this.dom) {
            this.dom = document.createElement("div");
            this.dom.className = this.top ? "cm-panels cm-panels-top" : "cm-panels cm-panels-bottom";
            this.dom.style[this.top ? "top" : "bottom"] = "0";
            let parent = this.container || this.view.dom;
            parent.insertBefore(this.dom, this.top ? parent.firstChild : null);
        }
        let curDOM = this.dom.firstChild;
        for (let panel of this.panels) {
            if (panel.dom.parentNode == this.dom) {
                while (curDOM != panel.dom)
                    curDOM = rm(curDOM);
                curDOM = curDOM.nextSibling;
            }
            else {
                this.dom.insertBefore(panel.dom, curDOM);
            }
        }
        while (curDOM)
            curDOM = rm(curDOM);
    }
    scrollMargin() {
        return !this.dom || this.container ? 0
            : Math.max(0, this.top ?
                this.dom.getBoundingClientRect().bottom - Math.max(0, this.view.scrollDOM.getBoundingClientRect().top) :
                Math.min(innerHeight, this.view.scrollDOM.getBoundingClientRect().bottom) - this.dom.getBoundingClientRect().top);
    }
    syncClasses() {
        if (!this.container || this.classes == this.view.themeClasses)
            return;
        for (let cls of this.classes.split(" "))
            if (cls)
                this.container.classList.remove(cls);
        for (let cls of (this.classes = this.view.themeClasses).split(" "))
            if (cls)
                this.container.classList.add(cls);
    }
}
function rm(node) {
    let next = node.nextSibling;
    node.remove();
    return next;
}
const baseTheme$2 = /*@__PURE__*/view.EditorView.baseTheme({
    ".cm-panels": {
        boxSizing: "border-box",
        position: "sticky",
        left: 0,
        right: 0
    },
    "&light .cm-panels": {
        backgroundColor: "#f5f5f5",
        color: "black"
    },
    "&light .cm-panels-top": {
        borderBottom: "1px solid #ddd"
    },
    "&light .cm-panels-bottom": {
        borderTop: "1px solid #ddd"
    },
    "&dark .cm-panels": {
        backgroundColor: "#333338",
        color: "white"
    }
});
/**
Opening a panel is done by providing a constructor function for
the panel through this facet. (The panel is closed again when its
constructor is no longer provided.) Values of `null` are ignored.
*/
const showPanel = /*@__PURE__*/state.Facet.define({
    enables: [panelPlugin, baseTheme$2]
});

class Viewbox {
    constructor(parent) {
        this.node = document.createElement("div");
    }
    appendStyle(cssStyle) {
        for (const prop in cssStyle) {
            this.node.style[prop] = cssStyle[prop];
        }
    }
    scroll(scrollTop) {
        this.node.style.transform = `translateY(${Math.round(scrollTop)}px)`;
    }
    hide() {
        this.node.style.visibility = "0";
    }
    show() {
        this.node.style.visibility = "1";
    }
    resize(width, height) {
        this.node.style.width = String(width) + "px";
        this.node.style.height = String(height) + "px";
    }
    attach(parent) {
        parent.appendChild(this.node);
    }
}
class CanvasElement {
    constructor(parent) {
        this.initLayer();
        this.attachCanvasLayer(parent);
    }
    attachCanvasLayer(parent) {
        parent.appendChild(this.node);
    }
    initLayer() {
        this.node = document.createElement("canvas");
        this.ctx = this.node.getContext("2d");
        this.node.className = "cm-minimap-layer";
        this.node.style.zIndex = "0";
    }
    resize(width, height) {
        this.node.style.width = String(width) + "px";
        this.node.style.height = String(height) + "px";
    }
}
class MinimapElement {
    constructor(view) {
        this.view = view;
        this.node = document.createElement("div");
        this.node.className = "cm-minimap";
        this.view.dom.style.overflow = "hidden";
    }
    resize(width, height) {
        if (width != this.width)
            this.node.style.width = String(width) + "px";
        if (height != this.height)
            this.node.style.height = String(height) + "px";
        this.width = width;
        this.height = height;
    }
    setSide(side) {
        this.view.dom.style.flexDirection =
            side === "left" ? "row-reverse" : "row";
    }
    setBackground(bg) {
        this.node.style.background = bg;
    }
}

/// The default maximum length of a `TreeBuffer` node.
const DefaultBufferLength = 1024;
let nextPropID = 0;
const CachedNode = new WeakMap();
/// Each [node type](#tree.NodeType) can have metadata associated with
/// it in props. Instances of this class represent prop names.
class NodeProp {
    /// Create a new node prop type. You can optionally pass a
    /// `deserialize` function.
    constructor({ deserialize } = {}) {
        this.id = nextPropID++;
        this.deserialize = deserialize || (() => {
            throw new Error("This node type doesn't define a deserialize function");
        });
    }
    /// Create a string-valued node prop whose deserialize function is
    /// the identity function.
    static string() { return new NodeProp({ deserialize: str => str }); }
    /// Create a number-valued node prop whose deserialize function is
    /// just `Number`.
    static number() { return new NodeProp({ deserialize: Number }); }
    /// Creates a boolean-valued node prop whose deserialize function
    /// returns true for any input.
    static flag() { return new NodeProp({ deserialize: () => true }); }
    /// Store a value for this prop in the given object. This can be
    /// useful when building up a prop object to pass to the
    /// [`NodeType`](#tree.NodeType) constructor. Returns its first
    /// argument.
    set(propObj, value) {
        propObj[this.id] = value;
        return propObj;
    }
    /// This is meant to be used with
    /// [`NodeSet.extend`](#tree.NodeSet.extend) or
    /// [`Parser.withProps`](#lezer.Parser.withProps) to compute prop
    /// values for each node type in the set. Takes a [match
    /// object](#tree.NodeType^match) or function that returns undefined
    /// if the node type doesn't get this prop, and the prop's value if
    /// it does.
    add(match) {
        if (typeof match != "function")
            match = NodeType.match(match);
        return (type) => {
            let result = match(type);
            return result === undefined ? null : [this, result];
        };
    }
}
/// Prop that is used to describe matching delimiters. For opening
/// delimiters, this holds an array of node names (written as a
/// space-separated string when declaring this prop in a grammar)
/// for the node types of closing delimiters that match it.
NodeProp.closedBy = new NodeProp({ deserialize: str => str.split(" ") });
/// The inverse of [`openedBy`](#tree.NodeProp^closedBy). This is
/// attached to closing delimiters, holding an array of node names
/// of types of matching opening delimiters.
NodeProp.openedBy = new NodeProp({ deserialize: str => str.split(" ") });
/// Used to assign node types to groups (for example, all node
/// types that represent an expression could be tagged with an
/// `"Expression"` group).
NodeProp.group = new NodeProp({ deserialize: str => str.split(" ") });
const noProps = Object.create(null);
/// Each node in a syntax tree has a node type associated with it.
class NodeType {
    /// @internal
    constructor(
    /// The name of the node type. Not necessarily unique, but if the
    /// grammar was written properly, different node types with the
    /// same name within a node set should play the same semantic
    /// role.
    name, 
    /// @internal
    props, 
    /// The id of this node in its set. Corresponds to the term ids
    /// used in the parser.
    id, 
    /// @internal
    flags = 0) {
        this.name = name;
        this.props = props;
        this.id = id;
        this.flags = flags;
    }
    static define(spec) {
        let props = spec.props && spec.props.length ? Object.create(null) : noProps;
        let flags = (spec.top ? 1 /* Top */ : 0) | (spec.skipped ? 2 /* Skipped */ : 0) |
            (spec.error ? 4 /* Error */ : 0) | (spec.name == null ? 8 /* Anonymous */ : 0);
        let type = new NodeType(spec.name || "", props, spec.id, flags);
        if (spec.props)
            for (let src of spec.props) {
                if (!Array.isArray(src))
                    src = src(type);
                if (src)
                    src[0].set(props, src[1]);
            }
        return type;
    }
    /// Retrieves a node prop for this type. Will return `undefined` if
    /// the prop isn't present on this node.
    prop(prop) { return this.props[prop.id]; }
    /// True when this is the top node of a grammar.
    get isTop() { return (this.flags & 1 /* Top */) > 0; }
    /// True when this node is produced by a skip rule.
    get isSkipped() { return (this.flags & 2 /* Skipped */) > 0; }
    /// Indicates whether this is an error node.
    get isError() { return (this.flags & 4 /* Error */) > 0; }
    /// When true, this node type doesn't correspond to a user-declared
    /// named node, for example because it is used to cache repetition.
    get isAnonymous() { return (this.flags & 8 /* Anonymous */) > 0; }
    /// Returns true when this node's name or one of its
    /// [groups](#tree.NodeProp^group) matches the given string.
    is(name) {
        if (typeof name == 'string') {
            if (this.name == name)
                return true;
            let group = this.prop(NodeProp.group);
            return group ? group.indexOf(name) > -1 : false;
        }
        return this.id == name;
    }
    /// Create a function from node types to arbitrary values by
    /// specifying an object whose property names are node or
    /// [group](#tree.NodeProp^group) names. Often useful with
    /// [`NodeProp.add`](#tree.NodeProp.add). You can put multiple
    /// names, separated by spaces, in a single property name to map
    /// multiple node names to a single value.
    static match(map) {
        let direct = Object.create(null);
        for (let prop in map)
            for (let name of prop.split(" "))
                direct[name] = map[prop];
        return (node) => {
            for (let groups = node.prop(NodeProp.group), i = -1; i < (groups ? groups.length : 0); i++) {
                let found = direct[i < 0 ? node.name : groups[i]];
                if (found)
                    return found;
            }
        };
    }
}
/// An empty dummy node type to use when no actual type is available.
NodeType.none = new NodeType("", Object.create(null), 0, 8 /* Anonymous */);
/// A node set holds a collection of node types. It is used to
/// compactly represent trees by storing their type ids, rather than a
/// full pointer to the type object, in a number array. Each parser
/// [has](#lezer.Parser.nodeSet) a node set, and [tree
/// buffers](#tree.TreeBuffer) can only store collections of nodes
/// from the same set. A set can have a maximum of 2**16 (65536)
/// node types in it, so that the ids fit into 16-bit typed array
/// slots.
class NodeSet {
    /// Create a set with the given types. The `id` property of each
    /// type should correspond to its position within the array.
    constructor(
    /// The node types in this set, by id.
    types) {
        this.types = types;
        for (let i = 0; i < types.length; i++)
            if (types[i].id != i)
                throw new RangeError("Node type ids should correspond to array positions when creating a node set");
    }
    /// Create a copy of this set with some node properties added. The
    /// arguments to this method should be created with
    /// [`NodeProp.add`](#tree.NodeProp.add).
    extend(...props) {
        let newTypes = [];
        for (let type of this.types) {
            let newProps = null;
            for (let source of props) {
                let add = source(type);
                if (add) {
                    if (!newProps)
                        newProps = Object.assign({}, type.props);
                    add[0].set(newProps, add[1]);
                }
            }
            newTypes.push(newProps ? new NodeType(type.name, newProps, type.id, type.flags) : type);
        }
        return new NodeSet(newTypes);
    }
}
/// A piece of syntax tree. There are two ways to approach these
/// trees: the way they are actually stored in memory, and the
/// convenient way.
///
/// Syntax trees are stored as a tree of `Tree` and `TreeBuffer`
/// objects. By packing detail information into `TreeBuffer` leaf
/// nodes, the representation is made a lot more memory-efficient.
///
/// However, when you want to actually work with tree nodes, this
/// representation is very awkward, so most client code will want to
/// use the `TreeCursor` interface instead, which provides a view on
/// some part of this data structure, and can be used to move around
/// to adjacent nodes.
class Tree {
    /// Construct a new tree. You usually want to go through
    /// [`Tree.build`](#tree.Tree^build) instead.
    constructor(type, 
    /// The tree's child nodes. Children small enough to fit in a
    /// `TreeBuffer will be represented as such, other children can be
    /// further `Tree` instances with their own internal structure.
    children, 
    /// The positions (offsets relative to the start of this tree) of
    /// the children.
    positions, 
    /// The total length of this tree
    length) {
        this.type = type;
        this.children = children;
        this.positions = positions;
        this.length = length;
    }
    /// @internal
    toString() {
        let children = this.children.map(c => c.toString()).join();
        return !this.type.name ? children :
            (/\W/.test(this.type.name) && !this.type.isError ? JSON.stringify(this.type.name) : this.type.name) +
                (children.length ? "(" + children + ")" : "");
    }
    /// Get a [tree cursor](#tree.TreeCursor) rooted at this tree. When
    /// `pos` is given, the cursor is [moved](#tree.TreeCursor.moveTo)
    /// to the given position and side.
    cursor(pos, side = 0) {
        let scope = (pos != null && CachedNode.get(this)) || this.topNode;
        let cursor = new TreeCursor(scope);
        if (pos != null) {
            cursor.moveTo(pos, side);
            CachedNode.set(this, cursor._tree);
        }
        return cursor;
    }
    /// Get a [tree cursor](#tree.TreeCursor) that, unlike regular
    /// cursors, doesn't skip [anonymous](#tree.NodeType.isAnonymous)
    /// nodes.
    fullCursor() {
        return new TreeCursor(this.topNode, true);
    }
    /// Get a [syntax node](#tree.SyntaxNode) object for the top of the
    /// tree.
    get topNode() {
        return new TreeNode(this, 0, 0, null);
    }
    /// Get the [syntax node](#tree.SyntaxNode) at the given position.
    /// If `side` is -1, this will move into nodes that end at the
    /// position. If 1, it'll move into nodes that start at the
    /// position. With 0, it'll only enter nodes that cover the position
    /// from both sides.
    resolve(pos, side = 0) {
        return this.cursor(pos, side).node;
    }
    /// Iterate over the tree and its children, calling `enter` for any
    /// node that touches the `from`/`to` region (if given) before
    /// running over such a node's children, and `leave` (if given) when
    /// leaving the node. When `enter` returns `false`, the given node
    /// will not have its children iterated over (or `leave` called).
    iterate(spec) {
        let { enter, leave, from = 0, to = this.length } = spec;
        for (let c = this.cursor();;) {
            let mustLeave = false;
            if (c.from <= to && c.to >= from && (c.type.isAnonymous || enter(c.type, c.from, c.to) !== false)) {
                if (c.firstChild())
                    continue;
                if (!c.type.isAnonymous)
                    mustLeave = true;
            }
            for (;;) {
                if (mustLeave && leave)
                    leave(c.type, c.from, c.to);
                mustLeave = c.type.isAnonymous;
                if (c.nextSibling())
                    break;
                if (!c.parent())
                    return;
                mustLeave = true;
            }
        }
    }
    /// Balance the direct children of this tree.
    balance(maxBufferLength = DefaultBufferLength) {
        return this.children.length <= BalanceBranchFactor ? this
            : balanceRange(this.type, NodeType.none, this.children, this.positions, 0, this.children.length, 0, maxBufferLength, this.length, 0);
    }
    /// Build a tree from a postfix-ordered buffer of node information,
    /// or a cursor over such a buffer.
    static build(data) { return buildTree(data); }
}
/// The empty tree
Tree.empty = new Tree(NodeType.none, [], [], 0);
// For trees that need a context hash attached, we're using this
// kludge which assigns an extra property directly after
// initialization (creating a single new object shape).
function withHash(tree, hash) {
    if (hash)
        tree.contextHash = hash;
    return tree;
}
/// Tree buffers contain (type, start, end, endIndex) quads for each
/// node. In such a buffer, nodes are stored in prefix order (parents
/// before children, with the endIndex of the parent indicating which
/// children belong to it)
class TreeBuffer {
    /// Create a tree buffer @internal
    constructor(
    /// @internal
    buffer, 
    // The total length of the group of nodes in the buffer.
    length, 
    /// @internal
    set, type = NodeType.none) {
        this.buffer = buffer;
        this.length = length;
        this.set = set;
        this.type = type;
    }
    /// @internal
    toString() {
        let result = [];
        for (let index = 0; index < this.buffer.length;) {
            result.push(this.childString(index));
            index = this.buffer[index + 3];
        }
        return result.join(",");
    }
    /// @internal
    childString(index) {
        let id = this.buffer[index], endIndex = this.buffer[index + 3];
        let type = this.set.types[id], result = type.name;
        if (/\W/.test(result) && !type.isError)
            result = JSON.stringify(result);
        index += 4;
        if (endIndex == index)
            return result;
        let children = [];
        while (index < endIndex) {
            children.push(this.childString(index));
            index = this.buffer[index + 3];
        }
        return result + "(" + children.join(",") + ")";
    }
    /// @internal
    findChild(startIndex, endIndex, dir, after) {
        let { buffer } = this, pick = -1;
        for (let i = startIndex; i != endIndex; i = buffer[i + 3]) {
            if (after != -100000000 /* None */) {
                let start = buffer[i + 1], end = buffer[i + 2];
                if (dir > 0) {
                    if (end > after)
                        pick = i;
                    if (end > after)
                        break;
                }
                else {
                    if (start < after)
                        pick = i;
                    if (end >= after)
                        break;
                }
            }
            else {
                pick = i;
                if (dir > 0)
                    break;
            }
        }
        return pick;
    }
}
class TreeNode {
    constructor(node, from, index, _parent) {
        this.node = node;
        this.from = from;
        this.index = index;
        this._parent = _parent;
    }
    get type() { return this.node.type; }
    get name() { return this.node.type.name; }
    get to() { return this.from + this.node.length; }
    nextChild(i, dir, after, full = false) {
        for (let parent = this;;) {
            for (let { children, positions } = parent.node, e = dir > 0 ? children.length : -1; i != e; i += dir) {
                let next = children[i], start = positions[i] + parent.from;
                if (after != -100000000 /* None */ && (dir < 0 ? start >= after : start + next.length <= after))
                    continue;
                if (next instanceof TreeBuffer) {
                    let index = next.findChild(0, next.buffer.length, dir, after == -100000000 /* None */ ? -100000000 /* None */ : after - start);
                    if (index > -1)
                        return new BufferNode(new BufferContext(parent, next, i, start), null, index);
                }
                else if (full || (!next.type.isAnonymous || hasChild(next))) {
                    let inner = new TreeNode(next, start, i, parent);
                    return full || !inner.type.isAnonymous ? inner : inner.nextChild(dir < 0 ? next.children.length - 1 : 0, dir, after);
                }
            }
            if (full || !parent.type.isAnonymous)
                return null;
            i = parent.index + dir;
            parent = parent._parent;
            if (!parent)
                return null;
        }
    }
    get firstChild() { return this.nextChild(0, 1, -100000000 /* None */); }
    get lastChild() { return this.nextChild(this.node.children.length - 1, -1, -100000000 /* None */); }
    childAfter(pos) { return this.nextChild(0, 1, pos); }
    childBefore(pos) { return this.nextChild(this.node.children.length - 1, -1, pos); }
    nextSignificantParent() {
        let val = this;
        while (val.type.isAnonymous && val._parent)
            val = val._parent;
        return val;
    }
    get parent() {
        return this._parent ? this._parent.nextSignificantParent() : null;
    }
    get nextSibling() {
        return this._parent ? this._parent.nextChild(this.index + 1, 1, -1) : null;
    }
    get prevSibling() {
        return this._parent ? this._parent.nextChild(this.index - 1, -1, -1) : null;
    }
    get cursor() { return new TreeCursor(this); }
    resolve(pos, side = 0) {
        return this.cursor.moveTo(pos, side).node;
    }
    getChild(type, before = null, after = null) {
        let r = getChildren(this, type, before, after);
        return r.length ? r[0] : null;
    }
    getChildren(type, before = null, after = null) {
        return getChildren(this, type, before, after);
    }
    /// @internal
    toString() { return this.node.toString(); }
}
function getChildren(node, type, before, after) {
    let cur = node.cursor, result = [];
    if (!cur.firstChild())
        return result;
    if (before != null)
        while (!cur.type.is(before))
            if (!cur.nextSibling())
                return result;
    for (;;) {
        if (after != null && cur.type.is(after))
            return result;
        if (cur.type.is(type))
            result.push(cur.node);
        if (!cur.nextSibling())
            return after == null ? result : [];
    }
}
class BufferContext {
    constructor(parent, buffer, index, start) {
        this.parent = parent;
        this.buffer = buffer;
        this.index = index;
        this.start = start;
    }
}
class BufferNode {
    constructor(context, _parent, index) {
        this.context = context;
        this._parent = _parent;
        this.index = index;
        this.type = context.buffer.set.types[context.buffer.buffer[index]];
    }
    get name() { return this.type.name; }
    get from() { return this.context.start + this.context.buffer.buffer[this.index + 1]; }
    get to() { return this.context.start + this.context.buffer.buffer[this.index + 2]; }
    child(dir, after) {
        let { buffer } = this.context;
        let index = buffer.findChild(this.index + 4, buffer.buffer[this.index + 3], dir, after == -100000000 /* None */ ? -100000000 /* None */ : after - this.context.start);
        return index < 0 ? null : new BufferNode(this.context, this, index);
    }
    get firstChild() { return this.child(1, -100000000 /* None */); }
    get lastChild() { return this.child(-1, -100000000 /* None */); }
    childAfter(pos) { return this.child(1, pos); }
    childBefore(pos) { return this.child(-1, pos); }
    get parent() {
        return this._parent || this.context.parent.nextSignificantParent();
    }
    externalSibling(dir) {
        return this._parent ? null : this.context.parent.nextChild(this.context.index + dir, dir, -1);
    }
    get nextSibling() {
        let { buffer } = this.context;
        let after = buffer.buffer[this.index + 3];
        if (after < (this._parent ? buffer.buffer[this._parent.index + 3] : buffer.buffer.length))
            return new BufferNode(this.context, this._parent, after);
        return this.externalSibling(1);
    }
    get prevSibling() {
        let { buffer } = this.context;
        let parentStart = this._parent ? this._parent.index + 4 : 0;
        if (this.index == parentStart)
            return this.externalSibling(-1);
        return new BufferNode(this.context, this._parent, buffer.findChild(parentStart, this.index, -1, -100000000 /* None */));
    }
    get cursor() { return new TreeCursor(this); }
    resolve(pos, side = 0) {
        return this.cursor.moveTo(pos, side).node;
    }
    /// @internal
    toString() { return this.context.buffer.childString(this.index); }
    getChild(type, before = null, after = null) {
        let r = getChildren(this, type, before, after);
        return r.length ? r[0] : null;
    }
    getChildren(type, before = null, after = null) {
        return getChildren(this, type, before, after);
    }
}
/// A tree cursor object focuses on a given node in a syntax tree, and
/// allows you to move to adjacent nodes.
class TreeCursor {
    /// @internal
    constructor(node, full = false) {
        this.full = full;
        this.buffer = null;
        this.stack = [];
        this.index = 0;
        this.bufferNode = null;
        if (node instanceof TreeNode) {
            this.yieldNode(node);
        }
        else {
            this._tree = node.context.parent;
            this.buffer = node.context;
            for (let n = node._parent; n; n = n._parent)
                this.stack.unshift(n.index);
            this.bufferNode = node;
            this.yieldBuf(node.index);
        }
    }
    /// Shorthand for `.type.name`.
    get name() { return this.type.name; }
    yieldNode(node) {
        if (!node)
            return false;
        this._tree = node;
        this.type = node.type;
        this.from = node.from;
        this.to = node.to;
        return true;
    }
    yieldBuf(index, type) {
        this.index = index;
        let { start, buffer } = this.buffer;
        this.type = type || buffer.set.types[buffer.buffer[index]];
        this.from = start + buffer.buffer[index + 1];
        this.to = start + buffer.buffer[index + 2];
        return true;
    }
    yield(node) {
        if (!node)
            return false;
        if (node instanceof TreeNode) {
            this.buffer = null;
            return this.yieldNode(node);
        }
        this.buffer = node.context;
        return this.yieldBuf(node.index, node.type);
    }
    /// @internal
    toString() {
        return this.buffer ? this.buffer.buffer.childString(this.index) : this._tree.toString();
    }
    /// @internal
    enter(dir, after) {
        if (!this.buffer)
            return this.yield(this._tree.nextChild(dir < 0 ? this._tree.node.children.length - 1 : 0, dir, after, this.full));
        let { buffer } = this.buffer;
        let index = buffer.findChild(this.index + 4, buffer.buffer[this.index + 3], dir, after == -100000000 /* None */ ? -100000000 /* None */ : after - this.buffer.start);
        if (index < 0)
            return false;
        this.stack.push(this.index);
        return this.yieldBuf(index);
    }
    /// Move the cursor to this node's first child. When this returns
    /// false, the node has no child, and the cursor has not been moved.
    firstChild() { return this.enter(1, -100000000 /* None */); }
    /// Move the cursor to this node's last child.
    lastChild() { return this.enter(-1, -100000000 /* None */); }
    /// Move the cursor to the first child that starts at or after `pos`.
    childAfter(pos) { return this.enter(1, pos); }
    /// Move to the last child that ends at or before `pos`.
    childBefore(pos) { return this.enter(-1, pos); }
    /// Move the node's parent node, if this isn't the top node.
    parent() {
        if (!this.buffer)
            return this.yieldNode(this.full ? this._tree._parent : this._tree.parent);
        if (this.stack.length)
            return this.yieldBuf(this.stack.pop());
        let parent = this.full ? this.buffer.parent : this.buffer.parent.nextSignificantParent();
        this.buffer = null;
        return this.yieldNode(parent);
    }
    /// @internal
    sibling(dir) {
        if (!this.buffer)
            return !this._tree._parent ? false
                : this.yield(this._tree._parent.nextChild(this._tree.index + dir, dir, -100000000 /* None */, this.full));
        let { buffer } = this.buffer, d = this.stack.length - 1;
        if (dir < 0) {
            let parentStart = d < 0 ? 0 : this.stack[d] + 4;
            if (this.index != parentStart)
                return this.yieldBuf(buffer.findChild(parentStart, this.index, -1, -100000000 /* None */));
        }
        else {
            let after = buffer.buffer[this.index + 3];
            if (after < (d < 0 ? buffer.buffer.length : buffer.buffer[this.stack[d] + 3]))
                return this.yieldBuf(after);
        }
        return d < 0 ? this.yield(this.buffer.parent.nextChild(this.buffer.index + dir, dir, -100000000 /* None */, this.full)) : false;
    }
    /// Move to this node's next sibling, if any.
    nextSibling() { return this.sibling(1); }
    /// Move to this node's previous sibling, if any.
    prevSibling() { return this.sibling(-1); }
    atLastNode(dir) {
        let index, parent, { buffer } = this;
        if (buffer) {
            if (dir > 0) {
                if (this.index < buffer.buffer.buffer.length)
                    return false;
            }
            else {
                for (let i = 0; i < this.index; i++)
                    if (buffer.buffer.buffer[i + 3] < this.index)
                        return false;
            }
            ({ index, parent } = buffer);
        }
        else {
            ({ index, _parent: parent } = this._tree);
        }
        for (; parent; { index, _parent: parent } = parent) {
            for (let i = index + dir, e = dir < 0 ? -1 : parent.node.children.length; i != e; i += dir) {
                let child = parent.node.children[i];
                if (this.full || !child.type.isAnonymous || child instanceof TreeBuffer || hasChild(child))
                    return false;
            }
        }
        return true;
    }
    move(dir) {
        if (this.enter(dir, -100000000 /* None */))
            return true;
        for (;;) {
            if (this.sibling(dir))
                return true;
            if (this.atLastNode(dir) || !this.parent())
                return false;
        }
    }
    /// Move to the next node in a
    /// [pre-order](https://en.wikipedia.org/wiki/Tree_traversal#Pre-order_(NLR))
    /// traversal, going from a node to its first child or, if the
    /// current node is empty, its next sibling or the next sibling of
    /// the first parent node that has one.
    next() { return this.move(1); }
    /// Move to the next node in a last-to-first pre-order traveral. A
    /// node is followed by ist last child or, if it has none, its
    /// previous sibling or the previous sibling of the first parent
    /// node that has one.
    prev() { return this.move(-1); }
    /// Move the cursor to the innermost node that covers `pos`. If
    /// `side` is -1, it will enter nodes that end at `pos`. If it is 1,
    /// it will enter nodes that start at `pos`.
    moveTo(pos, side = 0) {
        // Move up to a node that actually holds the position, if possible
        while (this.from == this.to ||
            (side < 1 ? this.from >= pos : this.from > pos) ||
            (side > -1 ? this.to <= pos : this.to < pos))
            if (!this.parent())
                break;
        // Then scan down into child nodes as far as possible
        for (;;) {
            if (side < 0 ? !this.childBefore(pos) : !this.childAfter(pos))
                break;
            if (this.from == this.to ||
                (side < 1 ? this.from >= pos : this.from > pos) ||
                (side > -1 ? this.to <= pos : this.to < pos)) {
                this.parent();
                break;
            }
        }
        return this;
    }
    /// Get a [syntax node](#tree.SyntaxNode) at the cursor's current
    /// position.
    get node() {
        if (!this.buffer)
            return this._tree;
        let cache = this.bufferNode, result = null, depth = 0;
        if (cache && cache.context == this.buffer) {
            scan: for (let index = this.index, d = this.stack.length; d >= 0;) {
                for (let c = cache; c; c = c._parent)
                    if (c.index == index) {
                        if (index == this.index)
                            return c;
                        result = c;
                        depth = d + 1;
                        break scan;
                    }
                index = this.stack[--d];
            }
        }
        for (let i = depth; i < this.stack.length; i++)
            result = new BufferNode(this.buffer, result, this.stack[i]);
        return this.bufferNode = new BufferNode(this.buffer, result, this.index);
    }
    /// Get the [tree](#tree.Tree) that represents the current node, if
    /// any. Will return null when the node is in a [tree
    /// buffer](#tree.TreeBuffer).
    get tree() {
        return this.buffer ? null : this._tree.node;
    }
}
function hasChild(tree) {
    return tree.children.some(ch => !ch.type.isAnonymous || ch instanceof TreeBuffer || hasChild(ch));
}
class FlatBufferCursor {
    constructor(buffer, index) {
        this.buffer = buffer;
        this.index = index;
    }
    get id() { return this.buffer[this.index - 4]; }
    get start() { return this.buffer[this.index - 3]; }
    get end() { return this.buffer[this.index - 2]; }
    get size() { return this.buffer[this.index - 1]; }
    get pos() { return this.index; }
    next() { this.index -= 4; }
    fork() { return new FlatBufferCursor(this.buffer, this.index); }
}
const BalanceBranchFactor = 8;
function buildTree(data) {
    var _a;
    let { buffer, nodeSet, topID = 0, maxBufferLength = DefaultBufferLength, reused = [], minRepeatType = nodeSet.types.length } = data;
    let cursor = Array.isArray(buffer) ? new FlatBufferCursor(buffer, buffer.length) : buffer;
    let types = nodeSet.types;
    let contextHash = 0;
    function takeNode(parentStart, minPos, children, positions, inRepeat) {
        let { id, start, end, size } = cursor;
        let startPos = start - parentStart;
        if (size < 0) {
            if (size == -1) { // Reused node
                children.push(reused[id]);
                positions.push(startPos);
            }
            else { // Context change
                contextHash = id;
            }
            cursor.next();
            return;
        }
        let type = types[id], node, buffer;
        if (end - start <= maxBufferLength && (buffer = findBufferSize(cursor.pos - minPos, inRepeat))) {
            // Small enough for a buffer, and no reused nodes inside
            let data = new Uint16Array(buffer.size - buffer.skip);
            let endPos = cursor.pos - buffer.size, index = data.length;
            while (cursor.pos > endPos)
                index = copyToBuffer(buffer.start, data, index, inRepeat);
            node = new TreeBuffer(data, end - buffer.start, nodeSet, inRepeat < 0 ? NodeType.none : types[inRepeat]);
            startPos = buffer.start - parentStart;
        }
        else { // Make it a node
            let endPos = cursor.pos - size;
            cursor.next();
            let localChildren = [], localPositions = [];
            let localInRepeat = id >= minRepeatType ? id : -1;
            while (cursor.pos > endPos) {
                if (cursor.id == localInRepeat)
                    cursor.next();
                else
                    takeNode(start, endPos, localChildren, localPositions, localInRepeat);
            }
            localChildren.reverse();
            localPositions.reverse();
            if (localInRepeat > -1 && localChildren.length > BalanceBranchFactor)
                node = balanceRange(type, type, localChildren, localPositions, 0, localChildren.length, 0, maxBufferLength, end - start, contextHash);
            else
                node = withHash(new Tree(type, localChildren, localPositions, end - start), contextHash);
        }
        children.push(node);
        positions.push(startPos);
    }
    function findBufferSize(maxSize, inRepeat) {
        // Scan through the buffer to find previous siblings that fit
        // together in a TreeBuffer, and don't contain any reused nodes
        // (which can't be stored in a buffer).
        // If `inRepeat` is > -1, ignore node boundaries of that type for
        // nesting, but make sure the end falls either at the start
        // (`maxSize`) or before such a node.
        let fork = cursor.fork();
        let size = 0, start = 0, skip = 0, minStart = fork.end - maxBufferLength;
        let result = { size: 0, start: 0, skip: 0 };
        scan: for (let minPos = fork.pos - maxSize; fork.pos > minPos;) {
            // Pretend nested repeat nodes of the same type don't exist
            if (fork.id == inRepeat) {
                // Except that we store the current state as a valid return
                // value.
                result.size = size;
                result.start = start;
                result.skip = skip;
                skip += 4;
                size += 4;
                fork.next();
                continue;
            }
            let nodeSize = fork.size, startPos = fork.pos - nodeSize;
            if (nodeSize < 0 || startPos < minPos || fork.start < minStart)
                break;
            let localSkipped = fork.id >= minRepeatType ? 4 : 0;
            let nodeStart = fork.start;
            fork.next();
            while (fork.pos > startPos) {
                if (fork.size < 0)
                    break scan;
                if (fork.id >= minRepeatType)
                    localSkipped += 4;
                fork.next();
            }
            start = nodeStart;
            size += nodeSize;
            skip += localSkipped;
        }
        if (inRepeat < 0 || size == maxSize) {
            result.size = size;
            result.start = start;
            result.skip = skip;
        }
        return result.size > 4 ? result : undefined;
    }
    function copyToBuffer(bufferStart, buffer, index, inRepeat) {
        let { id, start, end, size } = cursor;
        cursor.next();
        if (id == inRepeat)
            return index;
        let startIndex = index;
        if (size > 4) {
            let endPos = cursor.pos - (size - 4);
            while (cursor.pos > endPos)
                index = copyToBuffer(bufferStart, buffer, index, inRepeat);
        }
        if (id < minRepeatType) { // Don't copy repeat nodes into buffers
            buffer[--index] = startIndex;
            buffer[--index] = end - bufferStart;
            buffer[--index] = start - bufferStart;
            buffer[--index] = id;
        }
        return index;
    }
    let children = [], positions = [];
    while (cursor.pos > 0)
        takeNode(data.start || 0, 0, children, positions, -1);
    let length = (_a = data.length) !== null && _a !== void 0 ? _a : (children.length ? positions[0] + children[0].length : 0);
    return new Tree(types[topID], children.reverse(), positions.reverse(), length);
}
function balanceRange(outerType, innerType, children, positions, from, to, start, maxBufferLength, length, contextHash) {
    let localChildren = [], localPositions = [];
    if (length <= maxBufferLength) {
        for (let i = from; i < to; i++) {
            localChildren.push(children[i]);
            localPositions.push(positions[i] - start);
        }
    }
    else {
        let maxChild = Math.max(maxBufferLength, Math.ceil(length * 1.5 / BalanceBranchFactor));
        for (let i = from; i < to;) {
            let groupFrom = i, groupStart = positions[i];
            i++;
            for (; i < to; i++) {
                let nextEnd = positions[i] + children[i].length;
                if (nextEnd - groupStart > maxChild)
                    break;
            }
            if (i == groupFrom + 1) {
                let only = children[groupFrom];
                if (only instanceof Tree && only.type == innerType && only.length > maxChild << 1) { // Too big, collapse
                    for (let j = 0; j < only.children.length; j++) {
                        localChildren.push(only.children[j]);
                        localPositions.push(only.positions[j] + groupStart - start);
                    }
                    continue;
                }
                localChildren.push(only);
            }
            else if (i == groupFrom + 1) {
                localChildren.push(children[groupFrom]);
            }
            else {
                let inner = balanceRange(innerType, innerType, children, positions, groupFrom, i, groupStart, maxBufferLength, positions[i - 1] + children[i - 1].length - groupStart, contextHash);
                if (innerType != NodeType.none && !containsType(inner.children, innerType))
                    inner = withHash(new Tree(NodeType.none, inner.children, inner.positions, inner.length), contextHash);
                localChildren.push(inner);
            }
            localPositions.push(groupStart - start);
        }
    }
    return withHash(new Tree(outerType, localChildren, localPositions, length), contextHash);
}
function containsType(nodes, type) {
    for (let elt of nodes)
        if (elt.type == type)
            return true;
    return false;
}
// Creates an `Input` that is backed by a single, flat string.
function stringInput(input) { return new StringInput(input); }
class StringInput {
    constructor(string, length = string.length) {
        this.string = string;
        this.length = length;
    }
    get(pos) {
        return pos < 0 || pos >= this.length ? -1 : this.string.charCodeAt(pos);
    }
    lineAfter(pos) {
        if (pos < 0)
            return "";
        let end = this.string.indexOf("\n", pos);
        return this.string.slice(pos, end < 0 ? this.length : Math.min(end, this.length));
    }
    read(from, to) { return this.string.slice(from, Math.min(this.length, to)); }
    clip(at) { return new StringInput(this.string, at); }
}

/// A parse stack. These are used internally by the parser to track
/// parsing progress. They also provide some properties and methods
/// that external code such as a tokenizer can use to get information
/// about the parse state.
class Stack {
    /// @internal
    constructor(
    /// A the parse that this stack is part of @internal
    p, 
    /// Holds state, pos, value stack pos (15 bits array index, 15 bits
    /// buffer index) triplets for all but the top state
    /// @internal
    stack, 
    /// The current parse state @internal
    state, 
    // The position at which the next reduce should take place. This
    // can be less than `this.pos` when skipped expressions have been
    // added to the stack (which should be moved outside of the next
    // reduction)
    /// @internal
    reducePos, 
    /// The input position up to which this stack has parsed.
    pos, 
    /// The dynamic score of the stack, including dynamic precedence
    /// and error-recovery penalties
    /// @internal
    score, 
    // The output buffer. Holds (type, start, end, size) quads
    // representing nodes created by the parser, where `size` is
    // amount of buffer array entries covered by this node.
    /// @internal
    buffer, 
    // The base offset of the buffer. When stacks are split, the split
    // instance shared the buffer history with its parent up to
    // `bufferBase`, which is the absolute offset (including the
    // offset of previous splits) into the buffer at which this stack
    // starts writing.
    /// @internal
    bufferBase, 
    /// @internal
    curContext, 
    // A parent stack from which this was split off, if any. This is
    // set up so that it always points to a stack that has some
    // additional buffer content, never to a stack with an equal
    // `bufferBase`.
    /// @internal
    parent) {
        this.p = p;
        this.stack = stack;
        this.state = state;
        this.reducePos = reducePos;
        this.pos = pos;
        this.score = score;
        this.buffer = buffer;
        this.bufferBase = bufferBase;
        this.curContext = curContext;
        this.parent = parent;
    }
    /// @internal
    toString() {
        return `[${this.stack.filter((_, i) => i % 3 == 0).concat(this.state)}]@${this.pos}${this.score ? "!" + this.score : ""}`;
    }
    // Start an empty stack
    /// @internal
    static start(p, state, pos = 0) {
        let cx = p.parser.context;
        return new Stack(p, [], state, pos, pos, 0, [], 0, cx ? new StackContext(cx, cx.start) : null, null);
    }
    /// The stack's current [context](#lezer.ContextTracker) value, if
    /// any. Its type will depend on the context tracker's type
    /// parameter, or it will be `null` if there is no context
    /// tracker.
    get context() { return this.curContext ? this.curContext.context : null; }
    // Push a state onto the stack, tracking its start position as well
    // as the buffer base at that point.
    /// @internal
    pushState(state, start) {
        this.stack.push(this.state, start, this.bufferBase + this.buffer.length);
        this.state = state;
    }
    // Apply a reduce action
    /// @internal
    reduce(action) {
        let depth = action >> 19 /* ReduceDepthShift */, type = action & 65535 /* ValueMask */;
        let { parser } = this.p;
        let dPrec = parser.dynamicPrecedence(type);
        if (dPrec)
            this.score += dPrec;
        if (depth == 0) {
            // Zero-depth reductions are a special case—they add stuff to
            // the stack without popping anything off.
            if (type < parser.minRepeatTerm)
                this.storeNode(type, this.reducePos, this.reducePos, 4, true);
            this.pushState(parser.getGoto(this.state, type, true), this.reducePos);
            this.reduceContext(type);
            return;
        }
        // Find the base index into `this.stack`, content after which will
        // be dropped. Note that with `StayFlag` reductions we need to
        // consume two extra frames (the dummy parent node for the skipped
        // expression and the state that we'll be staying in, which should
        // be moved to `this.state`).
        let base = this.stack.length - ((depth - 1) * 3) - (action & 262144 /* StayFlag */ ? 6 : 0);
        let start = this.stack[base - 2];
        let bufferBase = this.stack[base - 1], count = this.bufferBase + this.buffer.length - bufferBase;
        // Store normal terms or `R -> R R` repeat reductions
        if (type < parser.minRepeatTerm || (action & 131072 /* RepeatFlag */)) {
            let pos = parser.stateFlag(this.state, 1 /* Skipped */) ? this.pos : this.reducePos;
            this.storeNode(type, start, pos, count + 4, true);
        }
        if (action & 262144 /* StayFlag */) {
            this.state = this.stack[base];
        }
        else {
            let baseStateID = this.stack[base - 3];
            this.state = parser.getGoto(baseStateID, type, true);
        }
        while (this.stack.length > base)
            this.stack.pop();
        this.reduceContext(type);
    }
    // Shift a value into the buffer
    /// @internal
    storeNode(term, start, end, size = 4, isReduce = false) {
        if (term == 0 /* Err */) { // Try to omit/merge adjacent error nodes
            let cur = this, top = this.buffer.length;
            if (top == 0 && cur.parent) {
                top = cur.bufferBase - cur.parent.bufferBase;
                cur = cur.parent;
            }
            if (top > 0 && cur.buffer[top - 4] == 0 /* Err */ && cur.buffer[top - 1] > -1) {
                if (start == end)
                    return;
                if (cur.buffer[top - 2] >= start) {
                    cur.buffer[top - 2] = end;
                    return;
                }
            }
        }
        if (!isReduce || this.pos == end) { // Simple case, just append
            this.buffer.push(term, start, end, size);
        }
        else { // There may be skipped nodes that have to be moved forward
            let index = this.buffer.length;
            if (index > 0 && this.buffer[index - 4] != 0 /* Err */)
                while (index > 0 && this.buffer[index - 2] > end) {
                    // Move this record forward
                    this.buffer[index] = this.buffer[index - 4];
                    this.buffer[index + 1] = this.buffer[index - 3];
                    this.buffer[index + 2] = this.buffer[index - 2];
                    this.buffer[index + 3] = this.buffer[index - 1];
                    index -= 4;
                    if (size > 4)
                        size -= 4;
                }
            this.buffer[index] = term;
            this.buffer[index + 1] = start;
            this.buffer[index + 2] = end;
            this.buffer[index + 3] = size;
        }
    }
    // Apply a shift action
    /// @internal
    shift(action, next, nextEnd) {
        if (action & 131072 /* GotoFlag */) {
            this.pushState(action & 65535 /* ValueMask */, this.pos);
        }
        else if ((action & 262144 /* StayFlag */) == 0) { // Regular shift
            let start = this.pos, nextState = action, { parser } = this.p;
            if (nextEnd > this.pos || next <= parser.maxNode) {
                this.pos = nextEnd;
                if (!parser.stateFlag(nextState, 1 /* Skipped */))
                    this.reducePos = nextEnd;
            }
            this.pushState(nextState, start);
            if (next <= parser.maxNode)
                this.buffer.push(next, start, nextEnd, 4);
            this.shiftContext(next);
        }
        else { // Shift-and-stay, which means this is a skipped token
            if (next <= this.p.parser.maxNode)
                this.buffer.push(next, this.pos, nextEnd, 4);
            this.pos = nextEnd;
        }
    }
    // Apply an action
    /// @internal
    apply(action, next, nextEnd) {
        if (action & 65536 /* ReduceFlag */)
            this.reduce(action);
        else
            this.shift(action, next, nextEnd);
    }
    // Add a prebuilt node into the buffer. This may be a reused node or
    // the result of running a nested parser.
    /// @internal
    useNode(value, next) {
        let index = this.p.reused.length - 1;
        if (index < 0 || this.p.reused[index] != value) {
            this.p.reused.push(value);
            index++;
        }
        let start = this.pos;
        this.reducePos = this.pos = start + value.length;
        this.pushState(next, start);
        this.buffer.push(index, start, this.reducePos, -1 /* size < 0 means this is a reused value */);
        if (this.curContext)
            this.updateContext(this.curContext.tracker.reuse(this.curContext.context, value, this.p.input, this));
    }
    // Split the stack. Due to the buffer sharing and the fact
    // that `this.stack` tends to stay quite shallow, this isn't very
    // expensive.
    /// @internal
    split() {
        let parent = this;
        let off = parent.buffer.length;
        // Because the top of the buffer (after this.pos) may be mutated
        // to reorder reductions and skipped tokens, and shared buffers
        // should be immutable, this copies any outstanding skipped tokens
        // to the new buffer, and puts the base pointer before them.
        while (off > 0 && parent.buffer[off - 2] > parent.reducePos)
            off -= 4;
        let buffer = parent.buffer.slice(off), base = parent.bufferBase + off;
        // Make sure parent points to an actual parent with content, if there is such a parent.
        while (parent && base == parent.bufferBase)
            parent = parent.parent;
        return new Stack(this.p, this.stack.slice(), this.state, this.reducePos, this.pos, this.score, buffer, base, this.curContext, parent);
    }
    // Try to recover from an error by 'deleting' (ignoring) one token.
    /// @internal
    recoverByDelete(next, nextEnd) {
        let isNode = next <= this.p.parser.maxNode;
        if (isNode)
            this.storeNode(next, this.pos, nextEnd);
        this.storeNode(0 /* Err */, this.pos, nextEnd, isNode ? 8 : 4);
        this.pos = this.reducePos = nextEnd;
        this.score -= 200 /* Token */;
    }
    /// Check if the given term would be able to be shifted (optionally
    /// after some reductions) on this stack. This can be useful for
    /// external tokenizers that want to make sure they only provide a
    /// given token when it applies.
    canShift(term) {
        for (let sim = new SimulatedStack(this);;) {
            let action = this.p.parser.stateSlot(sim.top, 4 /* DefaultReduce */) || this.p.parser.hasAction(sim.top, term);
            if ((action & 65536 /* ReduceFlag */) == 0)
                return true;
            if (action == 0)
                return false;
            sim.reduce(action);
        }
    }
    /// Find the start position of the rule that is currently being parsed.
    get ruleStart() {
        for (let state = this.state, base = this.stack.length;;) {
            let force = this.p.parser.stateSlot(state, 5 /* ForcedReduce */);
            if (!(force & 65536 /* ReduceFlag */))
                return 0;
            base -= 3 * (force >> 19 /* ReduceDepthShift */);
            if ((force & 65535 /* ValueMask */) < this.p.parser.minRepeatTerm)
                return this.stack[base + 1];
            state = this.stack[base];
        }
    }
    /// Find the start position of an instance of any of the given term
    /// types, or return `null` when none of them are found.
    ///
    /// **Note:** this is only reliable when there is at least some
    /// state that unambiguously matches the given rule on the stack.
    /// I.e. if you have a grammar like this, where the difference
    /// between `a` and `b` is only apparent at the third token:
    ///
    ///     a { b | c }
    ///     b { "x" "y" "x" }
    ///     c { "x" "y" "z" }
    ///
    /// Then a parse state after `"x"` will not reliably tell you that
    /// `b` is on the stack. You _can_ pass `[b, c]` to reliably check
    /// for either of those two rules (assuming that `a` isn't part of
    /// some rule that includes other things starting with `"x"`).
    ///
    /// When `before` is given, this keeps scanning up the stack until
    /// it finds a match that starts before that position.
    ///
    /// Note that you have to be careful when using this in tokenizers,
    /// since it's relatively easy to introduce data dependencies that
    /// break incremental parsing by using this method.
    startOf(types, before) {
        let state = this.state, frame = this.stack.length, { parser } = this.p;
        for (;;) {
            let force = parser.stateSlot(state, 5 /* ForcedReduce */);
            let depth = force >> 19 /* ReduceDepthShift */, term = force & 65535 /* ValueMask */;
            if (types.indexOf(term) > -1) {
                let base = frame - (3 * (force >> 19 /* ReduceDepthShift */)), pos = this.stack[base + 1];
                if (before == null || before > pos)
                    return pos;
            }
            if (frame == 0)
                return null;
            if (depth == 0) {
                frame -= 3;
                state = this.stack[frame];
            }
            else {
                frame -= 3 * (depth - 1);
                state = parser.getGoto(this.stack[frame - 3], term, true);
            }
        }
    }
    // Apply up to Recover.MaxNext recovery actions that conceptually
    // inserts some missing token or rule.
    /// @internal
    recoverByInsert(next) {
        if (this.stack.length >= 300 /* MaxInsertStackDepth */)
            return [];
        let nextStates = this.p.parser.nextStates(this.state);
        if (nextStates.length > 4 /* MaxNext */ << 1 || this.stack.length >= 120 /* DampenInsertStackDepth */) {
            let best = [];
            for (let i = 0, s; i < nextStates.length; i += 2) {
                if ((s = nextStates[i + 1]) != this.state && this.p.parser.hasAction(s, next))
                    best.push(nextStates[i], s);
            }
            if (this.stack.length < 120 /* DampenInsertStackDepth */)
                for (let i = 0; best.length < 4 /* MaxNext */ << 1 && i < nextStates.length; i += 2) {
                    let s = nextStates[i + 1];
                    if (!best.some((v, i) => (i & 1) && v == s))
                        best.push(nextStates[i], s);
                }
            nextStates = best;
        }
        let result = [];
        for (let i = 0; i < nextStates.length && result.length < 4 /* MaxNext */; i += 2) {
            let s = nextStates[i + 1];
            if (s == this.state)
                continue;
            let stack = this.split();
            stack.storeNode(0 /* Err */, stack.pos, stack.pos, 4, true);
            stack.pushState(s, this.pos);
            stack.shiftContext(nextStates[i]);
            stack.score -= 200 /* Token */;
            result.push(stack);
        }
        return result;
    }
    // Force a reduce, if possible. Return false if that can't
    // be done.
    /// @internal
    forceReduce() {
        let reduce = this.p.parser.stateSlot(this.state, 5 /* ForcedReduce */);
        if ((reduce & 65536 /* ReduceFlag */) == 0)
            return false;
        if (!this.p.parser.validAction(this.state, reduce)) {
            this.storeNode(0 /* Err */, this.reducePos, this.reducePos, 4, true);
            this.score -= 100 /* Reduce */;
        }
        this.reduce(reduce);
        return true;
    }
    /// @internal
    forceAll() {
        while (!this.p.parser.stateFlag(this.state, 2 /* Accepting */) && this.forceReduce()) { }
        return this;
    }
    /// Check whether this state has no further actions (assumed to be a direct descendant of the
    /// top state, since any other states must be able to continue
    /// somehow). @internal
    get deadEnd() {
        if (this.stack.length != 3)
            return false;
        let { parser } = this.p;
        return parser.data[parser.stateSlot(this.state, 1 /* Actions */)] == 65535 /* End */ &&
            !parser.stateSlot(this.state, 4 /* DefaultReduce */);
    }
    /// Restart the stack (put it back in its start state). Only safe
    /// when this.stack.length == 3 (state is directly below the top
    /// state). @internal
    restart() {
        this.state = this.stack[0];
        this.stack.length = 0;
    }
    /// @internal
    sameState(other) {
        if (this.state != other.state || this.stack.length != other.stack.length)
            return false;
        for (let i = 0; i < this.stack.length; i += 3)
            if (this.stack[i] != other.stack[i])
                return false;
        return true;
    }
    /// Get the parser used by this stack.
    get parser() { return this.p.parser; }
    /// Test whether a given dialect (by numeric ID, as exported from
    /// the terms file) is enabled.
    dialectEnabled(dialectID) { return this.p.parser.dialect.flags[dialectID]; }
    shiftContext(term) {
        if (this.curContext)
            this.updateContext(this.curContext.tracker.shift(this.curContext.context, term, this.p.input, this));
    }
    reduceContext(term) {
        if (this.curContext)
            this.updateContext(this.curContext.tracker.reduce(this.curContext.context, term, this.p.input, this));
    }
    /// @internal
    emitContext() {
        let cx = this.curContext;
        if (!cx.tracker.strict)
            return;
        let last = this.buffer.length - 1;
        if (last < 0 || this.buffer[last] != -2)
            this.buffer.push(cx.hash, this.reducePos, this.reducePos, -2);
    }
    updateContext(context) {
        if (context != this.curContext.context) {
            let newCx = new StackContext(this.curContext.tracker, context);
            if (newCx.hash != this.curContext.hash)
                this.emitContext();
            this.curContext = newCx;
        }
    }
}
class StackContext {
    constructor(tracker, context) {
        this.tracker = tracker;
        this.context = context;
        this.hash = tracker.hash(context);
    }
}
var Recover;
(function (Recover) {
    Recover[Recover["Token"] = 200] = "Token";
    Recover[Recover["Reduce"] = 100] = "Reduce";
    Recover[Recover["MaxNext"] = 4] = "MaxNext";
    Recover[Recover["MaxInsertStackDepth"] = 300] = "MaxInsertStackDepth";
    Recover[Recover["DampenInsertStackDepth"] = 120] = "DampenInsertStackDepth";
})(Recover || (Recover = {}));
// Used to cheaply run some reductions to scan ahead without mutating
// an entire stack
class SimulatedStack {
    constructor(stack) {
        this.stack = stack;
        this.top = stack.state;
        this.rest = stack.stack;
        this.offset = this.rest.length;
    }
    reduce(action) {
        let term = action & 65535 /* ValueMask */, depth = action >> 19 /* ReduceDepthShift */;
        if (depth == 0) {
            if (this.rest == this.stack.stack)
                this.rest = this.rest.slice();
            this.rest.push(this.top, 0, 0);
            this.offset += 3;
        }
        else {
            this.offset -= (depth - 1) * 3;
        }
        let goto = this.stack.p.parser.getGoto(this.rest[this.offset - 3], term, true);
        this.top = goto;
    }
}
// This is given to `Tree.build` to build a buffer, and encapsulates
// the parent-stack-walking necessary to read the nodes.
class StackBufferCursor {
    constructor(stack, pos, index) {
        this.stack = stack;
        this.pos = pos;
        this.index = index;
        this.buffer = stack.buffer;
        if (this.index == 0)
            this.maybeNext();
    }
    static create(stack) {
        return new StackBufferCursor(stack, stack.bufferBase + stack.buffer.length, stack.buffer.length);
    }
    maybeNext() {
        let next = this.stack.parent;
        if (next != null) {
            this.index = this.stack.bufferBase - next.bufferBase;
            this.stack = next;
            this.buffer = next.buffer;
        }
    }
    get id() { return this.buffer[this.index - 4]; }
    get start() { return this.buffer[this.index - 3]; }
    get end() { return this.buffer[this.index - 2]; }
    get size() { return this.buffer[this.index - 1]; }
    next() {
        this.index -= 4;
        this.pos -= 4;
        if (this.index == 0)
            this.maybeNext();
    }
    fork() {
        return new StackBufferCursor(this.stack, this.pos, this.index);
    }
}

/// Tokenizers write the tokens they read into instances of this class.
class Token {
    constructor() {
        /// The start of the token. This is set by the parser, and should not
        /// be mutated by the tokenizer.
        this.start = -1;
        /// This starts at -1, and should be updated to a term id when a
        /// matching token is found.
        this.value = -1;
        /// When setting `.value`, you should also set `.end` to the end
        /// position of the token. (You'll usually want to use the `accept`
        /// method.)
        this.end = -1;
    }
    /// Accept a token, setting `value` and `end` to the given values.
    accept(value, end) {
        this.value = value;
        this.end = end;
    }
}
/// @internal
class TokenGroup {
    constructor(data, id) {
        this.data = data;
        this.id = id;
    }
    token(input, token, stack) { readToken(this.data, input, token, stack, this.id); }
}
TokenGroup.prototype.contextual = TokenGroup.prototype.fallback = TokenGroup.prototype.extend = false;
// Tokenizer data is stored a big uint16 array containing, for each
// state:
//
//  - A group bitmask, indicating what token groups are reachable from
//    this state, so that paths that can only lead to tokens not in
//    any of the current groups can be cut off early.
//
//  - The position of the end of the state's sequence of accepting
//    tokens
//
//  - The number of outgoing edges for the state
//
//  - The accepting tokens, as (token id, group mask) pairs
//
//  - The outgoing edges, as (start character, end character, state
//    index) triples, with end character being exclusive
//
// This function interprets that data, running through a stream as
// long as new states with the a matching group mask can be reached,
// and updating `token` when it matches a token.
function readToken(data, input, token, stack, group) {
    let state = 0, groupMask = 1 << group, dialect = stack.p.parser.dialect;
    scan: for (let pos = token.start;;) {
        if ((groupMask & data[state]) == 0)
            break;
        let accEnd = data[state + 1];
        // Check whether this state can lead to a token in the current group
        // Accept tokens in this state, possibly overwriting
        // lower-precedence / shorter tokens
        for (let i = state + 3; i < accEnd; i += 2)
            if ((data[i + 1] & groupMask) > 0) {
                let term = data[i];
                if (dialect.allows(term) &&
                    (token.value == -1 || token.value == term || stack.p.parser.overrides(term, token.value))) {
                    token.accept(term, pos);
                    break;
                }
            }
        let next = input.get(pos++);
        // Do a binary search on the state's edges
        for (let low = 0, high = data[state + 2]; low < high;) {
            let mid = (low + high) >> 1;
            let index = accEnd + mid + (mid << 1);
            let from = data[index], to = data[index + 1];
            if (next < from)
                high = mid;
            else if (next >= to)
                low = mid + 1;
            else {
                state = data[index + 2];
                continue scan;
            }
        }
        break;
    }
}

// See lezer-generator/src/encode.ts for comments about the encoding
// used here
function decodeArray(input, Type = Uint16Array) {
    if (typeof input != "string")
        return input;
    let array = null;
    for (let pos = 0, out = 0; pos < input.length;) {
        let value = 0;
        for (;;) {
            let next = input.charCodeAt(pos++), stop = false;
            if (next == 126 /* BigValCode */) {
                value = 65535 /* BigVal */;
                break;
            }
            if (next >= 92 /* Gap2 */)
                next--;
            if (next >= 34 /* Gap1 */)
                next--;
            let digit = next - 32 /* Start */;
            if (digit >= 46 /* Base */) {
                digit -= 46 /* Base */;
                stop = true;
            }
            value += digit;
            if (stop)
                break;
            value *= 46 /* Base */;
        }
        if (array)
            array[out++] = value;
        else
            array = new Type(value);
    }
    return array;
}

// FIXME find some way to reduce recovery work done when the input
// doesn't match the grammar at all.
// Environment variable used to control console output
const verbose = typeof process != "undefined" && /\bparse\b/.test(process.env.LOG);
let stackIDs = null;
function cutAt(tree, pos, side) {
    let cursor = tree.cursor(pos);
    for (;;) {
        if (!(side < 0 ? cursor.childBefore(pos) : cursor.childAfter(pos)))
            for (;;) {
                if ((side < 0 ? cursor.to < pos : cursor.from > pos) && !cursor.type.isError)
                    return side < 0 ? Math.max(0, Math.min(cursor.to - 1, pos - 5)) : Math.min(tree.length, Math.max(cursor.from + 1, pos + 5));
                if (side < 0 ? cursor.prevSibling() : cursor.nextSibling())
                    break;
                if (!cursor.parent())
                    return side < 0 ? 0 : tree.length;
            }
    }
}
class FragmentCursor {
    constructor(fragments) {
        this.fragments = fragments;
        this.i = 0;
        this.fragment = null;
        this.safeFrom = -1;
        this.safeTo = -1;
        this.trees = [];
        this.start = [];
        this.index = [];
        this.nextFragment();
    }
    nextFragment() {
        let fr = this.fragment = this.i == this.fragments.length ? null : this.fragments[this.i++];
        if (fr) {
            this.safeFrom = fr.openStart ? cutAt(fr.tree, fr.from + fr.offset, 1) - fr.offset : fr.from;
            this.safeTo = fr.openEnd ? cutAt(fr.tree, fr.to + fr.offset, -1) - fr.offset : fr.to;
            while (this.trees.length) {
                this.trees.pop();
                this.start.pop();
                this.index.pop();
            }
            this.trees.push(fr.tree);
            this.start.push(-fr.offset);
            this.index.push(0);
            this.nextStart = this.safeFrom;
        }
        else {
            this.nextStart = 1e9;
        }
    }
    // `pos` must be >= any previously given `pos` for this cursor
    nodeAt(pos) {
        if (pos < this.nextStart)
            return null;
        while (this.fragment && this.safeTo <= pos)
            this.nextFragment();
        if (!this.fragment)
            return null;
        for (;;) {
            let last = this.trees.length - 1;
            if (last < 0) { // End of tree
                this.nextFragment();
                return null;
            }
            let top = this.trees[last], index = this.index[last];
            if (index == top.children.length) {
                this.trees.pop();
                this.start.pop();
                this.index.pop();
                continue;
            }
            let next = top.children[index];
            let start = this.start[last] + top.positions[index];
            if (start > pos) {
                this.nextStart = start;
                return null;
            }
            else if (start == pos && start + next.length <= this.safeTo) {
                return start == pos && start >= this.safeFrom ? next : null;
            }
            if (next instanceof TreeBuffer) {
                this.index[last]++;
                this.nextStart = start + next.length;
            }
            else {
                this.index[last]++;
                if (start + next.length >= pos) { // Enter this node
                    this.trees.push(next);
                    this.start.push(start);
                    this.index.push(0);
                }
            }
        }
    }
}
class CachedToken extends Token {
    constructor() {
        super(...arguments);
        this.extended = -1;
        this.mask = 0;
        this.context = 0;
    }
    clear(start) {
        this.start = start;
        this.value = this.extended = -1;
    }
}
const dummyToken = new Token;
class TokenCache$1 {
    constructor(parser) {
        this.tokens = [];
        this.mainToken = dummyToken;
        this.actions = [];
        this.tokens = parser.tokenizers.map(_ => new CachedToken);
    }
    getActions(stack, input) {
        let actionIndex = 0;
        let main = null;
        let { parser } = stack.p, { tokenizers } = parser;
        let mask = parser.stateSlot(stack.state, 3 /* TokenizerMask */);
        let context = stack.curContext ? stack.curContext.hash : 0;
        for (let i = 0; i < tokenizers.length; i++) {
            if (((1 << i) & mask) == 0)
                continue;
            let tokenizer = tokenizers[i], token = this.tokens[i];
            if (main && !tokenizer.fallback)
                continue;
            if (tokenizer.contextual || token.start != stack.pos || token.mask != mask || token.context != context) {
                this.updateCachedToken(token, tokenizer, stack, input);
                token.mask = mask;
                token.context = context;
            }
            if (token.value != 0 /* Err */) {
                let startIndex = actionIndex;
                if (token.extended > -1)
                    actionIndex = this.addActions(stack, token.extended, token.end, actionIndex);
                actionIndex = this.addActions(stack, token.value, token.end, actionIndex);
                if (!tokenizer.extend) {
                    main = token;
                    if (actionIndex > startIndex)
                        break;
                }
            }
        }
        while (this.actions.length > actionIndex)
            this.actions.pop();
        if (!main) {
            main = dummyToken;
            main.start = stack.pos;
            if (stack.pos == input.length)
                main.accept(stack.p.parser.eofTerm, stack.pos);
            else
                main.accept(0 /* Err */, stack.pos + 1);
        }
        this.mainToken = main;
        return this.actions;
    }
    updateCachedToken(token, tokenizer, stack, input) {
        token.clear(stack.pos);
        tokenizer.token(input, token, stack);
        if (token.value > -1) {
            let { parser } = stack.p;
            for (let i = 0; i < parser.specialized.length; i++)
                if (parser.specialized[i] == token.value) {
                    let result = parser.specializers[i](input.read(token.start, token.end), stack);
                    if (result >= 0 && stack.p.parser.dialect.allows(result >> 1)) {
                        if ((result & 1) == 0 /* Specialize */)
                            token.value = result >> 1;
                        else
                            token.extended = result >> 1;
                        break;
                    }
                }
        }
        else if (stack.pos == input.length) {
            token.accept(stack.p.parser.eofTerm, stack.pos);
        }
        else {
            token.accept(0 /* Err */, stack.pos + 1);
        }
    }
    putAction(action, token, end, index) {
        // Don't add duplicate actions
        for (let i = 0; i < index; i += 3)
            if (this.actions[i] == action)
                return index;
        this.actions[index++] = action;
        this.actions[index++] = token;
        this.actions[index++] = end;
        return index;
    }
    addActions(stack, token, end, index) {
        let { state } = stack, { parser } = stack.p, { data } = parser;
        for (let set = 0; set < 2; set++) {
            for (let i = parser.stateSlot(state, set ? 2 /* Skip */ : 1 /* Actions */);; i += 3) {
                if (data[i] == 65535 /* End */) {
                    if (data[i + 1] == 1 /* Next */) {
                        i = pair(data, i + 2);
                    }
                    else {
                        if (index == 0 && data[i + 1] == 2 /* Other */)
                            index = this.putAction(pair(data, i + 1), token, end, index);
                        break;
                    }
                }
                if (data[i] == token)
                    index = this.putAction(pair(data, i + 1), token, end, index);
            }
        }
        return index;
    }
}
var Rec;
(function (Rec) {
    Rec[Rec["Distance"] = 5] = "Distance";
    Rec[Rec["MaxRemainingPerStep"] = 3] = "MaxRemainingPerStep";
    Rec[Rec["MinBufferLengthPrune"] = 200] = "MinBufferLengthPrune";
    Rec[Rec["ForceReduceLimit"] = 10] = "ForceReduceLimit";
})(Rec || (Rec = {}));
/// A parse context can be used for step-by-step parsing. After
/// creating it, you repeatedly call `.advance()` until it returns a
/// tree to indicate it has reached the end of the parse.
class Parse {
    constructor(parser, input, startPos, context) {
        this.parser = parser;
        this.input = input;
        this.startPos = startPos;
        this.context = context;
        // The position to which the parse has advanced.
        this.pos = 0;
        this.recovering = 0;
        this.nextStackID = 0x2654;
        this.nested = null;
        this.nestEnd = 0;
        this.nestWrap = null;
        this.reused = [];
        this.tokens = new TokenCache$1(parser);
        this.topTerm = parser.top[1];
        this.stacks = [Stack.start(this, parser.top[0], this.startPos)];
        let fragments = context === null || context === void 0 ? void 0 : context.fragments;
        this.fragments = fragments && fragments.length ? new FragmentCursor(fragments) : null;
    }
    // Move the parser forward. This will process all parse stacks at
    // `this.pos` and try to advance them to a further position. If no
    // stack for such a position is found, it'll start error-recovery.
    //
    // When the parse is finished, this will return a syntax tree. When
    // not, it returns `null`.
    advance() {
        if (this.nested) {
            let result = this.nested.advance();
            this.pos = this.nested.pos;
            if (result) {
                this.finishNested(this.stacks[0], result);
                this.nested = null;
            }
            return null;
        }
        let stacks = this.stacks, pos = this.pos;
        // This will hold stacks beyond `pos`.
        let newStacks = this.stacks = [];
        let stopped, stoppedTokens;
        let maybeNest;
        // Keep advancing any stacks at `pos` until they either move
        // forward or can't be advanced. Gather stacks that can't be
        // advanced further in `stopped`.
        for (let i = 0; i < stacks.length; i++) {
            let stack = stacks[i], nest;
            for (;;) {
                if (stack.pos > pos) {
                    newStacks.push(stack);
                }
                else if (nest = this.checkNest(stack)) {
                    if (!maybeNest || maybeNest.stack.score < stack.score)
                        maybeNest = nest;
                }
                else if (this.advanceStack(stack, newStacks, stacks)) {
                    continue;
                }
                else {
                    if (!stopped) {
                        stopped = [];
                        stoppedTokens = [];
                    }
                    stopped.push(stack);
                    let tok = this.tokens.mainToken;
                    stoppedTokens.push(tok.value, tok.end);
                }
                break;
            }
        }
        if (maybeNest) {
            this.startNested(maybeNest);
            return null;
        }
        if (!newStacks.length) {
            let finished = stopped && findFinished(stopped);
            if (finished)
                return this.stackToTree(finished);
            if (this.parser.strict) {
                if (verbose && stopped)
                    console.log("Stuck with token " + this.parser.getName(this.tokens.mainToken.value));
                throw new SyntaxError("No parse at " + pos);
            }
            if (!this.recovering)
                this.recovering = 5 /* Distance */;
        }
        if (this.recovering && stopped) {
            let finished = this.runRecovery(stopped, stoppedTokens, newStacks);
            if (finished)
                return this.stackToTree(finished.forceAll());
        }
        if (this.recovering) {
            let maxRemaining = this.recovering == 1 ? 1 : this.recovering * 3 /* MaxRemainingPerStep */;
            if (newStacks.length > maxRemaining) {
                newStacks.sort((a, b) => b.score - a.score);
                while (newStacks.length > maxRemaining)
                    newStacks.pop();
            }
            if (newStacks.some(s => s.reducePos > pos))
                this.recovering--;
        }
        else if (newStacks.length > 1) {
            // Prune stacks that are in the same state, or that have been
            // running without splitting for a while, to avoid getting stuck
            // with multiple successful stacks running endlessly on.
            outer: for (let i = 0; i < newStacks.length - 1; i++) {
                let stack = newStacks[i];
                for (let j = i + 1; j < newStacks.length; j++) {
                    let other = newStacks[j];
                    if (stack.sameState(other) ||
                        stack.buffer.length > 200 /* MinBufferLengthPrune */ && other.buffer.length > 200 /* MinBufferLengthPrune */) {
                        if (((stack.score - other.score) || (stack.buffer.length - other.buffer.length)) > 0) {
                            newStacks.splice(j--, 1);
                        }
                        else {
                            newStacks.splice(i--, 1);
                            continue outer;
                        }
                    }
                }
            }
        }
        this.pos = newStacks[0].pos;
        for (let i = 1; i < newStacks.length; i++)
            if (newStacks[i].pos < this.pos)
                this.pos = newStacks[i].pos;
        return null;
    }
    // Returns an updated version of the given stack, or null if the
    // stack can't advance normally. When `split` and `stacks` are
    // given, stacks split off by ambiguous operations will be pushed to
    // `split`, or added to `stacks` if they move `pos` forward.
    advanceStack(stack, stacks, split) {
        let start = stack.pos, { input, parser } = this;
        let base = verbose ? this.stackID(stack) + " -> " : "";
        if (this.fragments) {
            let strictCx = stack.curContext && stack.curContext.tracker.strict, cxHash = strictCx ? stack.curContext.hash : 0;
            for (let cached = this.fragments.nodeAt(start); cached;) {
                let match = this.parser.nodeSet.types[cached.type.id] == cached.type ? parser.getGoto(stack.state, cached.type.id) : -1;
                if (match > -1 && cached.length && (!strictCx || (cached.contextHash || 0) == cxHash)) {
                    stack.useNode(cached, match);
                    if (verbose)
                        console.log(base + this.stackID(stack) + ` (via reuse of ${parser.getName(cached.type.id)})`);
                    return true;
                }
                if (!(cached instanceof Tree) || cached.children.length == 0 || cached.positions[0] > 0)
                    break;
                let inner = cached.children[0];
                if (inner instanceof Tree)
                    cached = inner;
                else
                    break;
            }
        }
        let defaultReduce = parser.stateSlot(stack.state, 4 /* DefaultReduce */);
        if (defaultReduce > 0) {
            stack.reduce(defaultReduce);
            if (verbose)
                console.log(base + this.stackID(stack) + ` (via always-reduce ${parser.getName(defaultReduce & 65535 /* ValueMask */)})`);
            return true;
        }
        let actions = this.tokens.getActions(stack, input);
        for (let i = 0; i < actions.length;) {
            let action = actions[i++], term = actions[i++], end = actions[i++];
            let last = i == actions.length || !split;
            let localStack = last ? stack : stack.split();
            localStack.apply(action, term, end);
            if (verbose)
                console.log(base + this.stackID(localStack) + ` (via ${(action & 65536 /* ReduceFlag */) == 0 ? "shift"
                    : `reduce of ${parser.getName(action & 65535 /* ValueMask */)}`} for ${parser.getName(term)} @ ${start}${localStack == stack ? "" : ", split"})`);
            if (last)
                return true;
            else if (localStack.pos > start)
                stacks.push(localStack);
            else
                split.push(localStack);
        }
        return false;
    }
    // Advance a given stack forward as far as it will go. Returns the
    // (possibly updated) stack if it got stuck, or null if it moved
    // forward and was given to `pushStackDedup`.
    advanceFully(stack, newStacks) {
        let pos = stack.pos;
        for (;;) {
            let nest = this.checkNest(stack);
            if (nest)
                return nest;
            if (!this.advanceStack(stack, null, null))
                return false;
            if (stack.pos > pos) {
                pushStackDedup(stack, newStacks);
                return true;
            }
        }
    }
    runRecovery(stacks, tokens, newStacks) {
        let finished = null, restarted = false;
        let maybeNest;
        for (let i = 0; i < stacks.length; i++) {
            let stack = stacks[i], token = tokens[i << 1], tokenEnd = tokens[(i << 1) + 1];
            let base = verbose ? this.stackID(stack) + " -> " : "";
            if (stack.deadEnd) {
                if (restarted)
                    continue;
                restarted = true;
                stack.restart();
                if (verbose)
                    console.log(base + this.stackID(stack) + " (restarted)");
                let done = this.advanceFully(stack, newStacks);
                if (done) {
                    if (done !== true)
                        maybeNest = done;
                    continue;
                }
            }
            let force = stack.split(), forceBase = base;
            for (let j = 0; force.forceReduce() && j < 10 /* ForceReduceLimit */; j++) {
                if (verbose)
                    console.log(forceBase + this.stackID(force) + " (via force-reduce)");
                let done = this.advanceFully(force, newStacks);
                if (done) {
                    if (done !== true)
                        maybeNest = done;
                    break;
                }
                if (verbose)
                    forceBase = this.stackID(force) + " -> ";
            }
            for (let insert of stack.recoverByInsert(token)) {
                if (verbose)
                    console.log(base + this.stackID(insert) + " (via recover-insert)");
                this.advanceFully(insert, newStacks);
            }
            if (this.input.length > stack.pos) {
                if (tokenEnd == stack.pos) {
                    tokenEnd++;
                    token = 0 /* Err */;
                }
                stack.recoverByDelete(token, tokenEnd);
                if (verbose)
                    console.log(base + this.stackID(stack) + ` (via recover-delete ${this.parser.getName(token)})`);
                pushStackDedup(stack, newStacks);
            }
            else if (!finished || finished.score < stack.score) {
                finished = stack;
            }
        }
        if (finished)
            return finished;
        if (maybeNest)
            for (let s of this.stacks)
                if (s.score > maybeNest.stack.score) {
                    maybeNest = undefined;
                    break;
                }
        if (maybeNest)
            this.startNested(maybeNest);
        return null;
    }
    forceFinish() {
        let stack = this.stacks[0].split();
        if (this.nested)
            this.finishNested(stack, this.nested.forceFinish());
        return this.stackToTree(stack.forceAll());
    }
    // Convert the stack's buffer to a syntax tree.
    stackToTree(stack, pos = stack.pos) {
        if (this.parser.context)
            stack.emitContext();
        return Tree.build({ buffer: StackBufferCursor.create(stack),
            nodeSet: this.parser.nodeSet,
            topID: this.topTerm,
            maxBufferLength: this.parser.bufferLength,
            reused: this.reused,
            start: this.startPos,
            length: pos - this.startPos,
            minRepeatType: this.parser.minRepeatTerm });
    }
    checkNest(stack) {
        let info = this.parser.findNested(stack.state);
        if (!info)
            return null;
        let spec = info.value;
        if (typeof spec == "function")
            spec = spec(this.input, stack);
        return spec ? { stack, info, spec } : null;
    }
    startNested(nest) {
        let { stack, info, spec } = nest;
        this.stacks = [stack];
        this.nestEnd = this.scanForNestEnd(stack, info.end, spec.filterEnd);
        this.nestWrap = typeof spec.wrapType == "number" ? this.parser.nodeSet.types[spec.wrapType] : spec.wrapType || null;
        if (spec.startParse) {
            this.nested = spec.startParse(this.input.clip(this.nestEnd), stack.pos, this.context);
        }
        else {
            this.finishNested(stack);
        }
    }
    scanForNestEnd(stack, endToken, filter) {
        for (let pos = stack.pos; pos < this.input.length; pos++) {
            dummyToken.start = pos;
            dummyToken.value = -1;
            endToken.token(this.input, dummyToken, stack);
            if (dummyToken.value > -1 && (!filter || filter(this.input.read(pos, dummyToken.end))))
                return pos;
        }
        return this.input.length;
    }
    finishNested(stack, tree) {
        if (this.nestWrap)
            tree = new Tree(this.nestWrap, tree ? [tree] : [], tree ? [0] : [], this.nestEnd - stack.pos);
        else if (!tree)
            tree = new Tree(NodeType.none, [], [], this.nestEnd - stack.pos);
        let info = this.parser.findNested(stack.state);
        stack.useNode(tree, this.parser.getGoto(stack.state, info.placeholder, true));
        if (verbose)
            console.log(this.stackID(stack) + ` (via unnest)`);
    }
    stackID(stack) {
        let id = (stackIDs || (stackIDs = new WeakMap)).get(stack);
        if (!id)
            stackIDs.set(stack, id = String.fromCodePoint(this.nextStackID++));
        return id + stack;
    }
}
function pushStackDedup(stack, newStacks) {
    for (let i = 0; i < newStacks.length; i++) {
        let other = newStacks[i];
        if (other.pos == stack.pos && other.sameState(stack)) {
            if (newStacks[i].score < stack.score)
                newStacks[i] = stack;
            return;
        }
    }
    newStacks.push(stack);
}
class Dialect {
    constructor(source, flags, disabled) {
        this.source = source;
        this.flags = flags;
        this.disabled = disabled;
    }
    allows(term) { return !this.disabled || this.disabled[term] == 0; }
}
/// A parser holds the parse tables for a given grammar, as generated
/// by `lezer-generator`.
class Parser {
    /// @internal
    constructor(spec) {
        /// @internal
        this.bufferLength = DefaultBufferLength;
        /// @internal
        this.strict = false;
        this.cachedDialect = null;
        if (spec.version != 13 /* Version */)
            throw new RangeError(`Parser version (${spec.version}) doesn't match runtime version (${13 /* Version */})`);
        let tokenArray = decodeArray(spec.tokenData);
        let nodeNames = spec.nodeNames.split(" ");
        this.minRepeatTerm = nodeNames.length;
        this.context = spec.context;
        for (let i = 0; i < spec.repeatNodeCount; i++)
            nodeNames.push("");
        let nodeProps = [];
        for (let i = 0; i < nodeNames.length; i++)
            nodeProps.push([]);
        function setProp(nodeID, prop, value) {
            nodeProps[nodeID].push([prop, prop.deserialize(String(value))]);
        }
        if (spec.nodeProps)
            for (let propSpec of spec.nodeProps) {
                let prop = propSpec[0];
                for (let i = 1; i < propSpec.length;) {
                    let next = propSpec[i++];
                    if (next >= 0) {
                        setProp(next, prop, propSpec[i++]);
                    }
                    else {
                        let value = propSpec[i + -next];
                        for (let j = -next; j > 0; j--)
                            setProp(propSpec[i++], prop, value);
                        i++;
                    }
                }
            }
        this.specialized = new Uint16Array(spec.specialized ? spec.specialized.length : 0);
        this.specializers = [];
        if (spec.specialized)
            for (let i = 0; i < spec.specialized.length; i++) {
                this.specialized[i] = spec.specialized[i].term;
                this.specializers[i] = spec.specialized[i].get;
            }
        this.states = decodeArray(spec.states, Uint32Array);
        this.data = decodeArray(spec.stateData);
        this.goto = decodeArray(spec.goto);
        let topTerms = Object.keys(spec.topRules).map(r => spec.topRules[r][1]);
        this.nodeSet = new NodeSet(nodeNames.map((name, i) => NodeType.define({
            name: i >= this.minRepeatTerm ? undefined : name,
            id: i,
            props: nodeProps[i],
            top: topTerms.indexOf(i) > -1,
            error: i == 0,
            skipped: spec.skippedNodes && spec.skippedNodes.indexOf(i) > -1
        })));
        this.maxTerm = spec.maxTerm;
        this.tokenizers = spec.tokenizers.map(value => typeof value == "number" ? new TokenGroup(tokenArray, value) : value);
        this.topRules = spec.topRules;
        this.nested = (spec.nested || []).map(([name, value, endToken, placeholder]) => {
            return { name, value, end: new TokenGroup(decodeArray(endToken), 0), placeholder };
        });
        this.dialects = spec.dialects || {};
        this.dynamicPrecedences = spec.dynamicPrecedences || null;
        this.tokenPrecTable = spec.tokenPrec;
        this.termNames = spec.termNames || null;
        this.maxNode = this.nodeSet.types.length - 1;
        this.dialect = this.parseDialect();
        this.top = this.topRules[Object.keys(this.topRules)[0]];
    }
    /// Parse a given string or stream.
    parse(input, startPos = 0, context = {}) {
        if (typeof input == "string")
            input = stringInput(input);
        let cx = new Parse(this, input, startPos, context);
        for (;;) {
            let done = cx.advance();
            if (done)
                return done;
        }
    }
    /// Start an incremental parse.
    startParse(input, startPos = 0, context = {}) {
        if (typeof input == "string")
            input = stringInput(input);
        return new Parse(this, input, startPos, context);
    }
    /// Get a goto table entry @internal
    getGoto(state, term, loose = false) {
        let table = this.goto;
        if (term >= table[0])
            return -1;
        for (let pos = table[term + 1];;) {
            let groupTag = table[pos++], last = groupTag & 1;
            let target = table[pos++];
            if (last && loose)
                return target;
            for (let end = pos + (groupTag >> 1); pos < end; pos++)
                if (table[pos] == state)
                    return target;
            if (last)
                return -1;
        }
    }
    /// Check if this state has an action for a given terminal @internal
    hasAction(state, terminal) {
        let data = this.data;
        for (let set = 0; set < 2; set++) {
            for (let i = this.stateSlot(state, set ? 2 /* Skip */ : 1 /* Actions */), next;; i += 3) {
                if ((next = data[i]) == 65535 /* End */) {
                    if (data[i + 1] == 1 /* Next */)
                        next = data[i = pair(data, i + 2)];
                    else if (data[i + 1] == 2 /* Other */)
                        return pair(data, i + 2);
                    else
                        break;
                }
                if (next == terminal || next == 0 /* Err */)
                    return pair(data, i + 1);
            }
        }
        return 0;
    }
    /// @internal
    stateSlot(state, slot) {
        return this.states[(state * 6 /* Size */) + slot];
    }
    /// @internal
    stateFlag(state, flag) {
        return (this.stateSlot(state, 0 /* Flags */) & flag) > 0;
    }
    /// @internal
    findNested(state) {
        let flags = this.stateSlot(state, 0 /* Flags */);
        return flags & 4 /* StartNest */ ? this.nested[flags >> 10 /* NestShift */] : null;
    }
    /// @internal
    validAction(state, action) {
        if (action == this.stateSlot(state, 4 /* DefaultReduce */))
            return true;
        for (let i = this.stateSlot(state, 1 /* Actions */);; i += 3) {
            if (this.data[i] == 65535 /* End */) {
                if (this.data[i + 1] == 1 /* Next */)
                    i = pair(this.data, i + 2);
                else
                    return false;
            }
            if (action == pair(this.data, i + 1))
                return true;
        }
    }
    /// Get the states that can follow this one through shift actions or
    /// goto jumps. @internal
    nextStates(state) {
        let result = [];
        for (let i = this.stateSlot(state, 1 /* Actions */);; i += 3) {
            if (this.data[i] == 65535 /* End */) {
                if (this.data[i + 1] == 1 /* Next */)
                    i = pair(this.data, i + 2);
                else
                    break;
            }
            if ((this.data[i + 2] & (65536 /* ReduceFlag */ >> 16)) == 0) {
                let value = this.data[i + 1];
                if (!result.some((v, i) => (i & 1) && v == value))
                    result.push(this.data[i], value);
            }
        }
        return result;
    }
    /// @internal
    overrides(token, prev) {
        let iPrev = findOffset(this.data, this.tokenPrecTable, prev);
        return iPrev < 0 || findOffset(this.data, this.tokenPrecTable, token) < iPrev;
    }
    /// Configure the parser. Returns a new parser instance that has the
    /// given settings modified. Settings not provided in `config` are
    /// kept from the original parser.
    configure(config) {
        // Hideous reflection-based kludge to make it easy to create a
        // slightly modified copy of a parser.
        let copy = Object.assign(Object.create(Parser.prototype), this);
        if (config.props)
            copy.nodeSet = this.nodeSet.extend(...config.props);
        if (config.top) {
            let info = this.topRules[config.top];
            if (!info)
                throw new RangeError(`Invalid top rule name ${config.top}`);
            copy.top = info;
        }
        if (config.tokenizers)
            copy.tokenizers = this.tokenizers.map(t => {
                let found = config.tokenizers.find(r => r.from == t);
                return found ? found.to : t;
            });
        if (config.dialect)
            copy.dialect = this.parseDialect(config.dialect);
        if (config.nested)
            copy.nested = this.nested.map(obj => {
                if (!Object.prototype.hasOwnProperty.call(config.nested, obj.name))
                    return obj;
                return { name: obj.name, value: config.nested[obj.name], end: obj.end, placeholder: obj.placeholder };
            });
        if (config.strict != null)
            copy.strict = config.strict;
        if (config.bufferLength != null)
            copy.bufferLength = config.bufferLength;
        return copy;
    }
    /// Returns the name associated with a given term. This will only
    /// work for all terms when the parser was generated with the
    /// `--names` option. By default, only the names of tagged terms are
    /// stored.
    getName(term) {
        return this.termNames ? this.termNames[term] : String(term <= this.maxNode && this.nodeSet.types[term].name || term);
    }
    /// The eof term id is always allocated directly after the node
    /// types. @internal
    get eofTerm() { return this.maxNode + 1; }
    /// Tells you whether this grammar has any nested grammars.
    get hasNested() { return this.nested.length > 0; }
    /// The type of top node produced by the parser.
    get topNode() { return this.nodeSet.types[this.top[1]]; }
    /// @internal
    dynamicPrecedence(term) {
        let prec = this.dynamicPrecedences;
        return prec == null ? 0 : prec[term] || 0;
    }
    /// @internal
    parseDialect(dialect) {
        if (this.cachedDialect && this.cachedDialect.source == dialect)
            return this.cachedDialect;
        let values = Object.keys(this.dialects), flags = values.map(() => false);
        if (dialect)
            for (let part of dialect.split(" ")) {
                let id = values.indexOf(part);
                if (id >= 0)
                    flags[id] = true;
            }
        let disabled = null;
        for (let i = 0; i < values.length; i++)
            if (!flags[i]) {
                for (let j = this.dialects[values[i]], id; (id = this.data[j++]) != 65535 /* End */;)
                    (disabled || (disabled = new Uint8Array(this.maxTerm + 1)))[id] = 1;
            }
        return this.cachedDialect = new Dialect(dialect, flags, disabled);
    }
    /// (used by the output of the parser generator) @internal
    static deserialize(spec) {
        return new Parser(spec);
    }
}
function pair(data, off) { return data[off] | (data[off + 1] << 16); }
function findOffset(data, start, term) {
    for (let i = start, next; (next = data[i]) != 65535 /* End */; i++)
        if (next == term)
            return i - start;
    return -1;
}
function findFinished(stacks) {
    let best = null;
    for (let stack of stacks) {
        if (stack.pos == stack.p.input.length &&
            stack.p.parser.stateFlag(stack.state, 2 /* Accepting */) &&
            (!best || best.score < stack.score))
            best = stack;
    }
    return best;
}

function makeLineNumGreaterThanZero(lineNum) {
    return lineNum <= 0 ? 1 : lineNum;
}
const translateDiagnostics = (input = {
    checkResult: { checkItems: [], rawSQL: "" },
}, view) => {
    const { checkResult: { checkItems = [] }, } = input;
    const found = [];
    const maxLines = view.state.doc.lines;
    checkItems.forEach((item) => {
        const { error, errorCode, lineOffsets } = item;
        lineOffsets.forEach((lineOffset) => {
            const { StartLine, EndLine, StartOffset, EndOffset } = lineOffset;
            let from, to;
            if (maxLines >= EndLine) {
                const { from: startFrom, to: startTo } = view.state.doc.line(makeLineNumGreaterThanZero(StartLine) + 1);
                const { from: endFrom, to: endTo } = view.state.doc.line(makeLineNumGreaterThanZero(EndLine) + 1);
                if (startFrom === endFrom && startFrom === endTo) {
                    // same line
                    from = startFrom + StartOffset;
                    to = startTo + EndOffset;
                }
                else {
                    from = startFrom + StartOffset;
                    to = endFrom + EndOffset;
                }
                found.push({
                    from,
                    to,
                    severity: errorCode == -1 ? "error" : "warning",
                    message: error,
                });
            }
        });
    });
    return found.sort((a, b) => {
        if (a.from < b.from) {
            return -1;
        }
        if (a.from > b.from) {
            return 1;
        }
        return 0;
    });
};
const isSameTextAccordingToDoc = (editor, value = "") => {
    const doc = editor.state.doc.toJSON();
    const nextDocText = value.split(editor.state.lineBreak);
    if (doc.length !== nextDocText.length) {
        return false;
    }
    let i = doc.length - 1;
    while (i >= 0) {
        if (doc[i] !== nextDocText[i]) {
            return false;
        }
        i--;
    }
    return true;
};
function getStyle(element, attr) {
    var _a;
    return ((_a = window === null || window === void 0 ? void 0 : window.getComputedStyle(element, null)[attr]) !== null && _a !== void 0 ? _a : "");
}

// const SpecialChars = [
//   "(",
//   ")",
//   "{",
//   "}",
//   "[",
//   "]",
//   ",",
//   ";",
//   "-",
//   "*",
// ]
const LineHeight = 1;
const WordGap = 5;
class Drawer {
    constructor(view, ctx, cache) {
        ///
        this.tree = Tree.empty;
        this.view = view;
        this.ctx = ctx;
        this.cache = cache;
    }
    /// 绘制内容
    draw(view) {
        const doc = view.state.doc;
        const totalLines = doc.lines;
        const textIter = doc.iter(1);
        while (!textIter.done) {
            console.log(textIter);
            textIter.next();
        }
        let index = 1;
        while (index <= totalLines) {
            this.drawLine(doc.line(index));
            index++;
        }
    }
    /// 处理一行的字符
    drawLine(line) {
        const { text, from } = line;
        let offset = 2;
        if (/^\s.*/.test(text)) {
            offset = /\s+/.exec(text)[0].length + 10;
        }
        const tokens = text.split(/\s/);
        // Reset Prev Token
        this.posX = null;
        tokens.forEach((token, index) => {
            const pos = from + line.text.indexOf(token) + 1;
            let color;
            const { node } = this.view.domAtPos(pos);
            const cacheToken = this.cache.get(token + String(line.number));
            if (!cacheToken ||
                cacheToken.text !== token ||
                cacheToken.pos !== pos ||
                token !== "") {
                this.cache.set(token + String(line.number), {
                    text: token,
                    pos,
                });
                color = getStyle(node.parentElement, "color");
                this.drawToken(line, {
                    color,
                    text: token,
                }, offset);
                // }
            }
        });
    }
    /// 处理单个字符
    drawToken(line, token, offset) {
        const color = token.color || "white";
        // let length: number, x: number, y: number
        let x = offset; // token渲染的左偏移
        const y = line.number + line.number * LineHeight; // token渲染的上偏移
        const length = token.text.trimStart().length * 5; // token的长度，原始长度太小，增加五倍的长度
        if (this.posX) {
            x = this.posX + WordGap;
        }
        this.drawRect(x, y, color, length, LineHeight);
        this.posX = x + length;
    }
    /// 绘制色块
    drawRect(x, y, color, width, height) {
        this.ctx.fillStyle = color;
        this.ctx.fillRect(x, y, width, height);
    }
    /// 清理canvas内容
    clear(x, y, w, h) {
        this.ctx.clearRect(x, y, w, h);
    }
}

class TokenCache {
    constructor() {
        this.tokens = new Map();
    }
    get(key) {
        return this.tokens.get(key);
    }
    set(key, value) {
        this.tokens.set(key, value);
    }
    clear() { }
}

class Minimap {
    constructor(view) {
        this.view = view;
        this.cache = new TokenCache();
        this.minimap = new MinimapElement(view);
        this.canvas = new CanvasElement(this.minimap.node);
        this.viewbox = new Viewbox(this.minimap.node);
        this.drawer = new Drawer(view, this.canvas.ctx, this.cache);
    }
    resize(width, height) {
        this.minimap.resize(width, height);
        this.canvas.resize(width, height);
        this.viewbox.resize(width, height);
    }
    setBackground(bg) {
        this.minimap.setBackground(bg);
    }
    // move .cm-minimap under .cm-editor
    detach() {
        const panels = this.view.dom.querySelector(".cm-panels");
        this.view.dom.removeChild(panels);
        return this;
    }
    attach() {
        this.view.dom.appendChild(this.minimap.node);
        return this;
    }
    render(view) {
        if (!view) {
            view = this.view;
        }
        const cursorLine = view.state.doc.lineAt(view.state.selection.main.head);
        // clear canvas area below input line
        this.drawer.clear(0, cursorLine.number * 2, 80 * 5, view.dom.clientHeight);
        this.drawer.draw(view ? view : this.view);
    }
}

const minimapConfig = state.Facet.define({
    combine(configs) {
        let minimap, minimapWidth, minimapSide;
        for (const config of configs) {
            minimap = config.minimap;
            minimapWidth = config.minimapWidth;
            minimapSide = config.minimapSide;
        }
        return {
            minimap: minimap,
            minimapSide: minimapSide || "left",
            minimapWidth: minimapWidth || 80,
        };
    },
});
const createMinimapPanel = (view) => {
    const config = view.state.facet(minimapConfig);
    const minimapPanel = new Minimap(view);
    return {
        top: true,
        dom: minimapPanel.minimap.node,
        update(update) {
            if (update.docChanged) {
                minimapPanel.render(update.view);
            }
        },
        mount() {
            view.requestMeasure({
                write() {
                    // FIXME 读取minimap是否是打开状态，是的话，才进行read方法里面的操作
                },
                read(view) {
                    minimapPanel.resize(config.minimapWidth, view.dom.clientHeight);
                    minimapPanel.detach().attach();
                    minimapPanel.minimap.setSide(config.minimapSide);
                    minimapPanel.render();
                },
            });
        },
    };
};
const toggleMinimap = state.StateEffect.define();
const minimap = state.StateField.define({
    create() {
        return true;
    },
    update(update, tr) {
        for (const e of tr.effects)
            if (e.is(toggleMinimap))
                update = e.value;
        return update;
    },
    provide(field) {
        return showPanel.from(field, (show) => show ? createMinimapPanel : null);
    },
});
[
    minimap,
    view.EditorView.theme({
        ".cm-panel": {
            backgroundColor: "#21252b",
            color: "#abb2bf",
        },
    }),
];

function startFormat(view, configs) {
    const str = configs.parser(view.state.doc.toJSON().join(view.state.lineBreak));
    void Promise.resolve(str).then((resolveStr) => {
        view.dispatch({
            changes: [
                {
                    from: 0,
                    to: view.state.doc.length,
                    insert: resolveStr,
                },
            ],
        });
    });
}

// todo scroll offset， 设置一个fact，可以设置一个滚动偏移量，提前对数据进行加载
const scrollOptions = state.Facet.define({
    combine(offsets) {
        return state.combineConfig(offsets, {
            topOffset: 0,
            bottomOffset: 0,
        });
    },
});
function scrollToEdge({ onScrollToTop, onScrollToBottom, }) {
    // scroll to edge, top or bottom
    return view.EditorView.domEventHandlers({
        scroll: (_event, view) => {
            const config = view.state.facet(scrollOptions);
            const { topOffset, bottomOffset } = config;
            const scrollTop = view.scrollDOM.scrollTop;
            const clientHeight = view.scrollDOM.clientHeight;
            const scrollHeight = view.scrollDOM.scrollHeight;
            if (scrollTop + clientHeight > scrollHeight - bottomOffset) {
                scrollTop > 0 && (onScrollToBottom === null || onScrollToBottom === void 0 ? void 0 : onScrollToBottom());
            }
            if (scrollTop === topOffset) {
                onScrollToTop === null || onScrollToTop === void 0 ? void 0 : onScrollToTop();
            }
        },
    });
}

/// feature flags
const MINIMAP_FLAG = false;

/// editable
const editableCompart = new state.Compartment();
function reconfigure$3(editable) {
    return editableCompart.reconfigure(view.EditorView.editable.of(editable));
}
function useEditableProp(value, view$1) {
    const [compart] = React.useState(() => editableCompart.of(view.EditorView.editable.of(value)));
    React.useEffect(() => {
        view$1 === null || view$1 === void 0 ? void 0 : view$1.dispatch({
            effects: [reconfigure$3(value)],
        });
    }, [value, view$1]);
    return compart;
}

function walk(tree, traveler) {
    let lastNode;
    tree.iterate({
        enter(node, from, to) {
            const travel = traveler[node.name];
            if (travel) {
                travel(node, from, to, lastNode);
            }
            lastNode = { node, from, to };
        },
    });
}
function analysis(state) {
    const syntaxNodes = language.syntaxTree(state);
    const localTableInfos = {};
    let activeTable = null;
    let activeTableKey = null;
    let activeTableKeyType = null;
    walk(syntaxNodes, {
        Identifier(_node, from, to, { node: prevNode, from: prevFrom, to: prevTo }) {
            const tokenName = state.doc.sliceString(from, to);
            if (prevNode.name.toLowerCase() === "keyword" &&
                state.doc.sliceString(prevFrom, prevTo).toLowerCase() ===
                    "table") {
                localTableInfos[tokenName] = [];
                activeTable = tokenName;
                activeTableKey = null;
                activeTableKeyType = null;
            }
            else {
                if (activeTable && activeTableKey) {
                    activeTableKeyType = tokenName;
                }
                if (activeTable && !activeTableKey) {
                    activeTableKey = tokenName;
                }
            }
        },
        Type(_node, from, to) {
            const tokenName = state.doc.sliceString(from, to);
            if (activeTable && activeTableKey && !activeTableKeyType) {
                activeTableKeyType = tokenName;
            }
        },
        String(_node, from, to) {
            const tokenName = state.doc.sliceString(from, to);
            if (activeTable && activeTableKey && !activeTableKeyType) {
                activeTableKeyType = tokenName;
            }
        },
        Punctuation(_node, from, to) {
            const tokenName = state.doc.sliceString(from, to);
            if (tokenName === ",") {
                if (activeTableKey && activeTableKeyType) {
                    localTableInfos[activeTable].push({
                        name: activeTableKey,
                        type: activeTableKeyType,
                    });
                    activeTableKey = null;
                    activeTableKeyType = null;
                }
            }
        },
        ")"() {
            if (activeTableKey && activeTableKeyType) {
                localTableInfos[activeTable].push({
                    name: activeTableKey,
                    type: activeTableKeyType,
                });
                activeTableKey = null;
                activeTableKeyType = null;
            }
        },
        Statement() {
            /// 新的语句开始，重置参数
            if (activeTable) {
                activeTable = null;
                activeTableKey = null;
                activeTableKeyType = null;
            }
        },
        Keyword(_node, from, to) {
            const tokenName = state.doc.sliceString(from, to);
            if (tokenName.toLowerCase() === "with") {
                if (activeTable && !activeTableKey && !activeTableKeyType) {
                    localTableInfos[activeTable].push({});
                }
            }
        },
    });
    return localTableInfos;
}
const sqlAnalysisEffect = state.StateEffect.define({});
const sqlAnalysisField = state.StateField.define({
    create(state) {
        return analysis(state);
    },
    update(value, { effects }) {
        for (let i = 0; i < effects.length; i++) {
            const effect = effects[i];
            if (effect.is(sqlAnalysisEffect)) {
                return effect.value;
            }
        }
        return value;
    },
});
function analysisSql() {
    let timer = null;
    return [
        sqlAnalysisField,
        view.EditorView.updateListener.of(({ docChanged, state, view }) => {
            if (docChanged) {
                if (timer) {
                    clearTimeout(timer);
                    timer = null;
                }
                const resultTable = analysis(state);
                timer = setTimeout(() => {
                    view.dispatch({
                        effects: [sqlAnalysisEffect.of(resultTable)],
                    });
                }, 400);
            }
        }),
    ];
}

function transferSnippetToCompletion(snippets) {
    const snippetsResult = {
        "": [],
    };
    snippets.forEach((snip) => {
        const snippetConfig = {
            label: snip.label,
            info: snip.info,
            apply: autocomplete.snippet(snip.template),
        };
        if (snip.hit) {
            //
            const hitKeywords = snip.hit.split("-").reverse();
            let tempMap;
            if (!snippetsResult[hitKeywords[0]]) {
                tempMap = snippetsResult[hitKeywords[0]] = {};
            }
            else {
                tempMap = snippetsResult[hitKeywords[0]];
            }
            hitKeywords.slice(1).forEach((keyword, index) => {
                if (!tempMap[keyword]) {
                    tempMap[keyword] = [];
                }
                if (index === hitKeywords.length - 2) {
                    tempMap[keyword].push(snippetConfig);
                }
                else {
                    tempMap = tempMap[keyword] = {};
                }
            });
        }
        else {
            snippetsResult[""].push(snippetConfig);
        }
    });
    return snippetsResult;
}
const snippetsFacet = state.Facet.define({
    combine(value) {
        if (value.length) {
            return transferSnippetToCompletion(value.slice(-1)[0]);
        }
        return [];
    },
});

function tokenBefore(tree) {
    const cursor = tree.cursor.moveTo(tree.from, -1);
    while (/Comment/.test(cursor.name)) {
        cursor.moveTo(cursor.from, -1);
    }
    return cursor.node;
}
function stripQuotes(name) {
    const quoted = /^[`'"](.*)[`'"]$/.exec(name);
    return quoted ? quoted[1] : name;
}
function sourceContext(state, startPos) {
    const pos = language.syntaxTree(state).resolve(startPos, -1);
    let empty = false;
    if (pos.name == "Identifier" || pos.name == "QuotedIdentifier") {
        empty = false;
        let parent = null;
        const dot = tokenBefore(pos);
        if (dot && dot.name == ".") {
            const before = tokenBefore(dot);
            if ((before && before.name == "Identifier") ||
                before.name == "QuotedIdentifier")
                parent = stripQuotes(state.sliceDoc(before.from, before.to));
        }
        return {
            parent,
            from: pos.from,
            quoted: pos.name == "QuotedIdentifier"
                ? state.sliceDoc(pos.from, pos.from + 1)
                : null,
        };
    }
    else if (pos.name == ".") {
        const before = tokenBefore(pos);
        if ((before && before.name == "Identifier") ||
            before.name == "QuotedIdentifier") {
            return {
                parent: stripQuotes(state.sliceDoc(before.from, before.to)),
                from: startPos,
                quoted: null,
            };
        }
        else {
            empty = true;
        }
    }
    else {
        empty = true;
    }
    return { parent: null, from: startPos, quoted: null, empty };
}
function maybeQuoteCompletions(quote, completions) {
    if (!quote)
        return completions;
    return completions.map((c) => (Object.assign(Object.assign({}, c), { label: quote + c.label + quote, apply: undefined })));
}
function transferTableSourceToCompletions(source) {
    const schema = {};
    const tableNames = Object.keys(source);
    tableNames.forEach((name) => {
        if (!schema[name]) {
            schema[name] = [];
        }
        for (let index = 0; index < source[name].length; index++) {
            const item = source[name][index];
            if (Object.keys(item).length === 0) {
                break;
            }
            schema[name].push({
                label: item.name,
                detail: `from ${name}`,
                info: ``,
                type: "property",
                boost: 2,
            });
        }
    });
    return schema;
}
const Span = /^\w*$/, QuotedSpan = /^[`'"]?\w*[`'"]?$/;
const usageStack = {};
function increaseUsageCountApply(view, completion, from, to) {
    if (!usageStack[completion.label]) {
        usageStack[completion.label] = 1;
    }
    else {
        usageStack[completion.label]++;
    }
    view.dispatch({
        changes: [
            {
                from,
                to,
                insert: completion.label,
            },
        ],
        selection: {
            anchor: from + completion.label.length,
            head: from + completion.label.length,
        },
    });
}
function rearrangeOptionsByContext(input, output) {
    const baseBoost = 0;
    /// 重置 Usage
    // usageStack = {}
    return [
        ...input.map((put, index) => {
            var _a;
            return (Object.assign(Object.assign({}, put), { boost: baseBoost - index + ((_a = usageStack[put.label]) !== null && _a !== void 0 ? _a : 0), apply: increaseUsageCountApply }));
        }),
        ...output.map((put, index) => {
            var _a;
            return (Object.assign(Object.assign({}, put), { boost: baseBoost + index + 1 + ((_a = usageStack[put.label]) !== null && _a !== void 0 ? _a : 0), apply: increaseUsageCountApply }));
        }),
    ];
}
/// 按照pos往前一个一个地截取字符，直到遇到匹配的补全。
function tokenBeforePos(state, pos) {
    let char = state.doc.sliceString(pos - 1, pos);
    while (char === " ") {
        pos--;
        char = state.doc.sliceString(pos - 1, pos);
    }
    const cursor = language.syntaxTree(state).cursor(pos, -1);
    return {
        node: {
            value: state.doc.sliceString(cursor.from, cursor.to),
            from: pos - 1,
            to: pos,
            cursor,
        },
    };
}
function getTokenNameByPos(state, from, to) {
    return state.doc.sliceString(from, to);
}
function completeFromSchema(schema, tables, defaultTable) {
    // eslint-disable-next-line @typescript-eslint/no-unsafe-assignment
    let byTable = Object.create(null);
    for (const table in schema)
        byTable[table] = schema[table].map((val) => {
            return typeof val == "string"
                ? { label: val, type: "property" }
                : val;
        });
    const topOptions = (tables ||
        Object.keys(byTable).map((name) => ({ label: name, type: "type" }))).concat((defaultTable && byTable[defaultTable]) || []);
    return (context) => {
        var _a;
        /// 补充代码段补全到CompletionResult
        const snippets = context.state.facet(snippetsFacet);
        const localTableSource = context.state.field(sqlAnalysisField);
        const localSchema = transferTableSourceToCompletions(localTableSource);
        const localOptions = Object.keys(localTableSource).map((name) => ({ label: name, type: "property" }));
        byTable = Object.assign({}, byTable, localSchema);
        const { parent, from, quoted, empty } = sourceContext(context.state, context.pos);
        if (empty && !context.explicit) {
            let targetOptions = snippets;
            let node = tokenBeforePos(context.state, context.pos).node;
            while (targetOptions[node.value] &&
                !Array.isArray(targetOptions[node.value])) {
                const char = getTokenNameByPos(context.state, node.cursor.from, node.cursor.to);
                node = tokenBeforePos(context.state, node.from).node;
                targetOptions = targetOptions[char];
            }
            if (context.pos === node.to) {
                return null;
            }
            const options = (_a = targetOptions[node.value]) !== null && _a !== void 0 ? _a : [];
            //// 匹配到 to 就展示输出源的 table 名称
            if (node.value.toLowerCase() === "into") {
                return {
                    from,
                    to: context.pos,
                    options: maybeQuoteCompletions(quoted, [
                        ...options,
                        ...rearrangeOptionsByContext([...topOptions], localOptions),
                    ]),
                    span: Span,
                };
            }
            //// 匹配到 from 就展示输入源的 table 名称
            if (node.value.toLowerCase() === "from") {
                return {
                    from,
                    to: context.pos,
                    options: maybeQuoteCompletions(quoted, [
                        ...options,
                        ...rearrangeOptionsByContext(localOptions, [
                            ...topOptions,
                        ]),
                    ]),
                    span: Span,
                };
            }
            //// 最后有匹配到所有的键，就给出匹配到的补全
            if (options.length) {
                return {
                    from,
                    to: context.pos,
                    options: maybeQuoteCompletions(quoted, options),
                    span: Span,
                };
            }
            return null;
        }
        let options = topOptions;
        options = options.concat(snippets[""]);
        if (parent) {
            const columns = byTable[parent];
            if (!columns)
                return null;
            options = columns;
        }
        const quoteAfter = quoted &&
            context.state.sliceDoc(context.pos, context.pos + 1) == quoted;
        return {
            from,
            to: quoteAfter ? context.pos + 1 : undefined,
            options: maybeQuoteCompletions(quoted, options),
            span: quoted ? QuotedSpan : Span,
        };
    };
}

const keymapPromptConfig = state.Facet.define({
    combine(KeymapPromptConfigs) {
        return state.combineConfig(KeymapPromptConfigs, {
            placement: "leftbottom",
        });
    },
});
/// 设置prompt的文字，展示在编辑器右下角
const setPromptText = state.StateEffect.define({
    map: (value) => value !== null && value !== void 0 ? value : "",
});
const prompt = view.ViewPlugin.fromClass(class {
    constructor(view) {
        this.onAninationEnd = () => {
            this.dom.removeEventListener("transitionend", this.onAninationEnd);
            this.view.dom.removeChild(this.dom);
            clearTimeout(this.timer);
            this.timer = null;
        };
        this.view = view;
    }
    update(update) {
        const { transactions } = update;
        for (let index = 0; index < transactions.length; index++) {
            const { effects } = transactions[index];
            for (let index = 0; index < effects.length; index++) {
                const effect = effects[index];
                if (effect.is(setPromptText) && effect.value) {
                    if (this.timer) {
                        this.setupTimer();
                    }
                    else {
                        this.init();
                    }
                    this.attach(effect.value);
                    break;
                }
            }
        }
    }
    init() {
        const dom = document.createElement("div");
        dom.classList.add("cm-keymap-prompt");
        dom.addEventListener("transitionend", this.onAninationEnd);
        this.view.dom.appendChild(dom);
        this.dom = dom;
        this.setupTimer();
    }
    attach(promptText) {
        // fadein
        this.dom.style.opacity = "0.8";
        this.dom.innerText = promptText;
    }
    detach() {
        // fadeout
        this.dom.style.opacity = "0";
    }
    setupTimer() {
        this.timer && clearTimeout(this.timer);
        this.timer = setTimeout(() => {
            this.detach();
        }, 2000);
    }
    destory() {
        this.detach();
    }
});
const propmtBaseTheme = view.EditorView.baseTheme({
    ".cm-keymap-prompt": {
        position: "absolute",
        width: "auto",
        height: "60px",
        background: "black",
        opacity: 0,
        bottom: "24px",
        right: "24px",
        borderRadius: "8px",
        transition: "opacity 0.2s",
        textAlign: "center",
        lineHeight: "60px",
        fontSize: "20px",
        padding: "0 16px",
    },
});
function keymapPrompt(config) {
    return [keymapPromptConfig.of(config), prompt, propmtBaseTheme];
}

/**
 * The base implementation of `_.findIndex` and `_.findLastIndex` without
 * support for iteratee shorthands.
 *
 * @private
 * @param {Array} array The array to inspect.
 * @param {Function} predicate The function invoked per iteration.
 * @param {number} fromIndex The index to search from.
 * @param {boolean} [fromRight] Specify iterating from right to left.
 * @returns {number} Returns the index of the matched value, else `-1`.
 */
function baseFindIndex(array, predicate, fromIndex, fromRight) {
  var length = array.length,
      index = fromIndex + (fromRight ? 1 : -1);

  while ((fromRight ? index-- : ++index < length)) {
    if (predicate(array[index], index, array)) {
      return index;
    }
  }
  return -1;
}

var _baseFindIndex = baseFindIndex;

/**
 * The base implementation of `_.isNaN` without support for number objects.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is `NaN`, else `false`.
 */
function baseIsNaN(value) {
  return value !== value;
}

var _baseIsNaN = baseIsNaN;

/**
 * A specialized version of `_.indexOf` which performs strict equality
 * comparisons of values, i.e. `===`.
 *
 * @private
 * @param {Array} array The array to inspect.
 * @param {*} value The value to search for.
 * @param {number} fromIndex The index to search from.
 * @returns {number} Returns the index of the matched value, else `-1`.
 */
function strictIndexOf(array, value, fromIndex) {
  var index = fromIndex - 1,
      length = array.length;

  while (++index < length) {
    if (array[index] === value) {
      return index;
    }
  }
  return -1;
}

var _strictIndexOf = strictIndexOf;

/**
 * The base implementation of `_.indexOf` without `fromIndex` bounds checks.
 *
 * @private
 * @param {Array} array The array to inspect.
 * @param {*} value The value to search for.
 * @param {number} fromIndex The index to search from.
 * @returns {number} Returns the index of the matched value, else `-1`.
 */
function baseIndexOf(array, value, fromIndex) {
  return value === value
    ? _strictIndexOf(array, value, fromIndex)
    : _baseFindIndex(array, _baseIsNaN, fromIndex);
}

var _baseIndexOf = baseIndexOf;

var commonjsGlobal = typeof globalThis !== 'undefined' ? globalThis : typeof window !== 'undefined' ? window : typeof global !== 'undefined' ? global : typeof self !== 'undefined' ? self : {};

function createCommonjsModule(fn) {
  var module = { exports: {} };
	return fn(module, module.exports), module.exports;
}

/** Detect free variable `global` from Node.js. */

var freeGlobal = typeof commonjsGlobal == 'object' && commonjsGlobal && commonjsGlobal.Object === Object && commonjsGlobal;

var _freeGlobal = freeGlobal;

/** Detect free variable `self`. */
var freeSelf = typeof self == 'object' && self && self.Object === Object && self;

/** Used as a reference to the global object. */
var root = _freeGlobal || freeSelf || Function('return this')();

var _root = root;

/** Built-in value references. */
var Symbol = _root.Symbol;

var _Symbol = Symbol;

/** Used for built-in method references. */
var objectProto$7 = Object.prototype;

/** Used to check objects for own properties. */
var hasOwnProperty$5 = objectProto$7.hasOwnProperty;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var nativeObjectToString$1 = objectProto$7.toString;

/** Built-in value references. */
var symToStringTag$1 = _Symbol ? _Symbol.toStringTag : undefined;

/**
 * A specialized version of `baseGetTag` which ignores `Symbol.toStringTag` values.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the raw `toStringTag`.
 */
function getRawTag(value) {
  var isOwn = hasOwnProperty$5.call(value, symToStringTag$1),
      tag = value[symToStringTag$1];

  try {
    value[symToStringTag$1] = undefined;
    var unmasked = true;
  } catch (e) {}

  var result = nativeObjectToString$1.call(value);
  if (unmasked) {
    if (isOwn) {
      value[symToStringTag$1] = tag;
    } else {
      delete value[symToStringTag$1];
    }
  }
  return result;
}

var _getRawTag = getRawTag;

/** Used for built-in method references. */
var objectProto$6 = Object.prototype;

/**
 * Used to resolve the
 * [`toStringTag`](http://ecma-international.org/ecma-262/7.0/#sec-object.prototype.tostring)
 * of values.
 */
var nativeObjectToString = objectProto$6.toString;

/**
 * Converts `value` to a string using `Object.prototype.toString`.
 *
 * @private
 * @param {*} value The value to convert.
 * @returns {string} Returns the converted string.
 */
function objectToString(value) {
  return nativeObjectToString.call(value);
}

var _objectToString = objectToString;

/** `Object#toString` result references. */
var nullTag = '[object Null]',
    undefinedTag = '[object Undefined]';

/** Built-in value references. */
var symToStringTag = _Symbol ? _Symbol.toStringTag : undefined;

/**
 * The base implementation of `getTag` without fallbacks for buggy environments.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the `toStringTag`.
 */
function baseGetTag(value) {
  if (value == null) {
    return value === undefined ? undefinedTag : nullTag;
  }
  return (symToStringTag && symToStringTag in Object(value))
    ? _getRawTag(value)
    : _objectToString(value);
}

var _baseGetTag = baseGetTag;

/**
 * Checks if `value` is the
 * [language type](http://www.ecma-international.org/ecma-262/7.0/#sec-ecmascript-language-types)
 * of `Object`. (e.g. arrays, functions, objects, regexes, `new Number(0)`, and `new String('')`)
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an object, else `false`.
 * @example
 *
 * _.isObject({});
 * // => true
 *
 * _.isObject([1, 2, 3]);
 * // => true
 *
 * _.isObject(_.noop);
 * // => true
 *
 * _.isObject(null);
 * // => false
 */
function isObject(value) {
  var type = typeof value;
  return value != null && (type == 'object' || type == 'function');
}

var isObject_1 = isObject;

/** `Object#toString` result references. */
var asyncTag = '[object AsyncFunction]',
    funcTag$1 = '[object Function]',
    genTag = '[object GeneratorFunction]',
    proxyTag = '[object Proxy]';

/**
 * Checks if `value` is classified as a `Function` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a function, else `false`.
 * @example
 *
 * _.isFunction(_);
 * // => true
 *
 * _.isFunction(/abc/);
 * // => false
 */
function isFunction(value) {
  if (!isObject_1(value)) {
    return false;
  }
  // The use of `Object#toString` avoids issues with the `typeof` operator
  // in Safari 9 which returns 'object' for typed arrays and other constructors.
  var tag = _baseGetTag(value);
  return tag == funcTag$1 || tag == genTag || tag == asyncTag || tag == proxyTag;
}

var isFunction_1 = isFunction;

/** Used as references for various `Number` constants. */
var MAX_SAFE_INTEGER$2 = 9007199254740991;

/**
 * Checks if `value` is a valid array-like length.
 *
 * **Note:** This method is loosely based on
 * [`ToLength`](http://ecma-international.org/ecma-262/7.0/#sec-tolength).
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a valid length, else `false`.
 * @example
 *
 * _.isLength(3);
 * // => true
 *
 * _.isLength(Number.MIN_VALUE);
 * // => false
 *
 * _.isLength(Infinity);
 * // => false
 *
 * _.isLength('3');
 * // => false
 */
function isLength(value) {
  return typeof value == 'number' &&
    value > -1 && value % 1 == 0 && value <= MAX_SAFE_INTEGER$2;
}

var isLength_1 = isLength;

/**
 * Checks if `value` is array-like. A value is considered array-like if it's
 * not a function and has a `value.length` that's an integer greater than or
 * equal to `0` and less than or equal to `Number.MAX_SAFE_INTEGER`.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is array-like, else `false`.
 * @example
 *
 * _.isArrayLike([1, 2, 3]);
 * // => true
 *
 * _.isArrayLike(document.body.children);
 * // => true
 *
 * _.isArrayLike('abc');
 * // => true
 *
 * _.isArrayLike(_.noop);
 * // => false
 */
function isArrayLike(value) {
  return value != null && isLength_1(value.length) && !isFunction_1(value);
}

var isArrayLike_1 = isArrayLike;

/**
 * Checks if `value` is classified as an `Array` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an array, else `false`.
 * @example
 *
 * _.isArray([1, 2, 3]);
 * // => true
 *
 * _.isArray(document.body.children);
 * // => false
 *
 * _.isArray('abc');
 * // => false
 *
 * _.isArray(_.noop);
 * // => false
 */
var isArray = Array.isArray;

var isArray_1 = isArray;

/**
 * Checks if `value` is object-like. A value is object-like if it's not `null`
 * and has a `typeof` result of "object".
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is object-like, else `false`.
 * @example
 *
 * _.isObjectLike({});
 * // => true
 *
 * _.isObjectLike([1, 2, 3]);
 * // => true
 *
 * _.isObjectLike(_.noop);
 * // => false
 *
 * _.isObjectLike(null);
 * // => false
 */
function isObjectLike(value) {
  return value != null && typeof value == 'object';
}

var isObjectLike_1 = isObjectLike;

/** `Object#toString` result references. */
var stringTag$1 = '[object String]';

/**
 * Checks if `value` is classified as a `String` primitive or object.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a string, else `false`.
 * @example
 *
 * _.isString('abc');
 * // => true
 *
 * _.isString(1);
 * // => false
 */
function isString(value) {
  return typeof value == 'string' ||
    (!isArray_1(value) && isObjectLike_1(value) && _baseGetTag(value) == stringTag$1);
}

var isString_1 = isString;

/** Used to match a single whitespace character. */
var reWhitespace = /\s/;

/**
 * Used by `_.trim` and `_.trimEnd` to get the index of the last non-whitespace
 * character of `string`.
 *
 * @private
 * @param {string} string The string to inspect.
 * @returns {number} Returns the index of the last non-whitespace character.
 */
function trimmedEndIndex(string) {
  var index = string.length;

  while (index-- && reWhitespace.test(string.charAt(index))) {}
  return index;
}

var _trimmedEndIndex = trimmedEndIndex;

/** Used to match leading whitespace. */
var reTrimStart = /^\s+/;

/**
 * The base implementation of `_.trim`.
 *
 * @private
 * @param {string} string The string to trim.
 * @returns {string} Returns the trimmed string.
 */
function baseTrim(string) {
  return string
    ? string.slice(0, _trimmedEndIndex(string) + 1).replace(reTrimStart, '')
    : string;
}

var _baseTrim = baseTrim;

/** `Object#toString` result references. */
var symbolTag = '[object Symbol]';

/**
 * Checks if `value` is classified as a `Symbol` primitive or object.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a symbol, else `false`.
 * @example
 *
 * _.isSymbol(Symbol.iterator);
 * // => true
 *
 * _.isSymbol('abc');
 * // => false
 */
function isSymbol(value) {
  return typeof value == 'symbol' ||
    (isObjectLike_1(value) && _baseGetTag(value) == symbolTag);
}

var isSymbol_1 = isSymbol;

/** Used as references for various `Number` constants. */
var NAN = 0 / 0;

/** Used to detect bad signed hexadecimal string values. */
var reIsBadHex = /^[-+]0x[0-9a-f]+$/i;

/** Used to detect binary string values. */
var reIsBinary = /^0b[01]+$/i;

/** Used to detect octal string values. */
var reIsOctal = /^0o[0-7]+$/i;

/** Built-in method references without a dependency on `root`. */
var freeParseInt = parseInt;

/**
 * Converts `value` to a number.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to process.
 * @returns {number} Returns the number.
 * @example
 *
 * _.toNumber(3.2);
 * // => 3.2
 *
 * _.toNumber(Number.MIN_VALUE);
 * // => 5e-324
 *
 * _.toNumber(Infinity);
 * // => Infinity
 *
 * _.toNumber('3.2');
 * // => 3.2
 */
function toNumber(value) {
  if (typeof value == 'number') {
    return value;
  }
  if (isSymbol_1(value)) {
    return NAN;
  }
  if (isObject_1(value)) {
    var other = typeof value.valueOf == 'function' ? value.valueOf() : value;
    value = isObject_1(other) ? (other + '') : other;
  }
  if (typeof value != 'string') {
    return value === 0 ? value : +value;
  }
  value = _baseTrim(value);
  var isBinary = reIsBinary.test(value);
  return (isBinary || reIsOctal.test(value))
    ? freeParseInt(value.slice(2), isBinary ? 2 : 8)
    : (reIsBadHex.test(value) ? NAN : +value);
}

var toNumber_1 = toNumber;

/** Used as references for various `Number` constants. */
var INFINITY$1 = 1 / 0,
    MAX_INTEGER = 1.7976931348623157e+308;

/**
 * Converts `value` to a finite number.
 *
 * @static
 * @memberOf _
 * @since 4.12.0
 * @category Lang
 * @param {*} value The value to convert.
 * @returns {number} Returns the converted number.
 * @example
 *
 * _.toFinite(3.2);
 * // => 3.2
 *
 * _.toFinite(Number.MIN_VALUE);
 * // => 5e-324
 *
 * _.toFinite(Infinity);
 * // => 1.7976931348623157e+308
 *
 * _.toFinite('3.2');
 * // => 3.2
 */
function toFinite(value) {
  if (!value) {
    return value === 0 ? value : 0;
  }
  value = toNumber_1(value);
  if (value === INFINITY$1 || value === -INFINITY$1) {
    var sign = (value < 0 ? -1 : 1);
    return sign * MAX_INTEGER;
  }
  return value === value ? value : 0;
}

var toFinite_1 = toFinite;

/**
 * Converts `value` to an integer.
 *
 * **Note:** This method is loosely based on
 * [`ToInteger`](http://www.ecma-international.org/ecma-262/7.0/#sec-tointeger).
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to convert.
 * @returns {number} Returns the converted integer.
 * @example
 *
 * _.toInteger(3.2);
 * // => 3
 *
 * _.toInteger(Number.MIN_VALUE);
 * // => 0
 *
 * _.toInteger(Infinity);
 * // => 1.7976931348623157e+308
 *
 * _.toInteger('3.2');
 * // => 3
 */
function toInteger(value) {
  var result = toFinite_1(value),
      remainder = result % 1;

  return result === result ? (remainder ? result - remainder : result) : 0;
}

var toInteger_1 = toInteger;

/**
 * A specialized version of `_.map` for arrays without support for iteratee
 * shorthands.
 *
 * @private
 * @param {Array} [array] The array to iterate over.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns the new mapped array.
 */
function arrayMap(array, iteratee) {
  var index = -1,
      length = array == null ? 0 : array.length,
      result = Array(length);

  while (++index < length) {
    result[index] = iteratee(array[index], index, array);
  }
  return result;
}

var _arrayMap = arrayMap;

/**
 * The base implementation of `_.values` and `_.valuesIn` which creates an
 * array of `object` property values corresponding to the property names
 * of `props`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {Array} props The property names to get values for.
 * @returns {Object} Returns the array of property values.
 */
function baseValues(object, props) {
  return _arrayMap(props, function(key) {
    return object[key];
  });
}

var _baseValues = baseValues;

/**
 * The base implementation of `_.times` without support for iteratee shorthands
 * or max array length checks.
 *
 * @private
 * @param {number} n The number of times to invoke `iteratee`.
 * @param {Function} iteratee The function invoked per iteration.
 * @returns {Array} Returns the array of results.
 */
function baseTimes(n, iteratee) {
  var index = -1,
      result = Array(n);

  while (++index < n) {
    result[index] = iteratee(index);
  }
  return result;
}

var _baseTimes = baseTimes;

/** `Object#toString` result references. */
var argsTag$1 = '[object Arguments]';

/**
 * The base implementation of `_.isArguments`.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an `arguments` object,
 */
function baseIsArguments(value) {
  return isObjectLike_1(value) && _baseGetTag(value) == argsTag$1;
}

var _baseIsArguments = baseIsArguments;

/** Used for built-in method references. */
var objectProto$5 = Object.prototype;

/** Used to check objects for own properties. */
var hasOwnProperty$4 = objectProto$5.hasOwnProperty;

/** Built-in value references. */
var propertyIsEnumerable = objectProto$5.propertyIsEnumerable;

/**
 * Checks if `value` is likely an `arguments` object.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is an `arguments` object,
 *  else `false`.
 * @example
 *
 * _.isArguments(function() { return arguments; }());
 * // => true
 *
 * _.isArguments([1, 2, 3]);
 * // => false
 */
var isArguments = _baseIsArguments(function() { return arguments; }()) ? _baseIsArguments : function(value) {
  return isObjectLike_1(value) && hasOwnProperty$4.call(value, 'callee') &&
    !propertyIsEnumerable.call(value, 'callee');
};

var isArguments_1 = isArguments;

/**
 * This method returns `false`.
 *
 * @static
 * @memberOf _
 * @since 4.13.0
 * @category Util
 * @returns {boolean} Returns `false`.
 * @example
 *
 * _.times(2, _.stubFalse);
 * // => [false, false]
 */
function stubFalse() {
  return false;
}

var stubFalse_1 = stubFalse;

var isBuffer_1 = createCommonjsModule(function (module, exports) {
/** Detect free variable `exports`. */
var freeExports = exports && !exports.nodeType && exports;

/** Detect free variable `module`. */
var freeModule = freeExports && 'object' == 'object' && module && !module.nodeType && module;

/** Detect the popular CommonJS extension `module.exports`. */
var moduleExports = freeModule && freeModule.exports === freeExports;

/** Built-in value references. */
var Buffer = moduleExports ? _root.Buffer : undefined;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeIsBuffer = Buffer ? Buffer.isBuffer : undefined;

/**
 * Checks if `value` is a buffer.
 *
 * @static
 * @memberOf _
 * @since 4.3.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a buffer, else `false`.
 * @example
 *
 * _.isBuffer(new Buffer(2));
 * // => true
 *
 * _.isBuffer(new Uint8Array(2));
 * // => false
 */
var isBuffer = nativeIsBuffer || stubFalse_1;

module.exports = isBuffer;
});

/** Used as references for various `Number` constants. */
var MAX_SAFE_INTEGER$1 = 9007199254740991;

/** Used to detect unsigned integer values. */
var reIsUint = /^(?:0|[1-9]\d*)$/;

/**
 * Checks if `value` is a valid array-like index.
 *
 * @private
 * @param {*} value The value to check.
 * @param {number} [length=MAX_SAFE_INTEGER] The upper bounds of a valid index.
 * @returns {boolean} Returns `true` if `value` is a valid index, else `false`.
 */
function isIndex(value, length) {
  var type = typeof value;
  length = length == null ? MAX_SAFE_INTEGER$1 : length;

  return !!length &&
    (type == 'number' ||
      (type != 'symbol' && reIsUint.test(value))) &&
        (value > -1 && value % 1 == 0 && value < length);
}

var _isIndex = isIndex;

/** `Object#toString` result references. */
var argsTag = '[object Arguments]',
    arrayTag = '[object Array]',
    boolTag = '[object Boolean]',
    dateTag = '[object Date]',
    errorTag = '[object Error]',
    funcTag = '[object Function]',
    mapTag$2 = '[object Map]',
    numberTag = '[object Number]',
    objectTag$1 = '[object Object]',
    regexpTag = '[object RegExp]',
    setTag$2 = '[object Set]',
    stringTag = '[object String]',
    weakMapTag$1 = '[object WeakMap]';

var arrayBufferTag = '[object ArrayBuffer]',
    dataViewTag$1 = '[object DataView]',
    float32Tag = '[object Float32Array]',
    float64Tag = '[object Float64Array]',
    int8Tag = '[object Int8Array]',
    int16Tag = '[object Int16Array]',
    int32Tag = '[object Int32Array]',
    uint8Tag = '[object Uint8Array]',
    uint8ClampedTag = '[object Uint8ClampedArray]',
    uint16Tag = '[object Uint16Array]',
    uint32Tag = '[object Uint32Array]';

/** Used to identify `toStringTag` values of typed arrays. */
var typedArrayTags = {};
typedArrayTags[float32Tag] = typedArrayTags[float64Tag] =
typedArrayTags[int8Tag] = typedArrayTags[int16Tag] =
typedArrayTags[int32Tag] = typedArrayTags[uint8Tag] =
typedArrayTags[uint8ClampedTag] = typedArrayTags[uint16Tag] =
typedArrayTags[uint32Tag] = true;
typedArrayTags[argsTag] = typedArrayTags[arrayTag] =
typedArrayTags[arrayBufferTag] = typedArrayTags[boolTag] =
typedArrayTags[dataViewTag$1] = typedArrayTags[dateTag] =
typedArrayTags[errorTag] = typedArrayTags[funcTag] =
typedArrayTags[mapTag$2] = typedArrayTags[numberTag] =
typedArrayTags[objectTag$1] = typedArrayTags[regexpTag] =
typedArrayTags[setTag$2] = typedArrayTags[stringTag] =
typedArrayTags[weakMapTag$1] = false;

/**
 * The base implementation of `_.isTypedArray` without Node.js optimizations.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
 */
function baseIsTypedArray(value) {
  return isObjectLike_1(value) &&
    isLength_1(value.length) && !!typedArrayTags[_baseGetTag(value)];
}

var _baseIsTypedArray = baseIsTypedArray;

/**
 * The base implementation of `_.unary` without support for storing metadata.
 *
 * @private
 * @param {Function} func The function to cap arguments for.
 * @returns {Function} Returns the new capped function.
 */
function baseUnary(func) {
  return function(value) {
    return func(value);
  };
}

var _baseUnary = baseUnary;

var _nodeUtil = createCommonjsModule(function (module, exports) {
/** Detect free variable `exports`. */
var freeExports = exports && !exports.nodeType && exports;

/** Detect free variable `module`. */
var freeModule = freeExports && 'object' == 'object' && module && !module.nodeType && module;

/** Detect the popular CommonJS extension `module.exports`. */
var moduleExports = freeModule && freeModule.exports === freeExports;

/** Detect free variable `process` from Node.js. */
var freeProcess = moduleExports && _freeGlobal.process;

/** Used to access faster Node.js helpers. */
var nodeUtil = (function() {
  try {
    // Use `util.types` for Node.js 10+.
    var types = freeModule && freeModule.require && freeModule.require('util').types;

    if (types) {
      return types;
    }

    // Legacy `process.binding('util')` for Node.js < 10.
    return freeProcess && freeProcess.binding && freeProcess.binding('util');
  } catch (e) {}
}());

module.exports = nodeUtil;
});

/* Node.js helper references. */
var nodeIsTypedArray = _nodeUtil && _nodeUtil.isTypedArray;

/**
 * Checks if `value` is classified as a typed array.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a typed array, else `false`.
 * @example
 *
 * _.isTypedArray(new Uint8Array);
 * // => true
 *
 * _.isTypedArray([]);
 * // => false
 */
var isTypedArray = nodeIsTypedArray ? _baseUnary(nodeIsTypedArray) : _baseIsTypedArray;

var isTypedArray_1 = isTypedArray;

/** Used for built-in method references. */
var objectProto$4 = Object.prototype;

/** Used to check objects for own properties. */
var hasOwnProperty$3 = objectProto$4.hasOwnProperty;

/**
 * Creates an array of the enumerable property names of the array-like `value`.
 *
 * @private
 * @param {*} value The value to query.
 * @param {boolean} inherited Specify returning inherited property names.
 * @returns {Array} Returns the array of property names.
 */
function arrayLikeKeys(value, inherited) {
  var isArr = isArray_1(value),
      isArg = !isArr && isArguments_1(value),
      isBuff = !isArr && !isArg && isBuffer_1(value),
      isType = !isArr && !isArg && !isBuff && isTypedArray_1(value),
      skipIndexes = isArr || isArg || isBuff || isType,
      result = skipIndexes ? _baseTimes(value.length, String) : [],
      length = result.length;

  for (var key in value) {
    if ((inherited || hasOwnProperty$3.call(value, key)) &&
        !(skipIndexes && (
           // Safari 9 has enumerable `arguments.length` in strict mode.
           key == 'length' ||
           // Node.js 0.10 has enumerable non-index properties on buffers.
           (isBuff && (key == 'offset' || key == 'parent')) ||
           // PhantomJS 2 has enumerable non-index properties on typed arrays.
           (isType && (key == 'buffer' || key == 'byteLength' || key == 'byteOffset')) ||
           // Skip index properties.
           _isIndex(key, length)
        ))) {
      result.push(key);
    }
  }
  return result;
}

var _arrayLikeKeys = arrayLikeKeys;

/** Used for built-in method references. */
var objectProto$3 = Object.prototype;

/**
 * Checks if `value` is likely a prototype object.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a prototype, else `false`.
 */
function isPrototype(value) {
  var Ctor = value && value.constructor,
      proto = (typeof Ctor == 'function' && Ctor.prototype) || objectProto$3;

  return value === proto;
}

var _isPrototype = isPrototype;

/**
 * Creates a unary function that invokes `func` with its argument transformed.
 *
 * @private
 * @param {Function} func The function to wrap.
 * @param {Function} transform The argument transform.
 * @returns {Function} Returns the new function.
 */
function overArg(func, transform) {
  return function(arg) {
    return func(transform(arg));
  };
}

var _overArg = overArg;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeKeys = _overArg(Object.keys, Object);

var _nativeKeys = nativeKeys;

/** Used for built-in method references. */
var objectProto$2 = Object.prototype;

/** Used to check objects for own properties. */
var hasOwnProperty$2 = objectProto$2.hasOwnProperty;

/**
 * The base implementation of `_.keys` which doesn't treat sparse arrays as dense.
 *
 * @private
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 */
function baseKeys(object) {
  if (!_isPrototype(object)) {
    return _nativeKeys(object);
  }
  var result = [];
  for (var key in Object(object)) {
    if (hasOwnProperty$2.call(object, key) && key != 'constructor') {
      result.push(key);
    }
  }
  return result;
}

var _baseKeys = baseKeys;

/**
 * Creates an array of the own enumerable property names of `object`.
 *
 * **Note:** Non-object values are coerced to objects. See the
 * [ES spec](http://ecma-international.org/ecma-262/7.0/#sec-object.keys)
 * for more details.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Object
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property names.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 *   this.b = 2;
 * }
 *
 * Foo.prototype.c = 3;
 *
 * _.keys(new Foo);
 * // => ['a', 'b'] (iteration order is not guaranteed)
 *
 * _.keys('hi');
 * // => ['0', '1']
 */
function keys(object) {
  return isArrayLike_1(object) ? _arrayLikeKeys(object) : _baseKeys(object);
}

var keys_1 = keys;

/**
 * Creates an array of the own enumerable string keyed property values of `object`.
 *
 * **Note:** Non-object values are coerced to objects.
 *
 * @static
 * @since 0.1.0
 * @memberOf _
 * @category Object
 * @param {Object} object The object to query.
 * @returns {Array} Returns the array of property values.
 * @example
 *
 * function Foo() {
 *   this.a = 1;
 *   this.b = 2;
 * }
 *
 * Foo.prototype.c = 3;
 *
 * _.values(new Foo);
 * // => [1, 2] (iteration order is not guaranteed)
 *
 * _.values('hi');
 * // => ['h', 'i']
 */
function values(object) {
  return object == null ? [] : _baseValues(object, keys_1(object));
}

var values_1 = values;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeMax = Math.max;

/**
 * Checks if `value` is in `collection`. If `collection` is a string, it's
 * checked for a substring of `value`, otherwise
 * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * is used for equality comparisons. If `fromIndex` is negative, it's used as
 * the offset from the end of `collection`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Collection
 * @param {Array|Object|string} collection The collection to inspect.
 * @param {*} value The value to search for.
 * @param {number} [fromIndex=0] The index to search from.
 * @param- {Object} [guard] Enables use as an iteratee for methods like `_.reduce`.
 * @returns {boolean} Returns `true` if `value` is found, else `false`.
 * @example
 *
 * _.includes([1, 2, 3], 1);
 * // => true
 *
 * _.includes([1, 2, 3], 1, 2);
 * // => false
 *
 * _.includes({ 'a': 1, 'b': 2 }, 1);
 * // => true
 *
 * _.includes('abcd', 'bc');
 * // => true
 */
function includes(collection, value, fromIndex, guard) {
  collection = isArrayLike_1(collection) ? collection : values_1(collection);
  fromIndex = (fromIndex && !guard) ? toInteger_1(fromIndex) : 0;

  var length = collection.length;
  if (fromIndex < 0) {
    fromIndex = nativeMax(length + fromIndex, 0);
  }
  return isString_1(collection)
    ? (fromIndex <= length && collection.indexOf(value, fromIndex) > -1)
    : (!!length && _baseIndexOf(collection, value, fromIndex) > -1);
}

var includes_1 = includes;

/**
 * Constants for token types
 */
var tokenTypes = {
    WHITESPACE: "whitespace",
    WORD: "word",
    STRING: "string",
    RESERVED: "reserved",
    RESERVED_TOP_LEVEL: "reserved-top-level",
    RESERVED_TOP_LEVEL_NO_INDENT: "reserved-top-level-no-indent",
    RESERVED_NEWLINE: "reserved-newline",
    OPERATOR: "operator",
    OPEN_PAREN: "open-paren",
    CLOSE_PAREN: "close-paren",
    LINE_COMMENT: "line-comment",
    BLOCK_COMMENT: "block-comment",
    NUMBER: "number",
    PLACEHOLDER: "placeholder",
};

/** Used as references for various `Number` constants. */
var MAX_SAFE_INTEGER = 9007199254740991;

/* Built-in method references for those with the same name as other `lodash` methods. */
var nativeFloor = Math.floor;

/**
 * The base implementation of `_.repeat` which doesn't coerce arguments.
 *
 * @private
 * @param {string} string The string to repeat.
 * @param {number} n The number of times to repeat the string.
 * @returns {string} Returns the repeated string.
 */
function baseRepeat(string, n) {
  var result = '';
  if (!string || n < 1 || n > MAX_SAFE_INTEGER) {
    return result;
  }
  // Leverage the exponentiation by squaring algorithm for a faster repeat.
  // See https://en.wikipedia.org/wiki/Exponentiation_by_squaring for more details.
  do {
    if (n % 2) {
      result += string;
    }
    n = nativeFloor(n / 2);
    if (n) {
      string += string;
    }
  } while (n);

  return result;
}

var _baseRepeat = baseRepeat;

/**
 * Performs a
 * [`SameValueZero`](http://ecma-international.org/ecma-262/7.0/#sec-samevaluezero)
 * comparison between two values to determine if they are equivalent.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to compare.
 * @param {*} other The other value to compare.
 * @returns {boolean} Returns `true` if the values are equivalent, else `false`.
 * @example
 *
 * var object = { 'a': 1 };
 * var other = { 'a': 1 };
 *
 * _.eq(object, object);
 * // => true
 *
 * _.eq(object, other);
 * // => false
 *
 * _.eq('a', 'a');
 * // => true
 *
 * _.eq('a', Object('a'));
 * // => false
 *
 * _.eq(NaN, NaN);
 * // => true
 */
function eq(value, other) {
  return value === other || (value !== value && other !== other);
}

var eq_1 = eq;

/**
 * Checks if the given arguments are from an iteratee call.
 *
 * @private
 * @param {*} value The potential iteratee value argument.
 * @param {*} index The potential iteratee index or key argument.
 * @param {*} object The potential iteratee object argument.
 * @returns {boolean} Returns `true` if the arguments are from an iteratee call,
 *  else `false`.
 */
function isIterateeCall(value, index, object) {
  if (!isObject_1(object)) {
    return false;
  }
  var type = typeof index;
  if (type == 'number'
        ? (isArrayLike_1(object) && _isIndex(index, object.length))
        : (type == 'string' && index in object)
      ) {
    return eq_1(object[index], value);
  }
  return false;
}

var _isIterateeCall = isIterateeCall;

/** Used as references for various `Number` constants. */
var INFINITY = 1 / 0;

/** Used to convert symbols to primitives and strings. */
var symbolProto = _Symbol ? _Symbol.prototype : undefined,
    symbolToString = symbolProto ? symbolProto.toString : undefined;

/**
 * The base implementation of `_.toString` which doesn't convert nullish
 * values to empty strings.
 *
 * @private
 * @param {*} value The value to process.
 * @returns {string} Returns the string.
 */
function baseToString(value) {
  // Exit early for strings to avoid a performance hit in some environments.
  if (typeof value == 'string') {
    return value;
  }
  if (isArray_1(value)) {
    // Recursively convert values (susceptible to call stack limits).
    return _arrayMap(value, baseToString) + '';
  }
  if (isSymbol_1(value)) {
    return symbolToString ? symbolToString.call(value) : '';
  }
  var result = (value + '');
  return (result == '0' && (1 / value) == -INFINITY) ? '-0' : result;
}

var _baseToString = baseToString;

/**
 * Converts `value` to a string. An empty string is returned for `null`
 * and `undefined` values. The sign of `-0` is preserved.
 *
 * @static
 * @memberOf _
 * @since 4.0.0
 * @category Lang
 * @param {*} value The value to convert.
 * @returns {string} Returns the converted string.
 * @example
 *
 * _.toString(null);
 * // => ''
 *
 * _.toString(-0);
 * // => '-0'
 *
 * _.toString([1, 2, 3]);
 * // => '1,2,3'
 */
function toString(value) {
  return value == null ? '' : _baseToString(value);
}

var toString_1 = toString;

/**
 * Repeats the given string `n` times.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category String
 * @param {string} [string=''] The string to repeat.
 * @param {number} [n=1] The number of times to repeat the string.
 * @param- {Object} [guard] Enables use as an iteratee for methods like `_.map`.
 * @returns {string} Returns the repeated string.
 * @example
 *
 * _.repeat('*', 3);
 * // => '***'
 *
 * _.repeat('abc', 2);
 * // => 'abcabc'
 *
 * _.repeat('abc', 0);
 * // => ''
 */
function repeat(string, n, guard) {
  if ((guard ? _isIterateeCall(string, n, guard) : n === undefined)) {
    n = 1;
  } else {
    n = toInteger_1(n);
  }
  return _baseRepeat(toString_1(string), n);
}

var repeat_1 = repeat;

/**
 * Gets the last element of `array`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Array
 * @param {Array} array The array to query.
 * @returns {*} Returns the last element of `array`.
 * @example
 *
 * _.last([1, 2, 3]);
 * // => 3
 */
function last(array) {
  var length = array == null ? 0 : array.length;
  return length ? array[length - 1] : undefined;
}

var last_1 = last;

const INDENT_TYPE_TOP_LEVEL = "top-level";
const INDENT_TYPE_BLOCK_LEVEL = "block-level";
/**
 * Manages indentation levels.
 *
 * There are two types of indentation levels:
 *
 * - BLOCK_LEVEL : increased by open-parenthesis
 * - TOP_LEVEL : increased by RESERVED_TOP_LEVEL words
 */
class Indentation {
    /**
     * @param {String} indent Indent value, default is "  " (2 spaces)
     */
    constructor(indent) {
        this.indent = indent || "  ";
        this.indentTypes = [];
    }
    /**
     * Returns current indentation string.
     * @return {String}
     */
    getIndent() {
        return repeat_1(this.indent, this.indentTypes.length);
    }
    /**
     * Increases indentation by one top-level indent.
     */
    increaseTopLevel() {
        this.indentTypes.push(INDENT_TYPE_TOP_LEVEL);
    }
    /**
     * Increases indentation by one block-level indent.
     */
    increaseBlockLevel() {
        this.indentTypes.push(INDENT_TYPE_BLOCK_LEVEL);
    }
    /**
     * Decreases indentation by one top-level indent.
     * Does nothing when the previous indent is not top-level.
     */
    decreaseTopLevel() {
        if (last_1(this.indentTypes) === INDENT_TYPE_TOP_LEVEL) {
            this.indentTypes.pop();
        }
    }
    /**
     * Decreases indentation by one block-level indent.
     * If there are top-level indents within the block-level indent,
     * throws away these as well.
     */
    decreaseBlockLevel() {
        while (this.indentTypes.length > 0) {
            const type = this.indentTypes.pop();
            if (type !== INDENT_TYPE_TOP_LEVEL) {
                break;
            }
        }
    }
    resetIndentation() {
        this.indentTypes = [];
    }
}

const INLINE_MAX_LENGTH = 50;
/**
 * Bookkeeper for inline blocks.
 *
 * Inline blocks are parenthesized expressions that are shorter than INLINE_MAX_LENGTH.
 * These blocks are formatted on a single line, unlike longer parenthesized
 * expressions where open-parenthesis causes newline and increase of indentation.
 */
class InlineBlock {
    constructor() {
        this.level = 0;
    }
    /**
     * Begins inline block when lookahead through upcoming tokens determines
     * that the block would be smaller than INLINE_MAX_LENGTH.
     * @param  {Object[]} tokens Array of all tokens
     * @param  {Number} index Current token position
     */
    beginIfPossible(tokens, index) {
        if (this.level === 0 && this.isInlineBlock(tokens, index)) {
            this.level = 1;
        }
        else if (this.level > 0) {
            this.level++;
        }
        else {
            this.level = 0;
        }
    }
    /**
     * Finishes current inline block.
     * There might be several nested ones.
     */
    end() {
        this.level--;
    }
    /**
     * True when inside an inline block
     * @return {Boolean}
     */
    isActive() {
        return this.level > 0;
    }
    // Check if this should be an inline parentheses block
    // Examples are "NOW()", "COUNT(*)", "int(10)", key(`some_column`), DECIMAL(7,2)
    isInlineBlock(tokens, index) {
        let length = 0;
        let level = 0;
        for (let i = index; i < tokens.length; i++) {
            const token = tokens[i];
            length += token.value.length;
            // Overran max length
            if (length > INLINE_MAX_LENGTH) {
                return false;
            }
            if (token.type === tokenTypes.OPEN_PAREN) {
                level++;
            }
            else if (token.type === tokenTypes.CLOSE_PAREN) {
                level--;
                if (level === 0) {
                    return true;
                }
            }
            if (this.isForbiddenToken(token)) {
                return false;
            }
        }
        return false;
    }
    // Reserved words that cause newlines, comments and semicolons
    // are not allowed inside inline parentheses block
    isForbiddenToken({ type, value }) {
        return (type === tokenTypes.RESERVED_TOP_LEVEL ||
            type === tokenTypes.RESERVED_NEWLINE ||
            type === tokenTypes.LINE_COMMENT ||
            type === tokenTypes.BLOCK_COMMENT ||
            value === ";");
    }
}

/**
 * Handles placeholder replacement with given params.
 */
class Params {
    /**
     * @param {Object} params
     */
    constructor(params) {
        this.params = params;
        this.index = 0;
    }
    /**
     * Returns param value that matches given placeholder with param key.
     * @param {Object} token
     *   @param {String} token.key Placeholder key
     *   @param {String} token.value Placeholder value
     * @return {String} param or token.value when params are missing
     */
    get({ key, value }) {
        if (!this.params) {
            return value;
        }
        if (key) {
            return this.params[key];
        }
        return this.params[this.index++];
    }
}

const trimSpacesEnd = (str) => str.replace(/[ \t]+$/u, "");
class Formatter {
    /**
     * @param {Object} cfg
     *  @param {String} cfg.language
     *  @param {String} cfg.indent
     *  @param {Bool} cfg.uppercase
     *  @param {Integer} cfg.linesBetweenQueries
     *  @param {Object} cfg.params
     * @param {Tokenizer} tokenizer
     */
    constructor(cfg, tokenizer, tokenOverride) {
        this.cfg = cfg || {};
        this.indentation = new Indentation(this.cfg.indent);
        this.inlineBlock = new InlineBlock();
        this.params = new Params(this.cfg.params);
        this.tokenizer = tokenizer;
        this.tokenOverride = tokenOverride;
        this.previousReservedWord = {};
        this.tokens = [];
        this.index = 0;
    }
    /**
     * Formats whitespace in a SQL string to make it easier to read.
     *
     * @param {String} query The SQL query string
     * @return {String} formatted query
     */
    format(query) {
        this.tokens = this.tokenizer.tokenize(query);
        const formattedQuery = this.getFormattedQueryFromTokens();
        return formattedQuery.trim();
    }
    getFormattedQueryFromTokens() {
        let formattedQuery = "";
        this.tokens.forEach((token, index) => {
            this.index = index;
            if (this.tokenOverride)
                token =
                    this.tokenOverride(token, this.previousReservedWord) ||
                        token;
            if (token.type === tokenTypes.WHITESPACE) ;
            else if (token.type === tokenTypes.LINE_COMMENT) {
                formattedQuery = this.formatLineComment(token, formattedQuery);
            }
            else if (token.type === tokenTypes.BLOCK_COMMENT) {
                formattedQuery = this.formatBlockComment(token, formattedQuery);
            }
            else if (token.type === tokenTypes.RESERVED_TOP_LEVEL) {
                formattedQuery = this.formatTopLevelReservedWord(token, formattedQuery);
                this.previousReservedWord = token;
            }
            else if (token.type === tokenTypes.RESERVED_TOP_LEVEL_NO_INDENT) {
                formattedQuery = this.formatTopLevelReservedWordNoIndent(token, formattedQuery);
                this.previousReservedWord = token;
            }
            else if (token.type === tokenTypes.RESERVED_NEWLINE) {
                formattedQuery = this.formatNewlineReservedWord(token, formattedQuery);
                this.previousReservedWord = token;
            }
            else if (token.type === tokenTypes.RESERVED) {
                formattedQuery = this.formatWithSpaces(token, formattedQuery);
                this.previousReservedWord = token;
            }
            else if (token.type === tokenTypes.OPEN_PAREN) {
                formattedQuery = this.formatOpeningParentheses(token, formattedQuery);
            }
            else if (token.type === tokenTypes.CLOSE_PAREN) {
                formattedQuery = this.formatClosingParentheses(token, formattedQuery);
            }
            else if (token.type === tokenTypes.PLACEHOLDER) {
                formattedQuery = this.formatPlaceholder(token, formattedQuery);
            }
            else if (token.value === ",") {
                formattedQuery = this.formatComma(token, formattedQuery);
            }
            else if (token.value === ":") {
                formattedQuery = this.formatWithSpaceAfter(token, formattedQuery);
            }
            else if (token.value === ".") {
                formattedQuery = this.formatWithoutSpaces(token, formattedQuery);
            }
            else if (token.value === ";") {
                formattedQuery = this.formatQuerySeparator(token, formattedQuery);
            }
            else {
                formattedQuery = this.formatWithSpaces(token, formattedQuery);
            }
        });
        return formattedQuery;
    }
    formatLineComment(token, query) {
        return this.addNewline(query + token.value);
    }
    formatBlockComment(token, query) {
        return this.addNewline(this.addNewline(query) + this.indentComment(token.value));
    }
    indentComment(comment) {
        return comment.replace(/\n[ \t]*/gu, "\n" + this.indentation.getIndent() + " ");
    }
    formatTopLevelReservedWordNoIndent(token, query) {
        this.indentation.decreaseTopLevel();
        query =
            this.addNewline(query) +
                this.equalizeWhitespace(this.formatReservedWord(token.value));
        return this.addNewline(query);
    }
    formatTopLevelReservedWord(token, query) {
        this.indentation.decreaseTopLevel();
        query = this.addNewline(query);
        this.indentation.increaseTopLevel();
        query += this.equalizeWhitespace(this.formatReservedWord(token.value));
        return this.addNewline(query);
    }
    formatNewlineReservedWord(token, query) {
        return (this.addNewline(query) +
            this.equalizeWhitespace(this.formatReservedWord(token.value)) +
            " ");
    }
    // Replace any sequence of whitespace characters with single space
    equalizeWhitespace(string) {
        return string.replace(/\s+/gu, " ");
    }
    // Opening parentheses increase the block indent level and start a new line
    formatOpeningParentheses(token, query) {
        // Take out the preceding space unless there was whitespace there in the original query
        // or another opening parens or line comment
        const preserveWhitespaceFor = [
            tokenTypes.WHITESPACE,
            tokenTypes.OPEN_PAREN,
            tokenTypes.LINE_COMMENT,
        ];
        if (!includes_1(preserveWhitespaceFor, this.previousToken().type)) {
            query = trimSpacesEnd(query);
        }
        query += this.cfg.uppercase
            ? token.value.toUpperCase()
            : token.value;
        this.inlineBlock.beginIfPossible(this.tokens, this.index);
        if (!this.inlineBlock.isActive()) {
            this.indentation.increaseBlockLevel();
            query = this.addNewline(query);
        }
        return query;
    }
    // Closing parentheses decrease the block indent level
    formatClosingParentheses(token, query) {
        token.value = this.cfg.uppercase
            ? token.value.toUpperCase()
            : token.value;
        if (this.inlineBlock.isActive()) {
            this.inlineBlock.end();
            return this.formatWithSpaceAfter(token, query);
        }
        else {
            this.indentation.decreaseBlockLevel();
            return this.formatWithSpaces(token, this.addNewline(query));
        }
    }
    formatPlaceholder(token, query) {
        return query + this.params.get(token) + " ";
    }
    // Commas start a new line (unless within inline parentheses or SQL "LIMIT" clause)
    formatComma(token, query) {
        query = trimSpacesEnd(query) + token.value + " ";
        if (this.inlineBlock.isActive()) {
            return query;
        }
        else if (/^LIMIT$/iu.test(this.previousReservedWord.value)) {
            return query;
        }
        else {
            return this.addNewline(query);
        }
    }
    formatWithSpaceAfter(token, query) {
        return trimSpacesEnd(query) + token.value + " ";
    }
    formatWithoutSpaces(token, query) {
        return trimSpacesEnd(query) + token.value;
    }
    formatWithSpaces(token, query) {
        const value = token.type === "reserved"
            ? this.formatReservedWord(token.value)
            : token.value;
        return query + value + " ";
    }
    formatReservedWord(value) {
        return this.cfg.uppercase ? value.toUpperCase() : value;
    }
    formatQuerySeparator(token, query) {
        this.indentation.resetIndentation();
        return (trimSpacesEnd(query) +
            token.value +
            "\n".repeat(this.cfg.linesBetweenQueries || 1));
    }
    addNewline(query) {
        query = trimSpacesEnd(query);
        if (!query.endsWith("\n"))
            query += "\n";
        return query + this.indentation.getIndent();
    }
    previousToken(offset = 1) {
        return this.tokens[this.index - offset] || {};
    }
}

/** Used to detect overreaching core-js shims. */
var coreJsData = _root['__core-js_shared__'];

var _coreJsData = coreJsData;

/** Used to detect methods masquerading as native. */
var maskSrcKey = (function() {
  var uid = /[^.]+$/.exec(_coreJsData && _coreJsData.keys && _coreJsData.keys.IE_PROTO || '');
  return uid ? ('Symbol(src)_1.' + uid) : '';
}());

/**
 * Checks if `func` has its source masked.
 *
 * @private
 * @param {Function} func The function to check.
 * @returns {boolean} Returns `true` if `func` is masked, else `false`.
 */
function isMasked(func) {
  return !!maskSrcKey && (maskSrcKey in func);
}

var _isMasked = isMasked;

/** Used for built-in method references. */
var funcProto$1 = Function.prototype;

/** Used to resolve the decompiled source of functions. */
var funcToString$1 = funcProto$1.toString;

/**
 * Converts `func` to its source code.
 *
 * @private
 * @param {Function} func The function to convert.
 * @returns {string} Returns the source code.
 */
function toSource(func) {
  if (func != null) {
    try {
      return funcToString$1.call(func);
    } catch (e) {}
    try {
      return (func + '');
    } catch (e) {}
  }
  return '';
}

var _toSource = toSource;

/**
 * Used to match `RegExp`
 * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
 */
var reRegExpChar$1 = /[\\^$.*+?()[\]{}|]/g;

/** Used to detect host constructors (Safari). */
var reIsHostCtor = /^\[object .+?Constructor\]$/;

/** Used for built-in method references. */
var funcProto = Function.prototype,
    objectProto$1 = Object.prototype;

/** Used to resolve the decompiled source of functions. */
var funcToString = funcProto.toString;

/** Used to check objects for own properties. */
var hasOwnProperty$1 = objectProto$1.hasOwnProperty;

/** Used to detect if a method is native. */
var reIsNative = RegExp('^' +
  funcToString.call(hasOwnProperty$1).replace(reRegExpChar$1, '\\$&')
  .replace(/hasOwnProperty|(function).*?(?=\\\()| for .+?(?=\\\])/g, '$1.*?') + '$'
);

/**
 * The base implementation of `_.isNative` without bad shim checks.
 *
 * @private
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is a native function,
 *  else `false`.
 */
function baseIsNative(value) {
  if (!isObject_1(value) || _isMasked(value)) {
    return false;
  }
  var pattern = isFunction_1(value) ? reIsNative : reIsHostCtor;
  return pattern.test(_toSource(value));
}

var _baseIsNative = baseIsNative;

/**
 * Gets the value at `key` of `object`.
 *
 * @private
 * @param {Object} [object] The object to query.
 * @param {string} key The key of the property to get.
 * @returns {*} Returns the property value.
 */
function getValue(object, key) {
  return object == null ? undefined : object[key];
}

var _getValue = getValue;

/**
 * Gets the native function at `key` of `object`.
 *
 * @private
 * @param {Object} object The object to query.
 * @param {string} key The key of the method to get.
 * @returns {*} Returns the function if it's native, else `undefined`.
 */
function getNative(object, key) {
  var value = _getValue(object, key);
  return _baseIsNative(value) ? value : undefined;
}

var _getNative = getNative;

/* Built-in method references that are verified to be native. */
var DataView = _getNative(_root, 'DataView');

var _DataView = DataView;

/* Built-in method references that are verified to be native. */
var Map$1 = _getNative(_root, 'Map');

var _Map = Map$1;

/* Built-in method references that are verified to be native. */
var Promise$1 = _getNative(_root, 'Promise');

var _Promise = Promise$1;

/* Built-in method references that are verified to be native. */
var Set = _getNative(_root, 'Set');

var _Set = Set;

/* Built-in method references that are verified to be native. */
var WeakMap$1 = _getNative(_root, 'WeakMap');

var _WeakMap = WeakMap$1;

/** `Object#toString` result references. */
var mapTag$1 = '[object Map]',
    objectTag = '[object Object]',
    promiseTag = '[object Promise]',
    setTag$1 = '[object Set]',
    weakMapTag = '[object WeakMap]';

var dataViewTag = '[object DataView]';

/** Used to detect maps, sets, and weakmaps. */
var dataViewCtorString = _toSource(_DataView),
    mapCtorString = _toSource(_Map),
    promiseCtorString = _toSource(_Promise),
    setCtorString = _toSource(_Set),
    weakMapCtorString = _toSource(_WeakMap);

/**
 * Gets the `toStringTag` of `value`.
 *
 * @private
 * @param {*} value The value to query.
 * @returns {string} Returns the `toStringTag`.
 */
var getTag = _baseGetTag;

// Fallback for data views, maps, sets, and weak maps in IE 11 and promises in Node.js < 6.
if ((_DataView && getTag(new _DataView(new ArrayBuffer(1))) != dataViewTag) ||
    (_Map && getTag(new _Map) != mapTag$1) ||
    (_Promise && getTag(_Promise.resolve()) != promiseTag) ||
    (_Set && getTag(new _Set) != setTag$1) ||
    (_WeakMap && getTag(new _WeakMap) != weakMapTag)) {
  getTag = function(value) {
    var result = _baseGetTag(value),
        Ctor = result == objectTag ? value.constructor : undefined,
        ctorString = Ctor ? _toSource(Ctor) : '';

    if (ctorString) {
      switch (ctorString) {
        case dataViewCtorString: return dataViewTag;
        case mapCtorString: return mapTag$1;
        case promiseCtorString: return promiseTag;
        case setCtorString: return setTag$1;
        case weakMapCtorString: return weakMapTag;
      }
    }
    return result;
  };
}

var _getTag = getTag;

/** `Object#toString` result references. */
var mapTag = '[object Map]',
    setTag = '[object Set]';

/** Used for built-in method references. */
var objectProto = Object.prototype;

/** Used to check objects for own properties. */
var hasOwnProperty = objectProto.hasOwnProperty;

/**
 * Checks if `value` is an empty object, collection, map, or set.
 *
 * Objects are considered empty if they have no own enumerable string keyed
 * properties.
 *
 * Array-like values such as `arguments` objects, arrays, buffers, strings, or
 * jQuery-like collections are considered empty if they have a `length` of `0`.
 * Similarly, maps and sets are considered empty if they have a `size` of `0`.
 *
 * @static
 * @memberOf _
 * @since 0.1.0
 * @category Lang
 * @param {*} value The value to check.
 * @returns {boolean} Returns `true` if `value` is empty, else `false`.
 * @example
 *
 * _.isEmpty(null);
 * // => true
 *
 * _.isEmpty(true);
 * // => true
 *
 * _.isEmpty(1);
 * // => true
 *
 * _.isEmpty([1, 2, 3]);
 * // => false
 *
 * _.isEmpty({ 'a': 1 });
 * // => false
 */
function isEmpty(value) {
  if (value == null) {
    return true;
  }
  if (isArrayLike_1(value) &&
      (isArray_1(value) || typeof value == 'string' || typeof value.splice == 'function' ||
        isBuffer_1(value) || isTypedArray_1(value) || isArguments_1(value))) {
    return !value.length;
  }
  var tag = _getTag(value);
  if (tag == mapTag || tag == setTag) {
    return !value.size;
  }
  if (_isPrototype(value)) {
    return !_baseKeys(value).length;
  }
  for (var key in value) {
    if (hasOwnProperty.call(value, key)) {
      return false;
    }
  }
  return true;
}

var isEmpty_1 = isEmpty;

/**
 * Used to match `RegExp`
 * [syntax characters](http://ecma-international.org/ecma-262/7.0/#sec-patterns).
 */
var reRegExpChar = /[\\^$.*+?()[\]{}|]/g,
    reHasRegExpChar = RegExp(reRegExpChar.source);

/**
 * Escapes the `RegExp` special characters "^", "$", "\", ".", "*", "+",
 * "?", "(", ")", "[", "]", "{", "}", and "|" in `string`.
 *
 * @static
 * @memberOf _
 * @since 3.0.0
 * @category String
 * @param {string} [string=''] The string to escape.
 * @returns {string} Returns the escaped string.
 * @example
 *
 * _.escapeRegExp('[lodash](https://lodash.com/)');
 * // => '\[lodash\]\(https://lodash\.com/\)'
 */
function escapeRegExp(string) {
  string = toString_1(string);
  return (string && reHasRegExpChar.test(string))
    ? string.replace(reRegExpChar, '\\$&')
    : string;
}

var escapeRegExp_1 = escapeRegExp;

class Tokenizer {
    /**
     * @param {Object} cfg
     *  @param {String[]} cfg.reservedWords Reserved words in SQL
     *  @param {String[]} cfg.reservedTopLevelWords Words that are set to new line separately
     *  @param {String[]} cfg.reservedNewlineWords Words that are set to newline
     *  @param {String[]} cfg.reservedTopLevelWordsNoIndent Words that are top level but have no indentation
     *  @param {String[]} cfg.stringTypes String types to enable: "", '', ``, [], N''
     *  @param {String[]} cfg.openParens Opening parentheses to enable, like (, [
     *  @param {String[]} cfg.closeParens Closing parentheses to enable, like ), ]
     *  @param {String[]} cfg.indexedPlaceholderTypes Prefixes for indexed placeholders, like ?
     *  @param {String[]} cfg.namedPlaceholderTypes Prefixes for named placeholders, like @ and :
     *  @param {String[]} cfg.lineCommentTypes Line comments to enable, like # and --
     *  @param {String[]} cfg.specialWordChars Special chars that can be found inside of words, like @ and #
     */
    constructor(cfg) {
        this.WHITESPACE_REGEX = /^(\s+)/u;
        this.NUMBER_REGEX =
            /^((-\s*)?[0-9]+(\.[0-9]+)?|0x[0-9a-fA-F]+|0b[01]+)\b/u;
        this.OPERATOR_REGEX =
            /^(!=|<>|==|<=|>=|!<|!>|\|\||::|->>|->|~~\*|~~|!~~\*|!~~|~\*|!~\*|!~|:=|.)/u;
        this.BLOCK_COMMENT_REGEX = /^(\/\*[^]*?(?:\*\/|$))/u;
        this.LINE_COMMENT_REGEX = this.createLineCommentRegex(cfg.lineCommentTypes);
        this.RESERVED_TOP_LEVEL_REGEX = this.createReservedWordRegex(cfg.reservedTopLevelWords);
        this.RESERVED_TOP_LEVEL_NO_INDENT_REGEX =
            this.createReservedWordRegex(cfg.reservedTopLevelWordsNoIndent);
        this.RESERVED_NEWLINE_REGEX = this.createReservedWordRegex(cfg.reservedNewlineWords);
        this.RESERVED_PLAIN_REGEX = this.createReservedWordRegex(cfg.reservedWords);
        this.WORD_REGEX = this.createWordRegex(cfg.specialWordChars);
        this.STRING_REGEX = this.createStringRegex(cfg.stringTypes);
        this.OPEN_PAREN_REGEX = this.createParenRegex(cfg.openParens);
        this.CLOSE_PAREN_REGEX = this.createParenRegex(cfg.closeParens);
        this.INDEXED_PLACEHOLDER_REGEX = this.createPlaceholderRegex(cfg.indexedPlaceholderTypes, "[0-9]*");
        this.IDENT_NAMED_PLACEHOLDER_REGEX = this.createPlaceholderRegex(cfg.namedPlaceholderTypes, "[a-zA-Z0-9._$]+");
        this.STRING_NAMED_PLACEHOLDER_REGEX = this.createPlaceholderRegex(cfg.namedPlaceholderTypes, this.createStringPattern(cfg.stringTypes));
    }
    createLineCommentRegex(lineCommentTypes) {
        return new RegExp(`^((?:${lineCommentTypes
            .map((c) => escapeRegExp_1(c))
            .join("|")}).*?(?:\r\n|\r|\n|$))`, "u");
    }
    createReservedWordRegex(reservedWords) {
        const reservedWordsPattern = reservedWords
            .join("|")
            .replace(/ /gu, "\\s+");
        return new RegExp(`^(${reservedWordsPattern})\\b`, "iu");
    }
    createWordRegex(specialChars = []) {
        return new RegExp(`^([\\p{Alphabetic}\\p{Mark}\\p{Decimal_Number}\\p{Connector_Punctuation}\\p{Join_Control}${specialChars.join("")}]+)`, "u");
    }
    createStringRegex(stringTypes) {
        return new RegExp("^(" + this.createStringPattern(stringTypes) + ")", "u");
    }
    // This enables the following string patterns:
    // 1. backtick quoted string using `` to escape
    // 2. square bracket quoted string (SQL Server) using ]] to escape
    // 3. double quoted string using "" or \" to escape
    // 4. single quoted string using '' or \' to escape
    // 5. national character quoted string using N'' or N\' to escape
    createStringPattern(stringTypes) {
        const patterns = {
            "``": "((`[^`]*($|`))+)",
            "[]": "((\\[[^\\]]*($|\\]))(\\][^\\]]*($|\\]))*)",
            '""': '(("[^"\\\\]*(?:\\\\.[^"\\\\]*)*("|$))+)',
            "''": "(('[^'\\\\]*(?:\\\\.[^'\\\\]*)*('|$))+)",
            "N''": "((N'[^N'\\\\]*(?:\\\\.[^N'\\\\]*)*('|$))+)",
        };
        return stringTypes.map((t) => patterns[t]).join("|");
    }
    createParenRegex(parens) {
        return new RegExp("^(" + parens.map((p) => this.escapeParen(p)).join("|") + ")", "iu");
    }
    escapeParen(paren) {
        if (paren.length === 1) {
            // A single punctuation character
            return escapeRegExp_1(paren);
        }
        else {
            // longer word
            return "\\b" + paren + "\\b";
        }
    }
    createPlaceholderRegex(types, pattern) {
        if (isEmpty_1(types)) {
            return new RegExp(".*");
        }
        const typesRegex = types.map(escapeRegExp_1).join("|");
        return new RegExp(`^((?:${typesRegex})(?:${pattern}))`, "u");
    }
    /**
     * Takes a SQL string and breaks it into tokens.
     * Each token is an object with type and value.
     *
     * @param {String} input The SQL string
     * @return {Object[]} tokens An array of tokens.
     *  @return {String} token.type
     *  @return {String} token.value
     */
    tokenize(input) {
        if (!input)
            return [];
        const tokens = [];
        let token;
        // Keep processing the string until it is empty
        while (input.length) {
            // Get the next token and the token type
            token = this.getNextToken(input, token);
            // Advance the string
            input = input.substring(token.value.length);
            tokens.push(token);
        }
        return tokens;
    }
    getNextToken(input, previousToken) {
        return (this.getWhitespaceToken(input) ||
            this.getCommentToken(input) ||
            this.getStringToken(input) ||
            this.getOpenParenToken(input) ||
            this.getCloseParenToken(input) ||
            this.getPlaceholderToken(input) ||
            this.getNumberToken(input) ||
            this.getReservedWordToken(input, previousToken) ||
            this.getWordToken(input) ||
            this.getOperatorToken(input));
    }
    getWhitespaceToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.WHITESPACE,
            regex: this.WHITESPACE_REGEX,
        });
    }
    getCommentToken(input) {
        return (this.getLineCommentToken(input) ||
            this.getBlockCommentToken(input));
    }
    getLineCommentToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.LINE_COMMENT,
            regex: this.LINE_COMMENT_REGEX,
        });
    }
    getBlockCommentToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.BLOCK_COMMENT,
            regex: this.BLOCK_COMMENT_REGEX,
        });
    }
    getStringToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.STRING,
            regex: this.STRING_REGEX,
        });
    }
    getOpenParenToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.OPEN_PAREN,
            regex: this.OPEN_PAREN_REGEX,
        });
    }
    getCloseParenToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.CLOSE_PAREN,
            regex: this.CLOSE_PAREN_REGEX,
        });
    }
    getPlaceholderToken(input) {
        return (this.getIdentNamedPlaceholderToken(input) ||
            this.getStringNamedPlaceholderToken(input) ||
            this.getIndexedPlaceholderToken(input));
    }
    getIdentNamedPlaceholderToken(input) {
        return this.getPlaceholderTokenWithKey({
            input,
            regex: this.IDENT_NAMED_PLACEHOLDER_REGEX,
            parseKey: (v) => v.slice(1),
        });
    }
    getStringNamedPlaceholderToken(input) {
        return this.getPlaceholderTokenWithKey({
            input,
            regex: this.STRING_NAMED_PLACEHOLDER_REGEX,
            parseKey: (v) => this.getEscapedPlaceholderKey({
                key: v.slice(2, -1),
                quoteChar: v.slice(-1),
            }),
        });
    }
    getIndexedPlaceholderToken(input) {
        return this.getPlaceholderTokenWithKey({
            input,
            regex: this.INDEXED_PLACEHOLDER_REGEX,
            parseKey: (v) => v.slice(1),
        });
    }
    getPlaceholderTokenWithKey({ input, regex, parseKey, }) {
        const token = this.getTokenOnFirstMatch({
            input,
            regex,
            type: tokenTypes.PLACEHOLDER,
        });
        if (token) {
            token.key = parseKey(token.value);
        }
        return token;
    }
    getEscapedPlaceholderKey({ key, quoteChar, }) {
        return key.replace(new RegExp(escapeRegExp_1("\\" + quoteChar), "gu"), quoteChar);
    }
    // Decimal, binary, or hex numbers
    getNumberToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.NUMBER,
            regex: this.NUMBER_REGEX,
        });
    }
    // Punctuation and symbols
    getOperatorToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.OPERATOR,
            regex: this.OPERATOR_REGEX,
        });
    }
    getReservedWordToken(input, previousToken) {
        // A reserved word cannot be preceded by a "."
        // this makes it so in "my_table.from", "from" is not considered a reserved word
        if (previousToken &&
            previousToken.value &&
            previousToken.value === ".") {
            return;
        }
        return (this.getTopLevelReservedToken(input) ||
            this.getNewlineReservedToken(input) ||
            this.getTopLevelReservedTokenNoIndent(input) ||
            this.getPlainReservedToken(input));
    }
    getTopLevelReservedToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.RESERVED_TOP_LEVEL,
            regex: this.RESERVED_TOP_LEVEL_REGEX,
        });
    }
    getNewlineReservedToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.RESERVED_NEWLINE,
            regex: this.RESERVED_NEWLINE_REGEX,
        });
    }
    getTopLevelReservedTokenNoIndent(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.RESERVED_TOP_LEVEL_NO_INDENT,
            regex: this.RESERVED_TOP_LEVEL_NO_INDENT_REGEX,
        });
    }
    getPlainReservedToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.RESERVED,
            regex: this.RESERVED_PLAIN_REGEX,
        });
    }
    getWordToken(input) {
        return this.getTokenOnFirstMatch({
            input,
            type: tokenTypes.WORD,
            regex: this.WORD_REGEX,
        });
    }
    getTokenOnFirstMatch({ input, type, regex, }) {
        /// NOTE replace `match` with `exec`
        const matches = regex.exec(input);
        if (matches) {
            return { type, value: matches[1] };
        }
    }
}

const reservedWords = [
    "ACCESSIBLE",
    "ACTION",
    "AGAINST",
    "AGGREGATE",
    "ALGORITHM",
    "ALL",
    "ALTER",
    "ANALYSE",
    "ANALYZE",
    "AS",
    "ASC",
    "AUTOCOMMIT",
    "AUTO_INCREMENT",
    "BACKUP",
    "BEGIN",
    "BETWEEN",
    "BINLOG",
    "BOTH",
    "CASCADE",
    "CASE",
    "CHANGE",
    "CHANGED",
    "CHARACTER SET",
    "CHARSET",
    "CHECK",
    "CHECKSUM",
    "COLLATE",
    "COLLATION",
    "COLUMN",
    "COLUMNS",
    "COMMENT",
    "COMMIT",
    "COMMITTED",
    "COMPRESSED",
    "CONCURRENT",
    "CONSTRAINT",
    "CONTAINS",
    "CONVERT",
    "CREATE",
    "CROSS",
    "CURRENT_TIMESTAMP",
    "DATABASE",
    "DATABASES",
    "DAY",
    "DAY_HOUR",
    "DAY_MINUTE",
    "DAY_SECOND",
    "DEFAULT",
    "DEFINER",
    "DELAYED",
    "DELETE",
    "DESC",
    "DESCRIBE",
    "DETERMINISTIC",
    "DISTINCT",
    "DISTINCTROW",
    "DIV",
    "DO",
    "DROP",
    "DUMPFILE",
    "DUPLICATE",
    "DYNAMIC",
    "ELSE",
    "ENCLOSED",
    "END",
    "ENGINE",
    "ENGINES",
    "ENGINE_TYPE",
    "ESCAPE",
    "ESCAPED",
    "EVENTS",
    "EXEC",
    "EXECUTE",
    "EXISTS",
    "EXPLAIN",
    "EXTENDED",
    "FAST",
    "FETCH",
    "FIELDS",
    "FILE",
    "FIRST",
    "FIXED",
    "FLUSH",
    "FOR",
    "FORCE",
    "FOREIGN",
    "FULL",
    "FULLTEXT",
    "FUNCTION",
    "GLOBAL",
    "GRANT",
    "GRANTS",
    "GROUP_CONCAT",
    "HEAP",
    "HIGH_PRIORITY",
    "HOSTS",
    "HOUR",
    "HOUR_MINUTE",
    "HOUR_SECOND",
    "IDENTIFIED",
    "IF",
    "IFNULL",
    "IGNORE",
    "IN",
    "INDEX",
    "INDEXES",
    "INFILE",
    "INSERT",
    "INSERT_ID",
    "INSERT_METHOD",
    "INTERVAL",
    "INTO",
    "INVOKER",
    "IS",
    "ISOLATION",
    "KEY",
    "KEYS",
    "KILL",
    "LAST_INSERT_ID",
    "LEADING",
    "LEVEL",
    "LIKE",
    "LINEAR",
    "LINES",
    "LOAD",
    "LOCAL",
    "LOCK",
    "LOCKS",
    "LOGS",
    "LOW_PRIORITY",
    "MARIA",
    "MASTER",
    "MASTER_CONNECT_RETRY",
    "MASTER_HOST",
    "MASTER_LOG_FILE",
    "MATCH",
    "MAX_CONNECTIONS_PER_HOUR",
    "MAX_QUERIES_PER_HOUR",
    "MAX_ROWS",
    "MAX_UPDATES_PER_HOUR",
    "MAX_USER_CONNECTIONS",
    "MEDIUM",
    "MERGE",
    "MINUTE",
    "MINUTE_SECOND",
    "MIN_ROWS",
    "MODE",
    "MODIFY",
    "MONTH",
    "MRG_MYISAM",
    "MYISAM",
    "NAMES",
    "NATURAL",
    "NOT",
    "NOW()",
    "NULL",
    "OFFSET",
    "ON DELETE",
    "ON UPDATE",
    "ON",
    "ONLY",
    "OPEN",
    "OPTIMIZE",
    "OPTION",
    "OPTIONALLY",
    "OUTFILE",
    "PACK_KEYS",
    "PAGE",
    "PARTIAL",
    "PARTITION",
    "PARTITIONS",
    "PASSWORD",
    "PRIMARY",
    "PRIVILEGES",
    "PROCEDURE",
    "PROCESS",
    "PROCESSLIST",
    "PURGE",
    "QUICK",
    "RAID0",
    "RAID_CHUNKS",
    "RAID_CHUNKSIZE",
    "RAID_TYPE",
    "RANGE",
    "READ",
    "READ_ONLY",
    "READ_WRITE",
    "REFERENCES",
    "REGEXP",
    "RELOAD",
    "RENAME",
    "REPAIR",
    "REPEATABLE",
    "REPLACE",
    "REPLICATION",
    "RESET",
    "RESTORE",
    "RESTRICT",
    "RETURN",
    "RETURNS",
    "REVOKE",
    "RLIKE",
    "ROLLBACK",
    "ROW",
    "ROWS",
    "ROW_FORMAT",
    "SECOND",
    "SECURITY",
    "SEPARATOR",
    "SERIALIZABLE",
    "SESSION",
    "SHARE",
    "SHOW",
    "SHUTDOWN",
    "SLAVE",
    "SONAME",
    "SOUNDS",
    "SQL",
    "SQL_AUTO_IS_NULL",
    "SQL_BIG_RESULT",
    "SQL_BIG_SELECTS",
    "SQL_BIG_TABLES",
    "SQL_BUFFER_RESULT",
    "SQL_CACHE",
    "SQL_CALC_FOUND_ROWS",
    "SQL_LOG_BIN",
    "SQL_LOG_OFF",
    "SQL_LOG_UPDATE",
    "SQL_LOW_PRIORITY_UPDATES",
    "SQL_MAX_JOIN_SIZE",
    "SQL_NO_CACHE",
    "SQL_QUOTE_SHOW_CREATE",
    "SQL_SAFE_UPDATES",
    "SQL_SELECT_LIMIT",
    "SQL_SLAVE_SKIP_COUNTER",
    "SQL_SMALL_RESULT",
    "SQL_WARNINGS",
    "START",
    "STARTING",
    "STATUS",
    "STOP",
    "STORAGE",
    "STRAIGHT_JOIN",
    "STRING",
    "STRIPED",
    "SUPER",
    "TABLE",
    "TABLES",
    "TEMPORARY",
    "TERMINATED",
    "THEN",
    "TO",
    "TRAILING",
    "TRANSACTIONAL",
    "TRUE",
    "TRUNCATE",
    "TYPE",
    "TYPES",
    "UNCOMMITTED",
    "UNIQUE",
    "UNLOCK",
    "UNSIGNED",
    "USAGE",
    "USE",
    "USING",
    "VARIABLES",
    "VIEW",
    "WHEN",
    "WITH",
    "WORK",
    "WRITE",
    "YEAR_MONTH",
];
const reservedTopLevelWords = [
    "ADD",
    "AFTER",
    "ALTER COLUMN",
    "ALTER TABLE",
    "DELETE FROM",
    "EXCEPT",
    "FETCH FIRST",
    "FROM",
    "GROUP BY",
    "GO",
    "HAVING",
    "INSERT INTO",
    "INSERT",
    "LIMIT",
    "MODIFY",
    "ORDER BY",
    "SELECT",
    "SET CURRENT SCHEMA",
    "SET SCHEMA",
    "SET",
    "UPDATE",
    "VALUES",
    "WHERE",
];
const reservedTopLevelWordsNoIndent = [
    "INTERSECT",
    "INTERSECT ALL",
    "MINUS",
    "UNION",
    "UNION ALL",
];
const reservedNewlineWords = [
    "AND",
    "CROSS APPLY",
    "CROSS JOIN",
    "ELSE",
    "INNER JOIN",
    "JOIN",
    "LEFT JOIN",
    "LEFT OUTER JOIN",
    "OR",
    "OUTER APPLY",
    "OUTER JOIN",
    "RIGHT JOIN",
    "RIGHT OUTER JOIN",
    "WHEN",
    "XOR",
];
let tokenizer;
class StandardSqlFormatter {
    /**
     * @param {SqlFormatConfig} cfg Different set of configurations
     */
    constructor(cfg) {
        this.cfg = cfg;
    }
    /**
     * Format the whitespace in a Standard SQL string to make it easier to read
     *
     * @param {String} query The Standard SQL string
     * @return {String} formatted string
     */
    format(query) {
        if (!tokenizer) {
            tokenizer = new Tokenizer({
                reservedWords,
                reservedTopLevelWords,
                reservedNewlineWords,
                reservedTopLevelWordsNoIndent,
                stringTypes: [`""`, "N''", "''", "``", "[]"],
                openParens: ["(", "CASE"],
                closeParens: [")", "END"],
                indexedPlaceholderTypes: ["?"],
                namedPlaceholderTypes: ["@", ":"],
                lineCommentTypes: ["#", "--"],
            });
        }
        return new Formatter(this.cfg, tokenizer).format(query);
    }
}

const defaultFormatter = new StandardSqlFormatter({
    uppercase: true,
});
function sql(config = {}) {
    const configs = Object.assign({}, config);
    if (!configs.schema) {
        configs.schema = {};
    }
    const lang = configs.dialect || langSql.StandardSQL;
    return new language.LanguageSupport(lang.language, [
        config.schema
            ? (config.dialect || langSql.StandardSQL).language.data.of({
                autocomplete: completeFromSchema(config.schema, config.tables, config.defaultTable),
            })
            : [],
        langSql.keywordCompletion(lang, !!configs.upperCaseKeywords),
        analysisSql(),
        view.keymap.of([
            {
                key: "Cmd-Shift-f",
                run: (view) => {
                    const query = view.state.doc
                        .toJSON()
                        .join(view.state.lineBreak);
                    void Promise.resolve(defaultFormatter.format(query)).then((formattedDoc) => {
                        return view.dispatch({
                            changes: [
                                {
                                    from: 0,
                                    to: view.state.doc.length,
                                    insert: formattedDoc,
                                },
                            ],
                            effects: [setPromptText.of("Cmd-Shift-f")],
                        });
                    });
                    return true;
                },
            },
        ]),
    ]);
}

// This file was generated by lezer-generator. You probably shouldn't edit it.
const parser = Parser.deserialize({
  version: 13,
  states:
    "!WQYQPOOOhQPO'#CdOOQO'#Ci'#CiOOQO'#Ce'#CeQYQPOOOOQO,59O,59OOyQPO,59OOOQO-E6c-E6cOOQO1G.j1G.j",
  stateData:
    "![~O[OSPOS~ORQOSQOTQOVPO~ORQOSQOTQOUTOVPO~ORQOSQOTQOUWOVPO~O",
  goto: "u^PPPPPPPP_ePPPoXQOPSUQSOQUPTVSUXROPSU",
  nodeNames:
    "⚠ LineComment Program Identifier String Boolean ) ( Application",
  maxTerm: 13,
  nodeProps: [
    [NodeProp.openedBy, 6, "("],
    [NodeProp.closedBy, 7, ")"],
  ],
  skippedNodes: [0, 1],
  repeatNodeCount: 1,
  tokenData:
    "#y~RYXYqYZq]^qpqqrs!Sst!qxy#Pyz#U!Q![#Z!]!^#n~vS[~XYqYZq]^qpqq~!VTOr!Srs!fs#O!S#O#P!k#P~!S~!kOS~~!nPO~!S~!tQ#Y#Z!z#h#i!z~#POT~~#UOV~~#ZOU~~#`RR~}!O#i!Q![#Z![!]#i~#nOR~~#sQP~OY#nZ~#n",
  tokenizers: [0],
  topRules: { Program: [0, 2] },
  tokenPrec: 0,
});

const LogLanguage = language.LezerLanguage.define({
  parser: parser.configure({
    props: [
      language.indentNodeProp.add({
        Application: language.delimitedIndent({ closing: ")", align: false }),
      }),
      language.foldNodeProp.add({
        Application: language.foldInside,
      }),
      highlight.styleTags({
        Identifier: highlight.tags.variableName,
        String: highlight.tags.string,
      }),
    ],
  }),
  languageData: {
    commentTokens: { line: ";" },
  },
});
function log() {
  return new language.LanguageSupport(LogLanguage)
}

/// language langOptions
const LANGUAGES = {
    javascript: langJavascript.javascript,
    sql,
    json: langJson.json,
    python: langPython.python,
    log,
};
const languageCompart = new state.Compartment();
function reconfigure$2(language, langOptions) {
    if (!LANGUAGES[language]) {
        return undefined;
    }
    return languageCompart.reconfigure(LANGUAGES[language](
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    //@ts-ignore
    langOptions));
}
function useLanguageProp(language, langOptions, view) {
    const prevLanguage = React.useRef({
        language,
        langOptions,
    });
    const [compart] = React.useState(() => {
        var _a;
        if (!LANGUAGES[language]) {
            return undefined;
        }
        return languageCompart.of((_a = LANGUAGES[language]) === null || _a === void 0 ? void 0 : _a.call(LANGUAGES, 
        /// 有些language方法并不需要参数，所以这里会有TS的一场类型错误
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore
        langOptions));
    });
    React.useEffect(() => {
        if (view &&
            (prevLanguage.current.language !== language ||
                !dequal.dequal(langOptions, prevLanguage.current.langOptions))) {
            view.dispatch({
                effects: [reconfigure$2(language, langOptions)],
            });
        }
    }, [language, view, langOptions]);
    return compart;
}

// Using https://github.com/one-dark/vscode-one-dark-theme/ as reference for the colors
const chalky$1 = "#e5c07b", coral$1 = "#e06c75", cyan$1 = "#56b6c2", invalid$1 = "#ffffff", ivory$1 = "#abb2bf", stone$1 = "#7d8799", // Brightened compared to original to increase contrast
malibu$1 = "#61afef", sage$1 = "#98c379", whiskey$1 = "#d19a66", violet$1 = "#c678dd", darkBackground$1 = "#21252b", highlightBackground$1 = "#2c313a", background$1 = "#282c34", selection$1 = "#3E4451", cursor$1 = "#528bff";
/// The editor theme styles for One Dark.
const oneDarkTheme$1 = view.EditorView.theme({
    "&": {
        color: ivory$1,
        backgroundColor: background$1,
        "& ::selection": { backgroundColor: selection$1 },
        caretColor: cursor$1,
        height: "100%",
    },
    "&.cm-focused .cm-cursor": { borderLeftColor: cursor$1 },
    "&.cm-focused .cm-selectionBackground, .cm-selectionBackground": {
        backgroundColor: selection$1,
    },
    ".cm-panels": { backgroundColor: darkBackground$1, color: ivory$1 },
    ".cm-panels.cm-panels-top": { borderBottom: "2px solid black" },
    ".cm-panels.cm-panels-bottom": { borderTop: "2px solid black" },
    ".cm-searchMatch": {
        backgroundColor: "#72a1ff59",
        outline: "1px solid #457dff",
    },
    ".cm-searchMatch.cm-searchMatch-selected": {
        backgroundColor: "#6199ff2f",
    },
    ".cm-activeLine": { backgroundColor: highlightBackground$1 },
    ".cm-selectionMatch": { backgroundColor: "#aafe661a" },
    ".cm-matchingBracket, .cm-nonmatchingBracket": {
        backgroundColor: "#bad0f847",
        outline: "1px solid #515a6b",
    },
    ".cm-gutters": {
        backgroundColor: background$1,
        color: stone$1,
        border: "none",
    },
    ".cm-lineNumbers .cm-gutterElement": { color: "inherit" },
    ".cm-foldPlaceholder": {
        backgroundColor: "transparent",
        border: "none",
        color: "#ddd",
    },
    ".cm-tooltip": {
        border: "1px solid #181a1f",
        backgroundColor: darkBackground$1,
    },
    ".cm-tooltip-autocomplete": {
        "& > ul > li[aria-selected]": {
            backgroundColor: highlightBackground$1,
            color: ivory$1,
        },
        "& > ul > li > cm-completionIcon": {
            paddingRight: "2em",
        },
    },
}, { dark: true });
/// The highlighting style for code in the One Dark theme.
const oneDarkHighlightStyle$1 = highlight.HighlightStyle.define([
    { tag: highlight.tags.keyword, color: violet$1 },
    {
        tag: [
            highlight.tags.name,
            highlight.tags.deleted,
            highlight.tags.character,
            highlight.tags.propertyName,
            highlight.tags.macroName,
        ],
        color: coral$1,
    },
    {
        tag: [highlight.tags.function(highlight.tags.variableName), highlight.tags.labelName],
        color: malibu$1,
    },
    {
        tag: [
            highlight.tags.color,
            highlight.tags.constant(highlight.tags.name),
            highlight.tags.standard(highlight.tags.name),
        ],
        color: whiskey$1,
    },
    { tag: [highlight.tags.definition(highlight.tags.name), highlight.tags.separator], color: ivory$1 },
    {
        tag: [
            highlight.tags.typeName,
            highlight.tags.className,
            highlight.tags.number,
            highlight.tags.changed,
            highlight.tags.annotation,
            highlight.tags.modifier,
            highlight.tags.self,
            highlight.tags.namespace,
        ],
        color: chalky$1,
    },
    {
        tag: [
            highlight.tags.operator,
            highlight.tags.operatorKeyword,
            highlight.tags.url,
            highlight.tags.escape,
            highlight.tags.regexp,
            highlight.tags.link,
            highlight.tags.special(highlight.tags.string),
        ],
        color: cyan$1,
    },
    { tag: [highlight.tags.meta, highlight.tags.comment], color: stone$1 },
    { tag: highlight.tags.strong, fontWeight: "bold" },
    { tag: highlight.tags.emphasis, fontStyle: "italic" },
    { tag: highlight.tags.link, color: stone$1, textDecoration: "underline" },
    { tag: highlight.tags.heading, fontWeight: "bold", color: coral$1 },
    {
        tag: [highlight.tags.atom, highlight.tags.bool, highlight.tags.special(highlight.tags.variableName)],
        color: whiskey$1,
    },
    {
        tag: [highlight.tags.processingInstruction, highlight.tags.string, highlight.tags.inserted],
        color: sage$1,
    },
    { tag: highlight.tags.invalid, color: invalid$1 },
]);
/// Extension to enable the One Dark theme (both the editor theme and
/// the highlight style).
const oneDark$1 = [oneDarkTheme$1, oneDarkHighlightStyle$1];

// Using https://github.com/one-dark/vscode-one-dark-theme/ as reference for the colors
const chalky = "#e5c07b", coral = "#e06c75", cyan = "#56b6c2", invalid = "#ffffff", ivory = "#abb2bf", stone = "#7d8799", // Brightened compared to original to increase contrast
malibu = "#61afef", sage = "#98c379", whiskey = "#d19a66", violet = "#c678dd", darkBackground = "#21252b", highlightBackground = "#e2e2e2", background = "#ffffff", selection = "#3E4451", cursor = "#528bff";
/// The editor theme styles for One Dark.
const oneDarkTheme = view.EditorView.theme({
    "&": {
        color: ivory,
        backgroundColor: background,
        "& ::selection": { backgroundColor: selection },
        caretColor: cursor,
        height: '100%'
    },
    "&.cm-focused .cm-cursor": { borderLeftColor: cursor },
    "&.cm-focused .cm-selectionBackground, .cm-selectionBackground": {
        backgroundColor: selection,
    },
    ".cm-panels": { backgroundColor: darkBackground, color: ivory },
    ".cm-panels.cm-panels-top": { borderBottom: "2px solid black" },
    ".cm-panels.cm-panels-bottom": { borderTop: "2px solid black" },
    ".cm-searchMatch": {
        backgroundColor: "#72a1ff59",
        outline: "1px solid #457dff",
    },
    ".cm-searchMatch.cm-searchMatch-selected": {
        backgroundColor: "#6199ff2f",
    },
    ".cm-activeLine": { backgroundColor: highlightBackground },
    ".cm-selectionMatch": { backgroundColor: "#aafe661a" },
    ".cm-matchingBracket, .cm-nonmatchingBracket": {
        backgroundColor: "#bad0f847",
        outline: "1px solid #515a6b",
    },
    ".cm-gutters": {
        backgroundColor: background,
        color: stone,
        border: "none",
    },
    ".cm-lineNumbers .cm-gutterElement": { color: "inherit" },
    ".cm-foldPlaceholder": {
        backgroundColor: "transparent",
        border: "none",
        color: "#ddd",
    },
    ".cm-tooltip": {
        border: "1px solid #181a1f",
        backgroundColor: darkBackground,
    },
    ".cm-tooltip-autocomplete": {
        "& > ul > li[aria-selected]": {
            backgroundColor: highlightBackground,
            color: ivory,
        },
        "& > ul > li > cm-completionIcon": {
            paddingRight: "2em",
        },
    },
}, { dark: true });
/// The highlighting style for code in the One Dark theme.
const oneDarkHighlightStyle = highlight.HighlightStyle.define([
    { tag: highlight.tags.keyword, color: violet },
    {
        tag: [
            highlight.tags.name,
            highlight.tags.deleted,
            highlight.tags.character,
            highlight.tags.propertyName,
            highlight.tags.macroName,
        ],
        color: coral,
    },
    {
        tag: [highlight.tags.function(highlight.tags.variableName), highlight.tags.labelName],
        color: malibu,
    },
    {
        tag: [
            highlight.tags.color,
            highlight.tags.constant(highlight.tags.name),
            highlight.tags.standard(highlight.tags.name),
        ],
        color: whiskey,
    },
    { tag: [highlight.tags.definition(highlight.tags.name), highlight.tags.separator], color: ivory },
    {
        tag: [
            highlight.tags.typeName,
            highlight.tags.className,
            highlight.tags.number,
            highlight.tags.changed,
            highlight.tags.annotation,
            highlight.tags.modifier,
            highlight.tags.self,
            highlight.tags.namespace,
        ],
        color: chalky,
    },
    {
        tag: [
            highlight.tags.operator,
            highlight.tags.operatorKeyword,
            highlight.tags.url,
            highlight.tags.escape,
            highlight.tags.regexp,
            highlight.tags.link,
            highlight.tags.special(highlight.tags.string),
        ],
        color: cyan,
    },
    { tag: [highlight.tags.meta, highlight.tags.comment], color: stone },
    { tag: highlight.tags.strong, fontWeight: "bold" },
    { tag: highlight.tags.emphasis, fontStyle: "italic" },
    { tag: highlight.tags.link, color: stone, textDecoration: "underline" },
    { tag: highlight.tags.heading, fontWeight: "bold", color: coral },
    {
        tag: [highlight.tags.atom, highlight.tags.bool, highlight.tags.special(highlight.tags.variableName)],
        color: whiskey,
    },
    {
        tag: [highlight.tags.processingInstruction, highlight.tags.string, highlight.tags.inserted],
        color: sage,
    },
    { tag: highlight.tags.invalid, color: invalid },
]);
/// Extension to enable the One Dark theme (both the editor theme and
/// the highlight style).
const oneDark = [oneDarkTheme, oneDarkHighlightStyle];

/// theme
const themeCompart = new state.Compartment();
const THEMES = {
    dark: oneDark$1,
    light: oneDark,
};
function reconfigure$1(theme) {
    return themeCompart.reconfigure(THEMES[theme]);
}
function useThemeProp(theme, view) {
    const [compart] = React.useState(() => themeCompart.of(THEMES[theme]));
    React.useEffect(() => {
        view === null || view === void 0 ? void 0 : view.dispatch({
            effects: [reconfigure$1(theme)],
        });
    }, [theme, view]);
    return compart;
}

function useDiagnostics(diagnostics, view) {
    React.useEffect(() => {
        if (view) {
            const founds = translateDiagnostics(diagnostics, view);
            const first = founds[0];
            // scroll into this first diagnostric result
            if (first) {
                view.scrollPosIntoView(first.from);
            }
            view.dispatch(lint.setDiagnostics(view.state, translateDiagnostics(diagnostics, view)));
        }
    }, [diagnostics, view]);
}

/// editable
const extensionsCompart = new state.Compartment();
function reconfigure(extensions) {
    return extensionsCompart.reconfigure(extensions);
}
function useExtensionsProp(extensions, view) {
    const [compart] = React.useState(() => extensionsCompart.of(extensions));
    React.useEffect(() => {
        view === null || view === void 0 ? void 0 : view.dispatch({
            effects: [reconfigure(extensions)],
        });
    }, [extensions, view]);
    return compart;
}

function useChangedValue(value, view) {
    React.useEffect(() => {
        if (view && !isSameTextAccordingToDoc(view, value)) {
            view.dispatch({
                changes: [
                    {
                        from: 0,
                        to: view.state.doc.length,
                        insert: value,
                    },
                ],
            });
        }
    }, [value, view]);
}
function listenValueChangeAndInvokeCallback(handler) {
    return view.EditorView.updateListener.of(({ docChanged, state }) => {
        const editable = editableCompart.get(state);
        if (docChanged &&
            handler &&
            editable &&
            // eslint-disable-next-line @typescript-eslint/ban-ts-comment
            // @ts-ignore
            state.facet(editable.facet)) {
            handler(state.doc.toJSON().join(state.lineBreak));
        }
    });
}

/// placeholder
const textOfPlaceholder = state.Facet.define({
    combine(input) {
        return input[input.length - 1];
    },
});
const placeholderPlugin = view.ViewPlugin.fromClass(class {
    constructor(view) {
        this.view = view;
        this.placeholder = view.state.facet(textOfPlaceholder);
        if (!this.placeholder) {
            return;
        }
        this.dom = this.initPlaceholderMask(view);
        this.updateMaskOverlay(this.placeholder);
    }
    update(update) {
        if (update.docChanged && update.state.doc.length === 0) {
            if (!this.placeholder) {
                return;
            }
            this.updateMaskOverlay(this.view.state.facet(textOfPlaceholder));
        }
        if (update.state.doc.length !== 0 && this.dom) {
            this.dom.innerHTML = "";
        }
    }
    focus() {
        this.view.contentDOM.focus();
    }
    bindClickEvent(dom) {
        dom.addEventListener("click", this.focus.bind(this));
    }
    destory() {
        this.view.dom.removeChild(this.dom);
    }
    initPlaceholderMask(view) {
        const mask = document.createElement("div");
        mask.classList.add("cm-placeholder-mask");
        view.scrollDOM.appendChild(mask);
        this.bindClickEvent(mask);
        return mask;
    }
    updateMaskOverlay(text) {
        if (typeof text === "string") {
            this.dom.innerText = text;
        }
        else {
            this.dom.innerHTML = text
                .map((str) => {
                return `<div class="cm-line" style="cursor: text" >${str}</div>`;
            })
                .join("");
        }
    }
}, {});
const baseTheme$1 = view.EditorView.baseTheme({
    ".cm-placeholder-mask": {
        position: "absolute",
        top: 0,
        left: "30px",
        lineHeight: 1.9,
        opacity: 0.4,
    },
});
function placeholder(placeholderText) {
    return [
        textOfPlaceholder.of(placeholderText),
        placeholderPlugin,
        baseTheme$1,
    ];
}

const pluginCompart = new state.Compartment();
function usePlaceholderProp(text, view) {
    const [compart] = React.useState(pluginCompart.of(placeholder(text)));
    React.useEffect(() => {
        view === null || view === void 0 ? void 0 : view.dispatch({
            effects: [pluginCompart.reconfigure(placeholder(text))],
        });
    }, [text, view]);
    return compart;
}

const ios = typeof navigator != "undefined" &&
    !/Edge\/(\d+)/.exec(navigator.userAgent) && /Apple Computer/.test(navigator.vendor) &&
    (/Mobile\/\w+/.test(navigator.userAgent) || navigator.maxTouchPoints > 2);
const Outside = "-10000px";
const tooltipPlugin = view.ViewPlugin.fromClass(class {
    constructor(view) {
        this.view = view;
        this.inView = true;
        this.measureReq = { read: this.readMeasure.bind(this), write: this.writeMeasure.bind(this), key: this };
        this.input = view.state.facet(showTooltip);
        this.tooltips = this.input.filter(t => t);
        this.tooltipViews = this.tooltips.map(tp => this.createTooltip(tp));
    }
    update(update) {
        let input = update.state.facet(showTooltip);
        if (input == this.input) {
            for (let t of this.tooltipViews)
                if (t.update)
                    t.update(update);
        }
        else {
            let tooltips = input.filter(x => x);
            let views = [];
            for (let i = 0; i < tooltips.length; i++) {
                let tip = tooltips[i], known = -1;
                if (!tip)
                    continue;
                for (let i = 0; i < this.tooltips.length; i++) {
                    let other = this.tooltips[i];
                    if (other && other.create == tip.create)
                        known = i;
                }
                if (known < 0) {
                    views[i] = this.createTooltip(tip);
                }
                else {
                    let tooltipView = views[i] = this.tooltipViews[known];
                    if (tooltipView.update)
                        tooltipView.update(update);
                }
            }
            for (let t of this.tooltipViews)
                if (views.indexOf(t) < 0)
                    t.dom.remove();
            this.input = input;
            this.tooltips = tooltips;
            this.tooltipViews = views;
            this.maybeMeasure();
        }
    }
    createTooltip(tooltip) {
        let tooltipView = tooltip.create(this.view);
        tooltipView.dom.classList.add("cm-tooltip");
        // FIXME drop this on the next breaking release
        if (tooltip.class)
            tooltipView.dom.classList.add(tooltip.class);
        tooltipView.dom.style.top = Outside;
        this.view.dom.appendChild(tooltipView.dom);
        if (tooltipView.mount)
            tooltipView.mount(this.view);
        return tooltipView;
    }
    destroy() {
        for (let { dom } of this.tooltipViews)
            dom.remove();
    }
    readMeasure() {
        return {
            editor: this.view.dom.getBoundingClientRect(),
            pos: this.tooltips.map(t => this.view.coordsAtPos(t.pos)),
            size: this.tooltipViews.map(({ dom }) => dom.getBoundingClientRect()),
            innerWidth: window.innerWidth,
            innerHeight: window.innerHeight
        };
    }
    writeMeasure(measured) {
        let { editor } = measured;
        for (let i = 0; i < this.tooltipViews.length; i++) {
            let tooltip = this.tooltips[i], tView = this.tooltipViews[i], { dom } = tView;
            let pos = measured.pos[i], size = measured.size[i];
            // Hide tooltips that are outside of the editor.
            if (!pos || pos.bottom <= editor.top || pos.top >= editor.bottom || pos.right <= editor.left || pos.left >= editor.right) {
                dom.style.top = Outside;
                continue;
            }
            let width = size.right - size.left, height = size.bottom - size.top;
            let left = this.view.textDirection == view.Direction.LTR ? Math.min(pos.left, measured.innerWidth - width)
                : Math.max(0, pos.left - width);
            let above = !!tooltip.above;
            if (!tooltip.strictSide &&
                (above ? pos.top - (size.bottom - size.top) < 0 : pos.bottom + (size.bottom - size.top) > measured.innerHeight))
                above = !above;
            if (ios) {
                dom.style.top = ((above ? pos.top - height : pos.bottom) - editor.top) + "px";
                dom.style.left = (left - editor.left) + "px";
                dom.style.position = "absolute";
            }
            else {
                dom.style.top = (above ? pos.top - height : pos.bottom) + "px";
                dom.style.left = left + "px";
            }
            dom.classList.toggle("cm-tooltip-above", above);
            dom.classList.toggle("cm-tooltip-below", !above);
            if (tView.positioned)
                tView.positioned();
        }
    }
    maybeMeasure() {
        if (this.tooltips.length) {
            if (this.view.inView || this.inView)
                this.view.requestMeasure(this.measureReq);
            this.inView = this.view.inView;
        }
    }
}, {
    eventHandlers: {
        scroll() { this.maybeMeasure(); }
    }
});
const baseTheme = view.EditorView.baseTheme({
    ".cm-tooltip": {
        position: "fixed",
        zIndex: 100
    },
    "&light .cm-tooltip": {
        border: "1px solid #ddd",
        backgroundColor: "#f5f5f5"
    },
    "&dark .cm-tooltip": {
        backgroundColor: "#333338",
        color: "white"
    }
});
/**
Behavior by which an extension can provide a tooltip to be shown.
*/
const showTooltip = state.Facet.define({
    enables: [tooltipPlugin, baseTheme]
});
const HoverTime = 750, HoverMaxDist = 6;
class HoverPlugin {
    constructor(view, source, field, setHover) {
        this.view = view;
        this.source = source;
        this.field = field;
        this.setHover = setHover;
        this.lastMouseMove = null;
        this.hoverTimeout = -1;
        this.restartTimeout = -1;
        this.pending = null;
        this.checkHover = this.checkHover.bind(this);
        view.dom.addEventListener("mouseleave", this.mouseleave = this.mouseleave.bind(this));
        view.dom.addEventListener("mousemove", this.mousemove = this.mousemove.bind(this));
    }
    update() {
        if (this.pending) {
            this.pending = null;
            clearTimeout(this.restartTimeout);
            this.restartTimeout = setTimeout(() => this.startHover(), 20);
        }
    }
    get active() {
        return this.view.state.field(this.field);
    }
    checkHover() {
        this.hoverTimeout = -1;
        if (this.active)
            return;
        let now = Date.now(), lastMove = this.lastMouseMove;
        if (now - lastMove.timeStamp < HoverTime)
            this.hoverTimeout = setTimeout(this.checkHover, HoverTime - (now - lastMove.timeStamp));
        else
            this.startHover();
    }
    startHover() {
        var _a;
        clearTimeout(this.restartTimeout);
        let lastMove = this.lastMouseMove;
        let coords = { x: lastMove.clientX, y: lastMove.clientY };
        let pos = this.view.contentDOM.contains(lastMove.target)
            ? this.view.posAtCoords(coords) : null;
        if (pos == null)
            return;
        let posCoords = this.view.coordsAtPos(pos);
        if (posCoords == null || coords.y < posCoords.top || coords.y > posCoords.bottom ||
            coords.x < posCoords.left - this.view.defaultCharacterWidth ||
            coords.x > posCoords.right + this.view.defaultCharacterWidth)
            return;
        let bidi = this.view.bidiSpans(this.view.state.doc.lineAt(pos)).find(s => s.from <= pos && s.to >= pos);
        let rtl = bidi && bidi.dir == view.Direction.RTL ? -1 : 1;
        let open = this.source(this.view, pos, (coords.x < posCoords.left ? -rtl : rtl));
        if ((_a = open) === null || _a === void 0 ? void 0 : _a.then) {
            let pending = this.pending = { pos };
            open.then(result => {
                if (this.pending == pending) {
                    this.pending = null;
                    if (result)
                        this.view.dispatch({ effects: this.setHover.of(result) });
                }
            }, e => view.logException(this.view.state, e, "hover tooltip"));
        }
        else if (open) {
            this.view.dispatch({ effects: this.setHover.of(open) });
        }
    }
    mousemove(event) {
        var _a;
        this.lastMouseMove = event;
        if (this.hoverTimeout < 0)
            this.hoverTimeout = setTimeout(this.checkHover, HoverTime);
        let tooltip = this.active;
        if (tooltip && !isInTooltip(event.target) || this.pending) {
            let { pos } = tooltip || this.pending, end = (_a = tooltip === null || tooltip === void 0 ? void 0 : tooltip.end) !== null && _a !== void 0 ? _a : pos;
            if ((pos == end ? this.view.posAtCoords({ x: event.clientX, y: event.clientY }) != pos
                : !isOverRange(this.view, pos, end, event.clientX, event.clientY, HoverMaxDist))) {
                this.view.dispatch({ effects: this.setHover.of(null) });
                this.pending = null;
            }
        }
    }
    mouseleave() {
        clearTimeout(this.hoverTimeout);
        this.hoverTimeout = -1;
        if (this.active)
            this.view.dispatch({ effects: this.setHover.of(null) });
    }
    destroy() {
        clearTimeout(this.hoverTimeout);
        this.view.dom.removeEventListener("mouseleave", this.mouseleave);
        this.view.dom.removeEventListener("mousemove", this.mousemove);
    }
}
function isInTooltip(elt) {
    for (let cur = elt; cur; cur = cur.parentNode)
        if (cur.nodeType == 1 && cur.classList.contains("cm-tooltip"))
            return true;
    return false;
}
function isOverRange(view, from, to, x, y, margin) {
    let range = document.createRange();
    let fromDOM = view.domAtPos(from), toDOM = view.domAtPos(to);
    range.setEnd(toDOM.node, toDOM.offset);
    range.setStart(fromDOM.node, fromDOM.offset);
    let rects = range.getClientRects();
    range.detach();
    for (let i = 0; i < rects.length; i++) {
        let rect = rects[i];
        let dist = Math.max(rect.top - y, y - rect.bottom, rect.left - x, x - rect.right);
        if (dist <= margin)
            return true;
    }
    return false;
}
/**
Enable a hover tooltip, which shows up when the pointer hovers
over ranges of text. The callback is called when the mouse hovers
over the document text. It should, if there is a tooltip
associated with position `pos` return the tooltip description
(either directly or in a promise). The `side` argument indicates
on which side of the position the pointer is—it will be -1 if the
pointer is before the position, 1 if after the position.
*/
function hoverTooltip(source, options = {}) {
    const setHover = state.StateEffect.define();
    const hoverState = state.StateField.define({
        create() { return null; },
        update(value, tr) {
            if (value && (options.hideOnChange && (tr.docChanged || tr.selection)))
                return null;
            for (let effect of tr.effects)
                if (effect.is(setHover))
                    return effect.value;
            if (value && tr.docChanged) {
                let newPos = tr.changes.mapPos(value.pos, -1, state.MapMode.TrackDel);
                if (newPos == null)
                    return null;
                let copy = Object.assign(Object.create(null), value);
                copy.pos = newPos;
                if (value.end != null)
                    copy.end = tr.changes.mapPos(value.end);
                return copy;
            }
            return value;
        },
        provide: f => showTooltip.from(f)
    });
    return [
        hoverState,
        view.ViewPlugin.define(view => new HoverPlugin(view, source, hoverState, setHover))
    ];
}

var functions = [
    {
        title: "Comparison Functions",
        label: ["Comparison functions", "Description"],
        data: [
            {
                key: "value1 = value2",
                value: "Returns TRUE if value1 is equal to value2; returns UNKNOWN if value1 or value2 is NULL.",
            },
            {
                key: "value1 <> value2",
                value: "Returns TRUE if value1 is not equal to value2; returns UNKNOWN if value1 or value2 is NULL.",
            },
            {
                key: "value1 > value2",
                value: "Returns TRUE if value1 is greater than value2; returns UNKNOWN if value1 or value2 is NULL.",
            },
            {
                key: "value1 >= value2",
                value: "Returns TRUE if value1 is greater than or equal to value2; returns UNKNOWN if value1 or value2 is NULL.",
            },
            {
                key: "value1 < value2",
                value: "Returns TRUE if value1 is less than value2; returns UNKNOWN if value1 or value2 is NULL.",
            },
            {
                key: "value1 <= value2",
                value: "Returns TRUE if value1 is less than or equal to value2; returns UNKNOWN if value1 or value2 is NULL.",
            },
            {
                key: "value IS NULL",
                value: "Returns TRUE if value is NULL.",
            },
            {
                key: "value IS NOT NULL",
                value: "Returns TRUE if value is not NULL.",
            },
            {
                key: "value1 IS DISTINCT FROM value2",
                value: "Returns TRUE if two values are not equal. NULL values are treated as identical here.\nE.g., 1 IS DISTINCT FROM NULL returns TRUE; NULL IS DISTINCT FROM NULL returns FALSE.",
            },
            {
                key: "value1 IS NOT DISTINCT FROM value2",
                value: "Returns TRUE if two values are equal. NULL values are treated as identical here.\nE.g., 1 IS NOT DISTINCT FROM NULL returns FALSE; NULL IS NOT DISTINCT FROM NULL returns TRUE.",
            },
            {
                key: "value1 BETWEEN [ ASYMMETRIC | SYMMETRIC ] value2 AND value3",
                value: "By default (or with the ASYMMETRIC keyword), returns TRUE if value1 is greater than or equal to value2 and less than or equal to value3. With the SYMMETRIC keyword, returns TRUE if value1 is inclusively between value2 and value3. When either value2 or value3 is NULL, returns FALSE or UNKNOWN.\nE.g., 12 BETWEEN 15 AND 12 returns FALSE; 12 BETWEEN SYMMETRIC 15 AND 12 returns TRUE; 12 BETWEEN 10 AND NULL returns UNKNOWN; 12 BETWEEN NULL AND 10 returns FALSE; 12 BETWEEN SYMMETRIC NULL AND 12 returns UNKNOWN.",
            },
            {
                key: "value1 NOT BETWEEN [ ASYMMETRIC | SYMMETRIC ] value2 AND value3",
                value: "By default (or with the ASYMMETRIC keyword), returns TRUE if value1 is less than value2 or greater than value3. With the SYMMETRIC keyword, returns TRUE if value1 is not inclusively between value2 and value3. When either value2 or value3 is NULL, returns TRUE or UNKNOWN.\nE.g., 12 NOT BETWEEN 15 AND 12 returns TRUE; 12 NOT BETWEEN SYMMETRIC 15 AND 12 returns FALSE; 12 NOT BETWEEN NULL AND 15 returns UNKNOWN; 12 NOT BETWEEN 15 AND NULL returns TRUE; 12 NOT BETWEEN SYMMETRIC 12 AND NULL returns UNKNOWN.",
            },
            {
                key: "string1 LIKE string2 [ ESCAPE char ]",
                value: "Returns TRUE if string1 matches pattern string2; returns UNKNOWN if string1 or string2 is NULL. An escape character can be defined if necessary.\nNote: The escape character has not been supported yet.",
            },
            {
                key: "string1 NOT LIKE string2 [ ESCAPE char ]",
                value: "Returns TRUE if string1 does not match pattern string2; returns UNKNOWN if string1 or string2 is NULL. An escape character can be defined if necessary.\nNote: The escape character has not been supported yet.",
            },
            {
                key: "string1 SIMILAR TO string2 [ ESCAPE char ]",
                value: "Returns TRUE if string1 matches SQL regular expression string2; returns UNKNOWN if string1 or string2 is NULL. An escape character can be defined if necessary.\nNote: The escape character has not been supported yet.",
            },
            {
                key: "string1 NOT SIMILAR TO string2 [ ESCAPE char ]",
                value: "Returns TRUE if string1 does not match SQL regular expression string2; returns UNKNOWN if string1 or string2 is NULL. An escape character can be defined if necessary.\nNote: The escape character has not been supported yet.",
            },
            {
                key: "value1 IN (value2 [, value3]* )",
                value: "Returns TRUE if value1 exists in the given list (value2, value3, ...). When (value2, value3, ...). contains NULL, returns TRUE if the element can be found and UNKNOWN otherwise. Always returns UNKNOWN if value1 is NULL.\nE.g., 4 IN (1, 2, 3) returns FALSE; 1 IN (1, 2, NULL) returns TRUE; 4 IN (1, 2, NULL) returns UNKNOWN.",
            },
            {
                key: "value1 NOT IN (value2 [, value3]* )",
                value: "Returns TRUE if value1 does not exist in the given list (value2, value3, ...). When (value2, value3, ...). contains NULL, returns FALSE if value1 can be found and UNKNOWN otherwise. Always returns UNKNOWN if value1 is NULL.\nE.g., 4 NOT IN (1, 2, 3) returns TRUE; 1 NOT IN (1, 2, NULL) returns FALSE; 4 NOT IN (1, 2, NULL) returns UNKNOWN.",
            },
            {
                key: "EXISTS (sub-query)",
                value: "Returns TRUE if sub-query returns at least one row. Only supported if the operation can be rewritten in a join and group operation.\nNote: For streaming queries the operation is rewritten in a join and group operation. The required state to compute the query result might grow infinitely depending on the number of distinct input rows. Please provide a query configuration with valid retention interval to prevent excessive state size. See Query Configuration for details.",
            },
            {
                key: "value IN (sub-query)",
                value: "Returns TRUE if value is equal to a row returned by sub-query.\nNote: For streaming queries the operation is rewritten in a join and group operation. The required state to compute the query result might grow infinitely depending on the number of distinct input rows. Please provide a query configuration with valid retention interval to prevent excessive state size. See Query Configuration for details.",
            },
            {
                key: "value NOT IN (sub-query)",
                value: "Returns TRUE if value is not equal to every row returned by sub-query.\nNote: For streaming queries the operation is rewritten in a join and group operation. The required state to compute the query result might grow infinitely depending on the number of distinct input rows. Please provide a query configuration with valid retention interval to prevent excessive state size. See Query Configuration for details.",
            },
        ],
    },
    {
        title: "Logical Functions",
        label: ["Logical functions", "Description"],
        data: [
            {
                key: "boolean1 OR boolean2",
                value: "Returns TRUE if boolean1 is TRUE or boolean2 is TRUE. Supports three-valued logic.\nE.g., TRUE OR UNKNOWN returns TRUE.",
            },
            {
                key: "boolean1 AND boolean2",
                value: "Returns TRUE if boolean1 and boolean2 are both TRUE. Supports three-valued logic.\nE.g., TRUE AND UNKNOWN returns UNKNOWN.",
            },
            {
                key: "NOT boolean",
                value: "Returns TRUE if boolean is FALSE; returns FALSE if boolean is TRUE; returns UNKNOWN if boolean is UNKNOWN.",
            },
            {
                key: "boolean IS FALSE",
                value: "Returns TRUE if boolean is FALSE; returns FALSE if boolean is TRUE or UNKNOWN.",
            },
            {
                key: "boolean IS NOT FALSE",
                value: "Returns TRUE if boolean is TRUE or UNKNOWN; returns FALSE if boolean is FALSE.",
            },
            {
                key: "boolean IS TRUE",
                value: "Returns TRUE if boolean is TRUE; returns FALSE if boolean is FALSE or UNKNOWN.",
            },
            {
                key: "boolean IS NOT TRUE",
                value: "Returns TRUE if boolean is FALSE or UNKNOWN; returns FALSE if boolean is TRUE.",
            },
            {
                key: "boolean IS UNKNOWN",
                value: "Returns TRUE if boolean is UNKNOWN; returns FALSE if boolean is TRUE or FALSE.",
            },
            {
                key: "boolean IS NOT UNKNOWN",
                value: "Returns TRUE if boolean is TRUE or FALSE; returns FALSE if boolean is UNKNOWN.",
            },
        ],
    },
    {
        title: "Arithmetic Functions",
        label: ["Arithmetic functions", "Description"],
        data: [
            {
                key: "+ numeric",
                value: "Returns numeric.",
            },
            {
                key: "- numeric",
                value: "Returns negative numeric.",
            },
            {
                key: "numeric1 + numeric2",
                value: "Returns numeric1 plus numeric2.",
            },
            {
                key: "numeric1 - numeric2",
                value: "Returns numeric1 minus numeric2.",
            },
            {
                key: "numeric1 * numeric2",
                value: "Returns numeric1 multiplied by numeric2.",
            },
            {
                key: "numeric1 / numeric2",
                value: "Returns numeric1 divided by numeric2.",
            },
            {
                key: "POWER(numeric1, numeric2)",
                value: "Returns numeric1 raised to the power of numeric2.",
            },
            {
                key: "ABS(numeric)",
                value: "Returns the absolute value of numeric.",
            },
            {
                key: "MOD(numeric1, numeric2)",
                value: "Returns the remainder (modulus) of numeric1 divided by numeric2. The result is negative only if numeric1 is negative.",
            },
            {
                key: "SQRT(numeric)",
                value: "Returns the square root of numeric.",
            },
            {
                key: "LN(numeric)",
                value: "Returns the natural logarithm (base e) of numeric.",
            },
            {
                key: "LOG10(numeric)",
                value: "Returns the base 10 logarithm of numeric.",
            },
            {
                key: "LOG2(numeric)",
                value: "Returns the base 2 logarithm of numeric.",
            },
            {
                key: "LOG(numeric2)\nLOG(numeric1, numeric2)",
                value: "When called with one argument, returns the natural logarithm of numeric2. When called with two arguments, this function returns the logarithm of numeric2 to the base numeric1.\nNote: Currently, numeric2 must be greater than 0 and numeric1 must be greater than 1.",
            },
            {
                key: "EXP(numeric)",
                value: "Returns e raised to the power of numeric.",
            },
            {
                key: "CEIL(numeric)\nCEILING(numeric)",
                value: "Rounds numeric up, and returns the smallest number that is greater than or equal to numeric.",
            },
            {
                key: "FLOOR(numeric)",
                value: "Rounds numeric down, and returns the largest number that is less than or equal to numeric.",
            },
            {
                key: "SIN(numeric)",
                value: "Returns the sine of numeric.",
            },
            {
                key: "SINH(numeric)",
                value: "Returns the hyperbolic sine of numeric.\nThe return type is DOUBLE.",
            },
            {
                key: "COS(numeric)",
                value: "Returns the cosine of numeric.",
            },
            {
                key: "TAN(numeric)",
                value: "Returns the tangent of numeric.",
            },
            {
                key: "TANH(numeric)",
                value: "Returns the hyperbolic tangent of numeric.\nThe return type is DOUBLE.",
            },
            {
                key: "COT(numeric)",
                value: "Returns the cotangent of a numeric.",
            },
            {
                key: "ASIN(numeric)",
                value: "Returns the arc sine of numeric.",
            },
            {
                key: "ACOS(numeric)",
                value: "Returns the arc cosine of numeric.",
            },
            {
                key: "ATAN(numeric)",
                value: "Returns the arc tangent of numeric.",
            },
            {
                key: "ATAN2(numeric1, numeric2)",
                value: "Returns the arc tangent of a coordinate (numeric1, numeric2).",
            },
            {
                key: "COSH(numeric)",
                value: "Returns the hyperbolic cosine of NUMERIC.\nReturn value type is DOUBLE.",
            },
            {
                key: "DEGREES(numeric)",
                value: "Returns the degree representation of a radian numeric.",
            },
            {
                key: "RADIANS(numeric)",
                value: "Returns the radian representation of a degree numeric.",
            },
            {
                key: "SIGN(numeric)",
                value: "Returns the signum of numeric.",
            },
            {
                key: "ROUND(numeric, integer)",
                value: "Returns a number rounded to integer decimal places for numeric.",
            },
            {
                key: "PI",
                value: "Returns a value that is closer than any other values to pi.",
            },
            {
                key: "E()",
                value: "Returns a value that is closer than any other values to e.",
            },
            {
                key: "RAND()",
                value: "Returns a pseudorandom double value between 0.0 (inclusive) and 1.0 (exclusive).",
            },
            {
                key: "RAND(integer)",
                value: "Returns a pseudorandom double value between 0.0 (inclusive) and 1.0 (exclusive) with an initial seed integer. Two RAND functions will return identical sequences of numbers if they have the same initial seed.",
            },
            {
                key: "RAND_INTEGER(integer)",
                value: "Returns a pseudorandom integer value between 0 (inclusive) and integer (exclusive).",
            },
            {
                key: "RAND_INTEGER(integer1, integer2)",
                value: "Returns a pseudorandom integer value between 0 (inclusive) and the specified value (exclusive) with an initial seed. Two RAND_INTEGER functions will return identical sequences of numbers if they have the same initial seed and bound.",
            },
            {
                key: "UUID()",
                value: 'Returns an UUID (Universally Unique Identifier) string (e.g., "3d3c68f7-f608-473f-b60c-b0c44ad4cc4e") according to RFC 4122 type 4 (pseudo randomly generated) UUID. The UUID is generated using a cryptographically strong pseudo random number generator.',
            },
            {
                key: "BIN(integer)",
                value: "Returns a string representation of integer in binary format. Returns NULL if integer is NULL.\nE.g. BIN(4) returns '100' and BIN(12) returns '1100'.",
            },
            {
                key: "HEX(numeric)\nHEX(string)\n      ",
                value: 'Returns a string representation of an integer numeric value or a string in hex format. Returns NULL if the argument is NULL.\nE.g. a numeric 20 leads to "14", a numeric 100 leads to "64", a string "hello,world" leads to "68656C6C6F2C776F726C64".',
            },
            {
                key: "TRUNCATE(numeric1, integer2)",
                value: "Returns a numeric of truncated to integer2 decimal places. Returns NULL if numeric1 or integer2 is NULL.If integer2 is 0,the result has no decimal point or fractional part.integer2 can be negative to cause integer2 digits left of the decimal point of the value to become zero.This function can also pass in only one numeric1 parameter and not set Integer2 to use.If Integer2 is not set, the function truncates as if Integer2 were 0.\nE.g. truncate(42.345, 2) to 42.34. and truncate(42.345) to 42.0.",
            },
            {
                key: "PI()",
                value: "Returns the value of π (pi).\nOnly supported in blink planner.",
            },
        ],
    },
    {
        title: "String Functions",
        label: ["String functions", "Description"],
        data: [
            {
                key: "string1 || string2",
                value: "Returns the concatenation of string1 and string2.",
            },
            {
                key: "CHAR_LENGTH(string)\nCHARACTER_LENGTH(string)",
                value: "Returns the number of characters in string.",
            },
            {
                key: "UPPER(string)",
                value: "Returns string in uppercase.",
            },
            {
                key: "LOWER(string)",
                value: "Returns string in lowercase.",
            },
            {
                key: "POSITION(string1 IN string2)",
                value: "Returns the position (start from 1) of the first occurrence of string1 in string2; returns 0 if string1 cannot be found in string2.",
            },
            {
                key: "TRIM([ BOTH | LEADING | TRAILING ] string1 FROM string2)",
                value: "Returns a string that removes leading and/or trailing characters string1 from string2. By default, whitespaces at both sides are removed.",
            },
            {
                key: "LTRIM(string)",
                value: "Returns a string that removes the left whitespaces from string.\nE.g., LTRIM(' This is a test String.') returns \"This is a test String.\".",
            },
            {
                key: "RTRIM(string)",
                value: "Returns a string that removes the right whitespaces from string.\nE.g., RTRIM('This is a test String. ') returns \"This is a test String.\".",
            },
            {
                key: "REPEAT(string, integer)",
                value: "Returns a string that repeats the base string integer times.\nE.g., REPEAT('This is a test String.', 2) returns \"This is a test String.This is a test String.\".",
            },
            {
                key: "REGEXP_REPLACE(string1, string2, string3)",
                value: "Returns a string from string1 with all the substrings that match a regular expression string2 consecutively being replaced with string3.\nE.g., REGEXP_REPLACE('foobar', 'oo|ar', '') returns \"fb\".",
            },
            {
                key: "OVERLAY(string1 PLACING string2 FROM integer1 [ FOR integer2 ])",
                value: "Returns a string that replaces integer2 (string2's length by default) characters of string1 with string2 from position integer1.\nE.g., OVERLAY('This is an old string' PLACING ' new' FROM 10 FOR 5) returns \"This is a new string\"",
            },
            {
                key: "SUBSTRING(string FROM integer1 [ FOR integer2 ])",
                value: "Returns a substring of string starting from position integer1 with length integer2 (to the end by default).",
            },
            {
                key: "REPLACE(string1, string2, string3)",
                value: "Returns a new string which replaces all the occurrences of string2 with string3 (non-overlapping) from string1\nE.g., REPLACE('hello world', 'world', 'flink') returns \"hello flink\"; REPLACE('ababab', 'abab', 'z') returns \"zab\".",
            },
            {
                key: "REGEXP_EXTRACT(string1, string2[, integer])",
                value: "Returns a string from string1 which extracted with a specified regular expression string2 and a regex match group index integer.\nNote: The regex match group index starts from 1 and 0 means matching the whole regex. In addition, the regex match group index should not exceed the number of the defined groups.\nE.g. REGEXP_EXTRACT('foothebar', 'foo(.*?)(bar)', 2)\" returns \"bar\".",
            },
            {
                key: "INITCAP(string)",
                value: "Returns a new form of string with the first character of each word converted to uppercase and the rest characters to lowercase. Here a word means a sequences of alphanumeric characters.",
            },
            {
                key: "CONCAT(string1, string2,...)",
                value: "Returns a string that concatenates string1, string2, .... Returns NULL if any argument is NULL.\nE.g., CONCAT('AA', 'BB', 'CC') returns \"AABBCC\".",
            },
            {
                key: "CONCAT_WS(string1, string2, string3,...)",
                value: "Returns a string that concatenates string2, string3, ... with a separator string1. The separator is added between the strings to be concatenated. Returns NULL If string1 is NULL. Compared with CONCAT(), CONCAT_WS() automatically skips NULL arguments.\nE.g., CONCAT_WS('~', 'AA', NULL, 'BB', '', 'CC') returns \"AA~BB~~CC\".",
            },
            {
                key: "LPAD(string1, integer, string2)",
                value: "Returns a new string from string1 left-padded with string2 to a length of integer characters. If the length of string1 is shorter than integer, returns string1 shortened to integer characters.\nE.g., LPAD('hi',4,'??') returns \"??hi\"; LPAD('hi',1,'??') returns \"h\".",
            },
            {
                key: "RPAD(string1, integer, string2)",
                value: "Returns a new string from string1 right-padded with string2 to a length of integer characters. If the length of string1 is shorter than integer, returns string1 shortened to integer characters.\nE.g., RPAD('hi',4,'??') returns \"hi??\", RPAD('hi',1,'??') returns \"h\".",
            },
            {
                key: "FROM_BASE64(string)",
                value: "Returns the base64-decoded result from string; returns NULL if string is NULL.\nE.g., FROM_BASE64('aGVsbG8gd29ybGQ=') returns \"hello world\".",
            },
            {
                key: "TO_BASE64(string)",
                value: "Returns the base64-encoded result from string; returns NULL if string is NULL.\nE.g., TO_BASE64('hello world') returns \"aGVsbG8gd29ybGQ=\".",
            },
            {
                key: "ASCII(string)",
                value: "Returns the numeric value of the first character of string. Returns NULL if string is NULL.\nOnly supported in blink planner.\nE.g., ascii('abc') returns 97, and ascii(CAST(NULL AS VARCHAR)) returns NULL.",
            },
            {
                key: "CHR(integer)",
                value: "Returns the ASCII character having the binary equivalent to integer. If integer is larger than 255, we will get the modulus of integer divided by 255 first, and returns CHR of the modulus. Returns NULL if integer is NULL.\nOnly supported in blink planner.\nE.g., chr(97) returns a, chr(353) returns a, and ascii(CAST(NULL AS VARCHAR)) returns NULL.",
            },
            {
                key: "DECODE(binary, string)",
                value: "Decodes the first argument into a String using the provided character set (one of 'US-ASCII', 'ISO-8859-1', 'UTF-8', 'UTF-16BE', 'UTF-16LE', 'UTF-16'). If either argument is null, the result will also be null.\nOnly supported in blink planner.",
            },
            {
                key: "ENCODE(string1, string2)",
                value: "Encodes the string1 into a BINARY using the provided string2 character set (one of 'US-ASCII', 'ISO-8859-1', 'UTF-8', 'UTF-16BE', 'UTF-16LE', 'UTF-16'). If either argument is null, the result will also be null.\nOnly supported in blink planner.",
            },
            {
                key: "INSTR(string1, string2)",
                value: "Returns the position of the first occurrence of string2 in string1. Returns NULL if any of arguments is NULL.</p>\nOnly supported in blink planner.",
            },
            {
                key: "LEFT(string, integer)",
                value: "Returns the leftmost integer characters from the string. Returns EMPTY String if integer is negative. Returns NULL if any argument is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "RIGHT(string, integer)",
                value: "Returns the rightmost integer characters from the string. Returns EMPTY String if integer is negative. Returns NULL if any argument is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "LOCATE(string1, string2[, integer])",
                value: "Returns the position of the first occurrence of string1 in string2 after position integer. Returns 0 if not found. Returns NULL if any of arguments is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "PARSE_URL(string1, string2[, string3])",
                value: "Returns the specified part from the URL. Valid values for string2 include 'HOST', 'PATH', 'QUERY', 'REF', 'PROTOCOL', 'AUTHORITY', 'FILE', and 'USERINFO'. Returns NULL if any of arguments is NULL.\nE.g., parse_url('http://facebook.com/path1/p.php?k1=v1&k2=v2#Ref1', 'HOST'), returns 'facebook.com'.\nAlso a value of a particular key in QUERY can be extracted by providing the key as the third argument string3.\nE.g., parse_url('http://facebook.com/path1/p.php?k1=v1&k2=v2#Ref1', 'QUERY', 'k1') returns 'v1'.\nOnly supported in blink planner.",
            },
            {
                key: "REGEXP(string1, string2)",
                value: "Returns TRUE if any (possibly empty) substring of string1 matches the Java regular expression string2, otherwise FALSE. Returns NULL if any of arguments is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "REVERSE(string)",
                value: "Returns the reversed string. Returns NULL if string is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "SPLIT_INDEX(string1, string2, integer1)",
                value: "Splits string1 by the delimiter string2, returns the integerth (zero-based) string of the split strings. Returns NULL if integer is negative. Returns NULL if any of arguments is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "STR_TO_MAP(string1[, string2, string3]])",
                value: "Returns a map after splitting the string1 into key/value pairs using delimiters. string2 is the pair delimiter, default is ','. And string3 is the key-value delimiter, default is '='.\nOnly supported in blink planner.",
            },
            {
                key: "SUBSTR(string[, integer1[, integer2]])",
                value: "Returns a substring of string starting from position integer1 with length integer2 (to the end by default).\nOnly supported in blink planner.",
            },
        ],
    },
    {
        title: "Temporal Functions",
        label: ["Temporal functions", "Description"],
        data: [
            {
                key: "DATE string",
                value: 'Returns a SQL date parsed from string in form of "yyyy-MM-dd".',
            },
            {
                key: "TIME string",
                value: 'Returns a SQL time parsed from string in form of "HH:mm:ss".',
            },
            {
                key: "TIMESTAMP string",
                value: 'Returns a SQL timestamp parsed from string in form of "yyyy-MM-dd HH:mm:ss[.SSS]".',
            },
            {
                key: "INTERVAL string range",
                value: "Parses an interval string in the form \"dd hh:mm:ss.fff\" for SQL intervals of milliseconds or \"yyyy-mm\" for SQL intervals of months. An interval range might be DAY, MINUTE, DAY TO HOUR, or DAY TO SECOND for intervals of milliseconds; YEAR or YEAR TO MONTH for intervals of months.\nE.g., INTERVAL '10 00:00:00.004' DAY TO SECOND, INTERVAL '10' DAY, or INTERVAL '2-10' YEAR TO MONTH return intervals.",
            },
            {
                key: "CURRENT_DATE",
                value: "Returns the current SQL date in the UTC time zone.",
            },
            {
                key: "CURRENT_TIME",
                value: "Returns the current SQL time in the UTC time zone.",
            },
            {
                key: "CURRENT_TIMESTAMP",
                value: "Returns the current SQL timestamp in the UTC time zone.",
            },
            {
                key: "LOCALTIME",
                value: "Returns the current SQL time in local time zone.",
            },
            {
                key: "LOCALTIMESTAMP",
                value: "Returns the current SQL timestamp in local time zone.",
            },
            {
                key: "EXTRACT(timeintervalunit FROM temporal)",
                value: "Returns a long value extracted from the timeintervalunit part of temporal.\nE.g., EXTRACT(DAY FROM DATE '2006-06-05') returns 5.",
            },
            {
                key: "YEAR(date)",
                value: "Returns the year from SQL date date. Equivalent to EXTRACT(YEAR FROM date).\nE.g., YEAR(DATE '1994-09-27') returns 1994.",
            },
            {
                key: "QUARTER(date)",
                value: "Returns the quarter of a year (an integer between 1 and 4) from SQL date date. Equivalent to EXTRACT(QUARTER FROM date).\nE.g., QUARTER(DATE '1994-09-27') returns 3.",
            },
            {
                key: "MONTH(date)",
                value: "Returns the month of a year (an integer between 1 and 12) from SQL date date. Equivalent to EXTRACT(MONTH FROM date).\nE.g., MONTH(DATE '1994-09-27') returns 9.",
            },
            {
                key: "WEEK(date)",
                value: "Returns the week of a year (an integer between 1 and 53) from SQL date date. Equivalent to EXTRACT(WEEK FROM date).\nE.g., WEEK(DATE '1994-09-27') returns 39.",
            },
            {
                key: "DAYOFYEAR(date)",
                value: "Returns the day of a year (an integer between 1 and 366) from SQL date date. Equivalent to EXTRACT(DOY FROM date).\nE.g., DAYOFYEAR(DATE '1994-09-27') returns 270.",
            },
            {
                key: "DAYOFMONTH(date)",
                value: "Returns the day of a month (an integer between 1 and 31) from SQL date date. Equivalent to EXTRACT(DAY FROM date).\nE.g., DAYOFMONTH(DATE '1994-09-27') returns 27.",
            },
            {
                key: "DAYOFWEEK(date)",
                value: "Returns the day of a week (an integer between 1 and 7; Sunday = 1) from SQL date date.Equivalent to EXTRACT(DOW FROM date).\nE.g., DAYOFWEEK(DATE '1994-09-27') returns 3.",
            },
            {
                key: "HOUR(timestamp)",
                value: "Returns the hour of a day (an integer between 0 and 23) from SQL timestamp timestamp. Equivalent to EXTRACT(HOUR FROM timestamp).\nE.g., HOUR(TIMESTAMP '1994-09-27 13:14:15') returns 13.",
            },
            {
                key: "MINUTE(timestamp)",
                value: "Returns the minute of an hour (an integer between 0 and 59) from SQL timestamp timestamp. Equivalent to EXTRACT(MINUTE FROM timestamp).\nE.g., MINUTE(TIMESTAMP '1994-09-27 13:14:15') returns 14.",
            },
            {
                key: "SECOND(timestamp)",
                value: "Returns the second of a minute (an integer between 0 and 59) from SQL timestamp. Equivalent to EXTRACT(SECOND FROM timestamp).\nE.g., SECOND(TIMESTAMP '1994-09-27 13:14:15') returns 15.",
            },
            {
                key: "FLOOR(timepoint TO timeintervalunit)",
                value: "Returns a value that rounds timepoint down to the time unit timeintervalunit.\nE.g., FLOOR(TIME '12:44:31' TO MINUTE) returns 12:44:00.",
            },
            {
                key: "CEIL(timepoint TO timeintervalunit)",
                value: "Returns a value that rounds timepoint up to the time unit timeintervalunit.\nE.g., CEIL(TIME '12:44:31' TO MINUTE) returns 12:45:00.",
            },
            {
                key: "(timepoint1, temporal1) OVERLAPS (timepoint2, temporal2)",
                value: "Returns TRUE if two time intervals defined by (timepoint1, temporal1) and (timepoint2, temporal2) overlap. The temporal values could be either a time point or a time interval.\nE.g., (TIME '2:55:00', INTERVAL '1' HOUR) OVERLAPS (TIME '3:30:00', INTERVAL '2' HOUR) returns TRUE; (TIME '9:00:00', TIME '10:00:00') OVERLAPS (TIME '10:15:00', INTERVAL '3' HOUR) returns FALSE.",
            },
            {
                key: "DATE_FORMAT(timestamp, string)",
                value: "Attention for old planner This function has serious bugs and should not be used for now. Please implement a custom UDF instead or use EXTRACT as a workaround.\nFor blink planner, this converts timestamp to a value of string in the format specified by the date format string. The format string is compatible with Java's SimpleDateFormat.",
            },
            {
                key: "TIMESTAMPADD(timeintervalunit, interval, timepoint)",
                value: "Returns a new time value that adds a (signed) integer interval to timepoint. The unit for interval is given by the unit argument, which should be one of the following values: SECOND, MINUTE, HOUR, DAY, WEEK, MONTH, QUARTER, or YEAR.\nE.g., TIMESTAMPADD(WEEK, 1, DATE '2003-01-02') returns 2003-01-09.",
            },
            {
                key: "TIMESTAMPDIFF(timepointunit, timepoint1, timepoint2)",
                value: "Returns the (signed) number of timepointunit between timepoint1 and timepoint2. The unit for the interval is given by the first argument, which should be one of the following values: SECOND, MINUTE, HOUR, DAY, MONTH, or YEAR. See also the Time Interval and Point Unit Specifiers table.\nE.g., TIMESTAMPDIFF(DAY, TIMESTAMP '2003-01-02 10:00:00', TIMESTAMP '2003-01-03 10:00:00') leads to 1.",
            },
            {
                key: "CONVERT_TZ(string1, string2, string3)",
                value: "Converts a datetime string1 (with default ISO timestamp format 'yyyy-MM-dd HH:mm:ss') from time zone string2 to time zone string3. The format of time zone should be either an abbreviation such as \"PST\", a full name such as \"America/Los_Angeles\", or a custom ID such as \"GMT-8:00\".\nE.g., CONVERT('1970-01-01 00:00:00', 'UTC', 'America/Los_Angeles') returns '1969-12-31 16:00:00'.\nOnly supported in blink planner.",
            },
            {
                key: "FROM_UNIXTIME(numeric[, string])",
                value: "Returns a representation of the numeric argument as a value in string format (default is 'YYYY-MM-DD hh:mm:ss'). numeric is an internal timestamp value representing seconds since '1970-01-01 00:00:00' UTC, such as produced by the UNIX_TIMESTAMP() function. The return value is expressed in the session time zone (specified in TableConfig).\nE.g., FROM_UNIXTIME(44) returns '1970-01-01 09:00:44' if in UTC time zone, but returns '1970-01-01 09:00:44' if in 'Asia/Tokyo' time zone.\nOnly supported in blink planner.",
            },
            {
                key: "UNIX_TIMESTAMP()",
                value: "Gets current Unix timestamp in seconds. This function is not deterministic.\nOnly supported in blink planner.",
            },
            {
                key: "UNIX_TIMESTAMP(string1[, string2])",
                value: "Converts date time string string1 in format string2 (by default: yyyy-MM-dd HH:mm:ss if not specified) to Unix timestamp (in seconds), using the specified timezone in table config.\nOnly supported in blink planner.",
            },
            {
                key: "TO_DATE(string1[, string2])",
                value: "Converts a date string string1 with format string2 (by default 'yyyy-MM-dd') to a date.\nOnly supported in blink planner.",
            },
            {
                key: "TO_TIMESTAMP(string1[, string2])",
                value: "Converts date time string string1 with format string2 (by default: 'yyyy-MM-dd HH:mm:ss') under the session time zone (specified by TableConfig) to a timestamp.\nOnly supported in blink planner.",
            },
            {
                key: "NOW()",
                value: "Returns the current SQL timestamp in the UTC time zone. This function is not deterministic.\nOnly supported in blink planner.",
            },
        ],
    },
    {
        title: "Conditional Functions",
        label: ["Conditional functions", "Description"],
        data: [
            {
                key: "CASE value\nWHEN value1_1 [, value1_2 ]* THEN result1\n[ WHEN value2_1 [, value2_2 ]* THEN result2 ]*\n[ ELSE resultZ ]\nEND",
                value: "Returns resultX when the first time value is contained in (valueX_1, valueX_2, ...). When no value matches, returns resultZ if it is provided and returns NULL otherwise.",
            },
            {
                key: "CASE\nWHEN condition1 THEN result1\n[ WHEN condition2 THEN result2 ]*\n[ ELSE resultZ ]\nEND",
                value: "Returns resultX when the first conditionX is met. When no condition is met, returns resultZ if it is provided and returns NULL otherwise.",
            },
            {
                key: "NULLIF(value1, value2)",
                value: "Returns NULL if value1 is equal to value2; returns value1 otherwise.\nE.g., NULLIF(5, 5) returns NULL; NULLIF(5, 0) returns 5.",
            },
            {
                key: "COALESCE(value1, value2 [, value3 ]* )",
                value: "Returns the first value that is not NULL from value1, value2, ....\nE.g., COALESCE(NULL, 5) returns 5.",
            },
            {
                key: "IF(condition, true_value, false_value)",
                value: "Returns the true_value if condition is met, otherwise false_value.\nOnly supported in blink planner.\nE.g., IF(5 > 3, 5, 3) returns 5.",
            },
            {
                key: "IS_ALPHA(string)",
                value: "Returns true if all characters in string are letter, otherwise false.\nOnly supported in blink planner.",
            },
            {
                key: "IS_DECIMAL(string)",
                value: "Returns true if string can be parsed to a valid numeric, otherwise false.\nOnly supported in blink planner.",
            },
            {
                key: "IS_DIGIT(string)",
                value: "Returns true if all characters in string are digit, otherwise false.\nOnly supported in blink planner.",
            },
        ],
    },
    {
        title: "Type Conversion Functions",
        label: ["Type conversion functions", "Description"],
        data: [
            {
                key: "CAST(value AS type)",
                value: "Returns a new value being cast to type type. See the supported types here.\nE.g., CAST('42' AS INT) returns 42; CAST(NULL AS VARCHAR) returns NULL of type VARCHAR.",
            },
        ],
    },
    {
        title: "Collection Functions",
        label: ["Collection functions", "Description"],
        data: [
            {
                key: "CARDINALITY(array)",
                value: "Returns the number of elements in array.",
            },
            {
                key: "array ‘[’ integer ‘]’",
                value: "Returns the element at position integer in array. The index starts from 1.",
            },
            {
                key: "ELEMENT(array)",
                value: "Returns the sole element of array (whose cardinality should be one); returns NULL if array is empty. Throws an exception if array has more than one element.",
            },
            {
                key: "CARDINALITY(map)",
                value: "Returns the number of entries in map.",
            },
            {
                key: "map ‘[’ value ‘]’",
                value: "Returns the value specified by key value in map.",
            },
        ],
    },
    {
        title: "Value Construction Functions",
        label: ["Value construction functions", "Description"],
        data: [
            {
                key: "ROW(value1, [, value2]*)\n(value1, [, value2]*)",
                value: "Returns a row created from a list of values (value1, value2,...).",
            },
            {
                key: "ARRAY ‘[’ value1 [, value2 ]* ‘]’",
                value: "Returns an array created from a list of values (value1, value2, ...).",
            },
            {
                key: "MAP ‘[’ value1, value2 [, value3, value4 ]* ‘]’",
                value: "Returns a map created from a list of key-value pairs ((value1, value2), (value3, value4), ...).",
            },
        ],
    },
    {
        title: "Value Access Functions",
        label: ["Value access functions", "Description"],
        data: [
            {
                key: "tableName.compositeType.field",
                value: "Returns the value of a field from a Flink composite type (e.g., Tuple, POJO) by name.",
            },
            {
                key: "tableName.compositeType.*",
                value: "Returns a flat representation of a Flink composite type (e.g., Tuple, POJO) that converts each of its direct subtype into a separate field. In most cases the fields of the flat representation are named similarly to the original fields but with a dollar separator (e.g., mypojo$mytuple$f0).",
            },
        ],
    },
    {
        title: "Grouping Functions",
        label: ["Grouping functions", "Description"],
        data: [
            {
                key: "GROUP_ID()",
                value: "Returns an integer that uniquely identifies the combination of grouping keys.",
            },
            {
                key: "GROUPING(expression1 [, expression2]* )\nGROUPING_ID(expression1 [, expression2]* )",
                value: "Returns a bit vector of the given grouping expressions.",
            },
        ],
    },
    {
        title: "Hash Functions",
        label: ["Hash functions", "Description"],
        data: [
            {
                key: "MD5(string)",
                value: "Returns the MD5 hash of string as a string of 32 hexadecimal digits; returns NULL if string is NULL.",
            },
            {
                key: "SHA1(string)",
                value: "Returns the SHA-1 hash of string as a string of 40 hexadecimal digits; returns NULL if string is NULL.",
            },
            {
                key: "SHA224(string)",
                value: "Returns the SHA-224 hash of string as a string of 56 hexadecimal digits; returns NULL if string is NULL.",
            },
            {
                key: "SHA256(string)",
                value: "Returns the SHA-256 hash of string as a string of 64 hexadecimal digits; returns NULL if string is NULL.",
            },
            {
                key: "SHA384(string)",
                value: "Returns the SHA-384 hash of string as a string of 96 hexadecimal digits; returns NULL if string is NULL.",
            },
            {
                key: "SHA512(string)",
                value: "Returns the SHA-512 hash of string as a string of 128 hexadecimal digits; returns NULL if string is NULL.",
            },
            {
                key: "SHA2(string, hashLength)",
                value: "Returns the hash using the SHA-2 family of hash functions (SHA-224, SHA-256, SHA-384, or SHA-512). The first argument string is the string to be hashed and the second argument hashLength is the bit length of the result (224, 256, 384, or 512). Returns NULL if string or hashLength is NULL.",
            },
        ],
    },
    {
        title: "Aggregate Functions",
        label: ["Aggregate functions", "Description"],
        data: [
            {
                key: "COUNT([ ALL ] expression | DISTINCT expression1 [, expression2]*)",
                value: "By default or with ALL, returns the number of input rows for which expression is not NULL. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "COUNT(*)\nCOUNT(1)",
                value: "Returns the number of input rows.",
            },
            {
                key: "AVG([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the average (arithmetic mean) of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "SUM([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the sum of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "MAX([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the maximum value of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "MIN([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the minimum value of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "STDDEV_POP([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the population standard deviation of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "STDDEV_SAMP([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the sample standard deviation of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "VAR_POP([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the population variance (square of the population standard deviation) of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "VAR_SAMP([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns the sample variance (square of the sample standard deviation) of expression across all input rows. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "COLLECT([ ALL | DISTINCT ] expression)",
                value: "By default or with keyword ALL, returns a multiset of expression across all input rows. NULL values will be ignored. Use DISTINCT for one unique instance of each value.",
            },
            {
                key: "VARIANCE([ ALL | DISTINCT ] expression)",
                value: "Synonyms for VAR_SAMP().\nOnly supported in blink planner.",
            },
            {
                key: "RANK()",
                value: "Returns the rank of a value in a group of values. The result is one plus the number of rows preceding or equal to the current row in the ordering of the partition. The values will produce gaps in the sequence.\nOnly supported in blink planner.",
            },
            {
                key: "DENSE_RANK()",
                value: "Returns the rank of a value in a group of values. The result is one plus the previously assigned rank value. Unlike the function rank, dense_rank will not produce gaps in the ranking sequence.\nOnly supported in blink planner.",
            },
            {
                key: "ROW_NUMBER()",
                value: "Assigns a unique, sequential number to each row, starting with one, according to the ordering of rows within the window partition.\nROW_NUMBER and RANK are similar. ROW_NUMBER numbers all rows sequentially (for example 1, 2, 3, 4, 5). RANK provides the same numeric value for ties (for example 1, 2, 2, 4, 5).\nOnly supported in blink planner.",
            },
            {
                key: "LEAD(expression [, offset] [, default] )",
                value: "Returns the value of expression at the offsetth row after the current row in the window. The default value of offset is 1 and the default value of default is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "LAG(expression [, offset] [, default])",
                value: "Returns the value of expression at the offsetth row after the current row in the window. The default value of offset is 1 and the default value of default is NULL.\nOnly supported in blink planner.",
            },
            {
                key: "FIRST_VALUE(expression)",
                value: "Returns the first value in an ordered set of values.\nOnly supported in blink planner.",
            },
            {
                key: "LAST_VALUE(expression)",
                value: "Returns the last value in an ordered set of values.\nOnly supported in blink planner.",
            },
            {
                key: "LISTAGG(expression [, separator])",
                value: "Concatenates the values of string expressions and places separator values between them. The separator is not added at the end of string. The default value of separator is ','.\nOnly supported in blink planner.",
            },
        ],
    },
];

const funcs = {};
const transformToKeyValue = (funcsMap) => {
    //// add func
    funcsMap.forEach((funcsCategory) => {
        funcsCategory.data.forEach((item) => {
            const key = /([A-z]+)/.exec(item.key)[0];
            if (key) {
                funcs[key.toLowerCase()] = {
                    value: item.value,
                    rawKey: item.key,
                    key: key,
                };
            }
        });
    });
};
const stripQuote = (str) => { var _a, _b; return (_b = (_a = /^[`'"](.*)[`'"]$/.exec(str)) === null || _a === void 0 ? void 0 : _a[1]) !== null && _b !== void 0 ? _b : str; };
function tokenAt(tree, pos, side) {
    const node = tree.resolve(pos, side);
    /// 只对Identifier做一个hover提示
    if (node.name === "Identifier") ;
    return node;
}
const createHead = (text, cls = "cm-hover-tooltip-head") => {
    const head = document.createElement("div");
    head.classList.add(cls);
    head.innerText = text;
    return head;
};
const createbody = (body, cls = "cm-hover-tooltip-body") => {
    const b = document.createElement("div");
    b.classList.add(cls);
    const bodyHtml = body
        .map((item) => {
        var _a, _b;
        return item.name
            ? `
      <div class='cm-hover-tooltip-item'>
        <span>${item.name}</span>
        <span style='color: gray;margin-left: 6px;display:inline-block'>${(_b = (_a = stripQuote(item.type)) === null || _a === void 0 ? void 0 : _a.toLowerCase()) !== null && _b !== void 0 ? _b : "-"}</span>
      </div>`
            : "";
    })
        .join("");
    b.innerHTML = bodyHtml;
    return b;
};
// eslint-disable-next-line @typescript-eslint/ban-ts-comment
// @ts-ignore
transformToKeyValue(functions);
const hoverField = hoverTooltip((view, pos, side) => {
    const token = tokenAt(language.syntaxTree(view.state), pos, side);
    const tokenName = view.state.doc.sliceString(token.from, token.to);
    const localTableSource = view.state.field(sqlAnalysisField);
    const dataSource = localTableSource[tokenName];
    const funcsDescription = funcs[tokenName.toLowerCase()];
    if (!dataSource && !funcsDescription) {
        return null;
    }
    return {
        pos,
        above: true,
        strictSide: false,
        class: "cm-cursor-tooltip",
        create: () => {
            if (dataSource) {
                const dom = document.createElement("div");
                const head = createHead(tokenName);
                const body = createbody(dataSource);
                dom.appendChild(head);
                dom.appendChild(body);
                return {
                    dom,
                };
            }
            if (funcsDescription) {
                const dom = document.createElement("div");
                const head = createHead(`${funcsDescription.key}`);
                const body = document.createElement("div");
                body.innerText = funcsDescription.value;
                // +
                // "\nexample:\n" +
                // funcsDescription.rawKey
                dom.appendChild(head);
                dom.appendChild(body);
                return {
                    dom,
                };
            }
            return null;
        },
    };
}, {
    hideOnChange: false,
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore
    hoverTime: -1,
});
const cursorTooltipBaseTheme = view.EditorView.baseTheme({
    ".cm-tooltip.cm-cursor-tooltip": {
        backgroundColor: "#222",
        color: "white",
        border: "none",
        padding: "4px 8px",
        borderRadius: "4px",
        fontSize: "14px",
    },
    ".cm-hover-tooltip-head": {
        fontSize: "14px",
        borderBottom: "1px solid gray",
        color: "gray",
        padding: "4px 0",
    },
    ".cm-hover-tooltip-item": {
        display: "flex",
        justifyContent: "space-between",
        padding: "1px",
    },
});
function cursorTooltip() {
    return [hoverField, cursorTooltipBaseTheme];
}

const snippetsCompart = new state.Compartment();
function useSnippetsProp(extraSnippets = [], view) {
    const [compart] = React.useState(snippetsCompart.of(snippetsFacet.of(extraSnippets)));
    React.useEffect(() => {
        view === null || view === void 0 ? void 0 : view.dispatch({
            effects: [
                snippetsCompart.reconfigure(snippetsFacet.of(extraSnippets)),
            ],
        });
    }, [extraSnippets, view]);
    return compart;
}

/**
 * 对 codemirror6 的封装
 */
const ReactCodemirror = (props, ref) => {
    var _a;
    const { onChange } = props;
    const element = React.useRef();
    const editor = React.useRef();
    React.useImperativeHandle(ref, () => ({
        format(configs) {
            startFormat(editor.current, configs);
        },
    }), [editor]);
    const themeCompart = useThemeProp(props.theme || "dark", editor.current);
    const editCompart = useEditableProp(props.editable != undefined ? props.editable : true, editor.current);
    const languageCompart = useLanguageProp(props.language, props.langOptions, editor.current);
    const extensionsCompart = useExtensionsProp(props.extensions || [], editor.current);
    const placeholderCompart = usePlaceholderProp(props.placeholder, editor.current);
    const snippets = useSnippetsProp(props.snippets || [], editor.current);
    useDiagnostics(props.diagnostics, editor.current);
    useChangedValue(props.value, editor.current);
    useMount(() => {
        var _a;
        const state$1 = state.EditorState.create({
            doc: props.defaultValue || props.value || "",
            extensions: [
                basicSetup,
                editCompart,
                themeCompart,
                languageCompart,
                extensionsCompart,
                scrollToEdge({
                    onScrollToBottom() {
                        var _a;
                        (_a = props === null || props === void 0 ? void 0 : props.onScrollToBottom) === null || _a === void 0 ? void 0 : _a.call(props);
                    },
                    onScrollToTop() {
                        var _a;
                        (_a = props === null || props === void 0 ? void 0 : props.onScrollToTop) === null || _a === void 0 ? void 0 : _a.call(props);
                    },
                }),
                scrollOptions.of((_a = props.scrollOptions) !== null && _a !== void 0 ? _a : {}),
                props.language !== "log" && gutter.lineNumbers(),
                /// features
                MINIMAP_FLAG ,
                placeholderCompart,
                listenValueChangeAndInvokeCallback(onChange),
                keymapPrompt({ placement: "leftbottom" }),
                /// 使用快捷键进行格式化，只针对sql语言
                props.language === "sql" && cursorTooltip(),
                props.language === "sql" && snippets,
            ].filter(Boolean),
        });
        editor.current = new view.EditorView({
            state: state$1,
            parent: element.current,
        });
    });
    useUnmount(() => editor.current.destroy());
    return (React__default['default'].createElement("div", { ref: element, className: `codemirror-editor-body ${(_a = props.className) !== null && _a !== void 0 ? _a : ""}`, style: props.style, 
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore 接受css属性，兼容 @emotion/react 的属性
        css: props.css }));
};
var ReactCodemirror$1 = React.forwardRef(ReactCodemirror);

exports.ReactCodemirror = ReactCodemirror$1;
